/***************************************************************************
                          ccmdprocessor.h  -  command processor
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : P Jul 5 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@host.sk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
Current status of cCmdProcessor class (Sep 13, 2002):
 - most of required functionality implemented
 - new features will probably be added (macro support, ...)
 - implemented parts work properly
*/

#ifndef CCMDPROCESSOR_H
#define CCMDPROCESSOR_H

#include <qobject.h>

#include "csessionbase.h"

/**
class cCmdProcessor handles processing of command after ENTER has been
pressed - it splits it into individual commands, checks every command
against all aliases, applying them if necessary (including calling of macros),
then it gives it to cSendCmd for sending...

Whole strings can now be used as cmdseparator, macro caller and friends...
It is caller's responsibility to verify correctness of supplied strings,
that is, none of them is a prefix of another one, single-char and multi-char
strings do not collide and so... Behaviour is not defined for strings
that violate these rules.
  *@author Tomas Mecir
  */

class cCmdProcessor : public QObject, public cSessionBase {
   Q_OBJECT
public: 
  cCmdProcessor();
  ~cCmdProcessor();

  /** backslash (\) expansion yes/no */
  void setExpandBackslashes (bool val) { expandbackslashes = val; };
  /** backslash expansion in aliases/triggers only; only applies if
  backslash expansion is on */
  void setBackslashesOnlyInAT (bool val) { backslashAT = val; };
  
  void setCmdSeparatorString (QString str);
  void setSpeedWalkString (QString str);
  void setScriptString (QString str);
  void setMultiCommandString (QString str);

  void setAllowEmptyWalkStr (bool val);

  /** set spaces stripping */
  void setStripSpaces (bool strip) { stripSpaces = strip; };
signals:
  void newCommand (const QString &);
public slots:
  /** process input received from anywhere but the trigger */
  void processInput (const QString &text);
  /** Process input received from some trigger. We need two separate
    functions because it's possible to limit backslash expansion to
    aliases/triggers only. Aliases are expanded inside that function, so
    we don't need another separate function for these. */
  void processTriggerInput (const QString &text);
  /** enable/disable command parsing */
  void setParsing (bool value);
protected:
  /** get next command from the input line or alias/trigger text*/
  bool getNext (QString &from, QString &command, bool isAliasTrigger = false);

  /** called by processInput and processTrigger */
  void doProcessInput (const QString &next, bool isAliasTrigger,
      bool expandAliases = true);

  /** if this is a speed-walk command, returns length of speed-walk char
  or string (it returns 0 if it isn't speedwalk) */
  int isSpeedWalkCommand (QString command);
  /** if the command is speed-walk, this will process it, starting
  at position <pos> (computed by isSpeedWalkCommand */
  void processSpeedWalk (const QString &command, int pos);
  
  /** Looks if this is a macro call. If it is, then it also splits the
  command into parts (macro name and parameters).
  Note that macros are NOT scripts, although scripts are called using
  macros.*/
  bool isMacroCall (const QString &command, QString &mname,
    QString &params, bool isAliasTrigger);
  /** Evaluate macro <mname> with parameters <params> */
  void processMacroCall (const QString &mname, const QString &params,
      bool isAliasTrigger);
  /** Runs script <sname> with parameters <params> */
  void runScript (const QString &sname, const QString &params,
      bool isAliasTrigger);

  /** handles repeater (multi-command) character. !!! modifies
  parameter command if there is a repeater !!!
  @param command Where to look for the repeater, repeater sequence is
  stripped off the command, if any */
  unsigned int repeater (QString &command);

  /** will we expand backslashes? */
  bool willExpandBS (bool isAliasTrigger);
  
  /** fix (=remove) backslashes in command */
  void fixBackslashes (QString &command, bool isAliasTrigger);
  /** protect BSs before alias expansion */
  void protectBackslashes (QString &command, bool isAliasTrigger);
  /** unprotect those BSs */
  void unprotectBackslashes (QString &command, bool isAliasTrigger);
  
  bool parsing, expandbackslashes, backslashAT;
  bool stripSpaces;
  QString separstr, walkstr, scriptstr, multistr;
  bool allowemptywalkstr;
};

#endif
