/***************************************************************************
                          cconnection.cpp  -  handles connections
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : Po Jul 22 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@kmuddy.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cconnection.h"

#include <config.h>

#include <kmessagebox.h>
#include <klocale.h>
#include <kapplication.h>
#include <qdatetime.h>
#include <qtimer.h>

#include "kmuddy.h"
#include "cconsole.h"
#include "cconnectioninfo.h"
#include "cconnprefs.h"
#include "cpluginmanager.h"
#include "cprofiles.h"
#include "csession.h"
#include "csendcmd.h"
#include "ctelnet.h"

#include "caliaslist.h"
#include "cshortcutlist.h"
#include "ctimerlist.h"
#include "ctranscript.h"
#include "ctriggerlist.h"
#include "cwindowlist.h"


class cActionList;
class cScriptList;
class cVariableList;
class cVarTriggerList;
class cWindowList;

#include "dialogs/dlgprofilesettings.h"
#include "dialogs/dlgaliastrigger.h"
#include "dialogs/dlgscripts.h"
#include "dialogs/dlgtimers.h"
#include "dialogs/dlgshortcuts.h"
#include "dialogs/dlgvariables.h"
#include "dialogs/dlgvartriglist.h"
#include "dialogs/dlgwindows.h"

extern KMuddy *kmuddy;

#define AUTOSAVE_INTERVAL 5

cConnection::cConnection ()
{
  ci = 0;
  cp = 0;
  destroying = false;
  connecting = NULL;
  connclosed = false;
  canCT = false;

  saver = new QTimer;
  connect (saver, SIGNAL (timeout ()), this, SLOT (saverTimeout ()));
}

cConnection::~cConnection ()
{
  destroying = true;
  disconnect ();  //maybe we aren't connected, but disconnect() can handle it
  if (ci != 0)
  {
    delete ci;
    ci = 0;
  }
  delete saver;
}

void cConnection::showConnPrefsDialog ()
{
  if (cp == NULL) //not a profile-based connection - do nothing
    return;

  //so first we have to create the dialog...
  sdlg = new dlgProfileSettings (kmuddy);

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (sdlg, SIGNAL (okClicked()), this, SLOT (getSettingsFromDialog()));
  connect (sdlg, SIGNAL (applyClicked()), this, SLOT (getSettingsFromDialog()));

  //next we fill in its data
  putSettingsToDialog ();

  //dialog is ready - show it!
  sdlg->exec ();

  //further action is handled via slots issued by buttons, so that we only
  //have to destroy the dialog...
  delete sdlg;
}

void cConnection::getSettingsFromDialog ()
{
  cp->setAnsiColors (sdlg->useAnsi ());
  cp->setLimitTriggers (sdlg->limitTriggers ());
  cp->setLimitRepeater (sdlg->limitRepeater ());
  cp->setPromptLabel (sdlg->promptLabel ());
  cp->setLPMudStyle(sdlg->LPMudStyle());
  cp->setStatusPrompt(sdlg->statusPrompt());
  cp->setConsolePrompt(sdlg->consolePrompt());
  cp->setAutoAdvTranscript(sdlg->autoAdvTranscript());

  for (int i = 0; i < 10; i++)
    cp->setCmd (i, sdlg->movementCmd (i));
  cp->setQuit (sdlg->quitCmd ());
  cp->setScriptDir (sdlg->scriptDir ());
  cp->setWorkDir (sdlg->scriptWorkDir ());
  cp->setTranscriptDir (sdlg->transcriptDir ());

#ifdef HAVE_SOUND
  cp->setSoundDirs (sdlg->soundDirList());
  cp->setUseMSP (sdlg->useMSP());
  cp->setAlwaysMSP (sdlg->alwaysMSP());
  cp->setMidlineMSP (sdlg->midlineMSP());
#endif

  //inform plug-ins... not really useful as of now, but that will change...
  pluginmanager->event (sess, "settings-changed");
}

void cConnection::putSettingsToDialog ()
{
  sdlg->setUseAnsi (cp->ansiColors ());
  sdlg->setLimitTriggers (cp->limitTriggers());
  sdlg->setLimitRepeater (cp->limitRepeater());
  sdlg->setPromptLabel(cp->promptLabel ());
  sdlg->setLPMudStyle(cp->LPMudStyle());
  sdlg->setStatusPrompt(cp->statusPrompt());
  sdlg->setConsolePrompt(cp->consolePrompt());
  sdlg->setAutoAdvTranscript(cp->autoAdvTranscript());
  
  for (int i = 0; i < 10; i++)
    sdlg->setMovementCmd (i, cp->cmd (i));
  sdlg->setQuitCmd (cp->quit ());
  sdlg->setScriptDir (cp->scriptDir ());
  sdlg->setScriptWorkDir (cp->workDir ());
  sdlg->setTranscriptDir (cp->transcriptDir ());

#ifdef HAVE_SOUND
  sdlg->setSoundDirList (cp->soundDirs());
  sdlg->setUseMSP (cp->useMSP());
  sdlg->setAlwaysMSP (cp->alwaysMSP());
  sdlg->setMidlineMSP (cp->midlineMSP());
#endif
}

bool cConnection::handleDisconnect ()
{
  if (!isConnected())
    return true;    //no longer connected
  //ask if he really wants to disconnect
  if (KMessageBox::questionYesNoCancel (kmuddy,
        i18n("Do you want to close connection to "+sess->connection()->getConnectionInfo()->name()+"?"),
        i18n("Disconnect")) == KMessageBox::Yes)
  {
    disconnect ();
    return true;
  }
  return false;
}

void cConnection::establishConnection (cConnectionInfo *cci, bool sendNothing)
{
  if ((ci != cci) && (ci != 0))
  //old connection info remembered and it's not the current one
  {
    delete ci;
    ci = 0;
  }

  if (sess->telnet()->isConnected ())
    disconnect ();
  if (sess->telnet()->connectIt(cci->server(), cci->port ()))
  {
    //connection established - great!
    ci = cci;
    ci->setIsQuick (false);
    //we can no longer reconnect
    connclosed = false;
    canCT = false;
    cp = new cConnPrefs (cci->name (), sess); //this loads preferences for the connection
    updateMenus ();
    kmuddy->enableConnectedOptions ();
    kmuddy->enableAdvancedOptions ();

    //update caption
    caption = ((cci->login().length()>0) ?
          (cci->login() + "@" + cci->name()) : cci->name());
    kmuddy->setSessionName (kmuddy->sessionId(sess), caption);

    //inform plug-ins
    pluginmanager->event (sess, "connected", "profile");

    // Here be autologging
    if(cp->autoAdvTranscript())
    {
      sess->transcript()->load();
      sess->transcript()->startAdvTranscript();
    }

    //everything's ready - schedule sending of login sequence if needed...
    if (!sendNothing)
    {
      loginSeq = cci->connStr();
      seqIter = loginSeq.begin();
      if ((loginSeq.count() > 1) || ((loginSeq.count() == 1) && (loginSeq[0].length() > 0)))
      {
        connecting = new QTimer;
        connect (connecting, SIGNAL (timeout()), this, SLOT (sendLoginAndPassword()));
        sess->telnet()->waitingForData();
        connecting->start (400);
      }
    }

    //also schedule automatic saving of settings and objects
    saver->start (AUTOSAVE_INTERVAL * 60 * 1000);
    (KApplication::kApplication())->processEvents ();
  }
  else
  {
    //do not update reconnect info, so that we can still reconnect if needed

    QString error = i18n ("Could not establish connection.");
    error += "\n\n";
    error += i18n ("Reason: ");
    error += sess->telnet()->lastError();
    KMessageBox::sorry (kmuddy, error, i18n ("Connect"));
  }
}

void cConnection::sendLoginAndPassword ()
{
  //This function is used to send login sequence (usually login name and
  //password) to the MUD.
  //Problem: it's not very good at detecting login prompt - name can be sent
  //too early, causing password to be sent too early as well...
  //I can only hope that the server will handle that properly.

  cTelnet *telnet = sess->telnet();
  if (telnet == 0)
    return;
  static int wasData = false; //have some data already arrived?

  if (telnet->isConnected())
  {
    if (wasData)
    {
      if (telnet->newData() || (seqIter == loginSeq.begin()))
      //only send next command if something new has arrived (should be reply
      //to previous sent command). Also send it if it's the first command,
      //where the required data was received during previous call - used to
      //wait for intro screens and such).
      {
        //get next command from login sequence
        QString command = *seqIter;
        QString cmd = "", cmddisp = "";
        
        //expand $name and $password
        int len = command.length();
        QString varname = QString::null;
        
        bool invar = false;
        for (int i = 0; i < len; i++)
        {
          if (invar)
          {
            varname += command[i];
            if (varname == "$name")
            {
              varname = "";
              invar = false;
              cmd += ci->login();
              cmddisp += ci->login();
            }
            if (varname == "$password")
            {
              varname = "";
              invar = false;
              cmd += ci->password();
              cmddisp += i18n("(password is hidden)");
            }
          }
          else
            if (command[i].latin1() == '$')
            {
              invar = true;
              varname = "$";
            }
            else
            {
              cmd += command[i];
              cmddisp += command[i];
            }
        }

        //send resulting command and display notification
        telnet->sendData(cmd);
        telnet->sendData("\n");
        sess->console()->addCommand (cmddisp);
        telnet->waitingForData();

        //advance to next command (that will be sent on next call)
        ++seqIter;

        //finish execution if there's nothing more to send
        if (seqIter == loginSeq.end())
        {
          connecting->stop ();
          //I have no idea why this is needed
          QObject::disconnect (connecting, SIGNAL (timeout()), this, SLOT (sendLoginAndPassword()));
          delete connecting;
          connecting = 0;
          wasData = false;
        }
      }
    }
    else
      if (telnet->newData())
      {
        telnet->waitingForData();
        wasData = true;
      }
  }
  else
  {
    connecting->stop ();
    //I have no idea why this is needed
    QObject::disconnect (connecting, SIGNAL (timeout()), this, SLOT (sendLoginAndPassword()));
    delete connecting;
    connecting = NULL;
    wasData = false;
  }
}

void cConnection::establishQuickConnection (cConnectionInfo *cci)
{
  if (sess->telnet()->isConnected ())
    disconnect ();
  if (sess->telnet()->connectIt(cci->server(), cci->port ()))
  {
    //connection established - great!
    ci = cci;
    ci->setIsQuick (true);
    cp = NULL;
    connclosed = false;
    canCT = false;
    updateMenus ();
    kmuddy->enableConnectedOptions ();
    kmuddy->disableAdvancedOptions ();

    //update application caption
    //update caption
    caption = cci->server();
    kmuddy->setSessionName (kmuddy->sessionId(sess), caption);

    //inform plug-ins
    pluginmanager->event (sess, "connected", "quick");
  }
  else
  {
    QString error = i18n ("Could not establish connection.");
    error += "\n\n";
    error += i18n ("Reason: ");
    error += sess->telnet()->lastError();
    KMessageBox::sorry (kmuddy, error, i18n ("QuickConnect"));
  }
}

bool cConnection::isConnected ()
{
  return sess->telnet()->isConnected ();
}

void cConnection::disconnect ()
{
  if (connecting != NULL)
  {
    connecting->stop ();
    delete connecting;
    connecting = NULL;
  }
  saver->stop ();
  if (cp != 0)
  {
    cp->save ();    //save all settings etc.
    delete cp;
    cp = 0;
  }

  //inform plug-ins
  if (pluginmanager)
    pluginmanager->event (sess, "disconnected");

  //clear command queue, so that pending commands don't close the tab
  //immediately
  sess->sendcmd()->clearCommandQueue ();

/*
//CODE DISABLED - we need connection info in case of reconnecting
  if (ci != 0)
  {
    delete ci;
    ci = 0;
  }
*/
  if (!destroying)
  {
    kmuddy->disableConnectedOptions ();
    kmuddy->disableAdvancedOptions ();
  }
  if (sess->telnet() != NULL)
    if (sess->telnet()->isConnected())
      sess->telnet()->disconnect (false);
}

void cConnection::reconnect ()
{
  if (ci == 0)
  {
    //We should never get here; Reconnect should ONLY be avail when it makes
    //sense; but you never know...
    KMessageBox::sorry (kmuddy, i18n ("No previous connection - cannot reconnect."), i18n ("Reconnect"));
    return;     //cannot reconnect - no connection info remembered!
  }

  //inform plug-ins
  pluginmanager->event (sess, "reconnecting");

  //reconnect...
  if (ci->isQuick())
    establishQuickConnection (ci);
  else
    establishConnection (ci);
}

void cConnection::handleAliasesDialog ()
{
  //not a profile-based connection => no aliases
  if (cp == NULL)
    return;

  cAliasList *al = cp->aliasList();
  //no aliases class (???) => nothing to do
  if (al == NULL)
    return;

  //so first we have to create the dialog...
  dlgAliasTrigger *adlg = new dlgAliasTrigger (al, kmuddy);

  //dialog is ready - show it!
  adlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete adlg;
}

void cConnection::handleTriggersDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  //not a profile-based connection => no triggers
  if (cp == NULL)
    return;

  cTriggerList *tl = cp->triggerList();
  //no triggers class (???) => nothing to do
  if (tl == NULL)
    return;

  //so first we have to create the dialog...
  dlgAliasTrigger *tdlg = new dlgAliasTrigger (tl, kmuddy);

  //dialog is ready - show it!
  tdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete tdlg;
}

void cConnection::handleActionsDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  //not a profile-based connection => no actions
  if (cp == NULL)
    return;

  cActionList *al = cp->actionList();
  //no actions class (???) => nothing to do
  if (al == NULL)
    return;

  //so first we have to create the dialog...
  dlgAliasTrigger *tdlg = new dlgAliasTrigger (al, kmuddy);

  //dialog is ready - show it!
  tdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete tdlg;
}

void cConnection::handleScriptsDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  //not a profile-based connection => no scripts
  if (cp == NULL)
    return;

  cScriptList *sl = cp->scriptList();
  //no scripts class (???) => nothing to do
  if (sl == NULL)
    return;

  //so first we have to create the dialog...
  dlgScripts *scdlg = new dlgScripts (sl, kmuddy);

  //dialog is ready - show it!
  scdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete scdlg;
}

void cConnection::handleTimersDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  //not a profile-based connection => no timers
  if (cp == NULL)
    return;

  cTimerList *tl = cp->timerList();
  //no timers class (???) => nothing to do
  if (tl == NULL)
    return;

  //so first we have to create the dialog...
  dlgTimers *timdlg = new dlgTimers (tl, kmuddy);

  //dialog is ready - show it!
  timdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete timdlg;
}

void cConnection::handleShortcutsDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  //not a profile-based connection => no short-cuts
  if (cp == NULL)
    return;

  cShortcutList *sl = cp->shortcutList();
  //no shortcuts class (???) => nothing to do
  if (sl == NULL)
    return;

  //so first we have to create the dialog...
  dlgShortcuts *shdlg = new dlgShortcuts (sl, kmuddy);

  //dialog is ready - show it!
  shdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete shdlg;
}

void cConnection::handleVariablesDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  //not a profile-based connection => no variables
  if (cp == NULL)
    return;

  cVariableList *vl = cp->variableList();
  //no variables class (???) => nothing to do
  if (vl == 0)
    return;

  //so first we have to create the dialog...
  dlgVariables *vdlg = new dlgVariables (vl, kmuddy);

  //dialog is ready - show it!
  vdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete vdlg;
}

void cConnection::handleVarTrigsDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  //not a profile-based connection => no variables
  if (cp == NULL)
    return;

  cVarTriggerList *vl = cp->vartrigList();
  //no variables class (???) => nothing to do
  if (vl == 0)
    return;

  //so first we have to create the dialog...
  dlgVarTrigList *vdlg = new dlgVarTrigList (vl, kmuddy);

  //dialog is ready - show it!
  vdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete vdlg;
}

void cConnection::handleWindowsDialog ()
{
  if (cp == NULL)
    return;

  cWindowList *wl = cp->windowList();

  if (wl == 0)
    return;
  
  dlgWindows *wdlg = new dlgWindows (wl, kmuddy);
  
  wdlg->exec ();
  
  delete wdlg;
}

void cConnection::switchAliases (bool val)
{
  //no connection preferences => nothing to do
  if (cp == NULL)
    return;
  cAliasList *al = cp->aliasList ();
  //no aliases class (???) => nothing to do
  if (al == NULL)
    return;
  al->setEnabled (val);
  sess->console()->systemMessage (val ? i18n ("Aliases are now enabled.") :
        i18n ("Aliases are now disabled."));
}

void cConnection::switchTriggers (bool val)
{
  //no connection preferences => nothing to do
  if (cp == NULL)
    return;
  cTriggerList *tl = cp->triggerList ();
  //no triggers class (???) => nothing to do
  if (tl == NULL)
    return;
  tl->setEnabled (val);
  sess->console()->systemMessage (val ? i18n ("Triggers are now enabled.") :
        i18n ("Triggers are now disabled."));
}

void cConnection::switchTimers (bool val)
{
  //no connection preferences => nothing to do
  if (cp == NULL)
    return;
  cTimerList *tl = cp->timerList ();
  //no timers class (???) => nothing to do
  if (tl == NULL)
    return;
  if (val)
    tl->enableTimers ();
  else
    tl->disableTimers ();
  sess->console()->systemMessage (val ? i18n ("Timers are now enabled.") :
        i18n ("Timers are now disabled."));
}

void cConnection::switchShortcuts (bool val)
{
  //no connection preferences => nothing to do
  if (cp == NULL)
    return;
  cShortcutList *sl = cp->shortcutList ();
  //no shortcuts class (???) => nothing to do
  if (sl == NULL)
    return;
  if (val)
    sl->setEnabled (true);
  else
    sl->setEnabled (false);
  sess->console()->systemMessage (val ? i18n ("Macro keys are now enabled.") :
        i18n ("Macro keys are now disabled."));
}

void cConnection::updateMenus ()
{
  if (sess->telnet()->isConnected())
  {
    kmuddy->unsetClosedConn ();
    kmuddy->enableConnectedOptions();
    if (cp != NULL)
      kmuddy->enableAdvancedOptions();
    else
      kmuddy->disableAdvancedOptions();

    if (cp != NULL)
    {
      cAliasList *al = cp->aliasList ();
      cTriggerList *tl = cp->triggerList ();
      cTimerList *ml = cp->timerList ();
      cShortcutList *sl = cp->shortcutList ();
      kmuddy->setMenuAliasesEnabled (al->enabled ());
      kmuddy->setMenuTriggersEnabled (tl->enabled ());
      kmuddy->setMenuTimersEnabled (ml->enabled ());
      kmuddy->setMenuShortcutsEnabled (sl->enabled ());
    }
  }
  else
  {
    kmuddy->disableConnectedOptions();
    if (connclosed)
      kmuddy->setClosedConn (canCT);
    else
      kmuddy->unsetClosedConn ();
  }
}

void cConnection::saverTimeout ()
{
  if (cp == 0) return;
  //auto-save
  cp->saveAll ();
}
