/***************************************************************************
                          cconsole.h  -  main displaying widget
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : So Jun 22 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@host.sk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
Current status of cConsole class (Jun 14, 2003):
 - requested functionality implemented
 - new features may still be added
 - code has been tested and it seems to work
 - added improved addText and cHistoryBuffer; contributed by Alex Bache
*/

#ifndef CCONSOLE_H
#define CCONSOLE_H

#include <qgridview.h>

#include "csessionbase.h"

#include <stdio.h>

class QPainter;
class QPixmap;
class QSplitter;

// Cyclic buffer class - buffers a specified number of items

//CODE DISABLED, because I keep getting all sorts of strange crashes here :(
/*

template <class T>
class cHistoryBuffer
{
 private:
  T   *buffer;
  int  start;
  int  numItems;
  int  bufSize;
 public:
  cHistoryBuffer(const int bufferMaxSize=0);
  ~cHistoryBuffer();
  T &operator[] (const int offset);
  int maxSize() const   { return bufSize; }
  int size() const      { return numItems; }
  void add(const T &item);
  void resize(const int bufferMaxSize);
  void flush();
}; // cHistoryBuffer<T>
*/

#define CL_BLACK 0
#define CL_RED 1
#define CL_GREEN 2
#define CL_YELLOW 3
#define CL_BLUE 4
#define CL_MAGENTA 5
#define CL_CYAN 6
#define CL_WHITE 7

#define CL_BRIGHT 8

//maximum cache size
#define MAXCACHE 50

/**
Main console - displays output of the MUD server...
Inherits QGridView - has one column and one row per line (each cell contains
one line of text - removes flicker)

Includes another cConsole object - used to display last few lines of output
while scrolling the main console...
  *@author Tomas Mecir
  */

class cConsole : public QGridView, public cSessionBase {
   Q_OBJECT
public: 
  /** auxiliary console is used to display last few lines when scrolling... */
  cConsole(bool auxiliary=false, QWidget *parent=0, const char *name=0);
  ~cConsole();

  /** get auxiliary console */
  cConsole *auxConsole ();

  /** set font */
  void setFont (QFont f);
  /** get currently used font */
  QFont font () { return myfont; };
  /** get color in color palette */
  QColor color (int index);
  /** change color palette */
  void setColor (QColor color, int index);
  void setDefaultBkColor (int index);
  /** change color of local echo (typed commands etc.) */
  void setEchoColor (int index);
  /** change color of system messages */
  void setSystemColor (int index);
  /** change color of script output */
  void setScriptOutputColor (int index);

  void setWordWrapping (bool val);
  void setIndentation (int val);
  
  void setEnableEcho (bool value);
  void setEnableMessages (bool value);

  void setEnableBlinking (bool value);
  
  void setPromptLabel (bool value);
      
  int numCols () { return charCount; };
  void setNumCols (int numcols) { charCount = numcols; };
  int cellWidth () { return charWidth; };
  void setCellWidth (int) { /* do nothing! */ };
  /** this function returns all the colors */
  void getAllColors (int *_echo, int *_system);

  /** return current number of rows */
  int curRows() { return currows; };
  /** return current number of columns */
  int curCols() { return curcols; };

  /** forces amission of dimensionsChanged signal; used by toolbar hiding
  functions, where this fails for unknown reasons */
  void forceEmitSize ();
  /** adds a new system message to the widget - calls addText and also
  does some extra things*/
  void systemMessage (const QString &text);
  /** decision message (Tools/Decide). The same as systemMessage, but
  it is shown even if system messages are off*/
  void decisionMessage (const QString &text);
  /** dump all history buffer to that file */
  void dumpBuffer (bool fromcurrent, FILE *file);
  static void setInitialHistorySize (int size) { initHistory = size; };

public slots:
  /** clear the widget */
  void clear ();
  /** adds line to the widget */
  void addLine (const QString &text, const QString &tc, const QString &bc);
  /** as addLine, but does not end line */
  void addText (const QString &text, const QString &tc, const QString &bc);
  /** adds a new command to the widget - calls addText and also
  does some extra things*/
  void addCommand (const QString &text);
  /** adds one line of script output to the console */
  void addScriptOutput (const QString &text);
  /** this is my decision helper similar to one found in zmud */
  void makeDecision ();
  /** adds selection to clipboard (adding to mouse selection buffer is done
      automatically) */
  void addSelectionToClipboard (bool toclipboard = true);

  /** shifting it around (SHIFT+keys) */
  void lineUp ();
  void lineDown ();
  void pageUp ();
  void pageDown ();

  /** resizing aconsole (CTRL+keys) */
  void aconUp ();
  void aconDown ();
  
  void updatePrompt(const QString &text);
signals:
  void dimensionsChanged (int cols, int rows);
  void commandDisplayed (const QString &text);
protected slots:
  void sliderChanged (int value);
  void blinkTimerTimeout ();
protected:
  /** called when a cell is to be painted */
  void paintCell (QPainter *p, int row, int);
  /** resize event */
  void resizeEvent (QResizeEvent *e);
  /** called when resizing and when changing font */
  void fixupOutput ();
  /** updates a range of cells - used in events below */
  void updateRowRange (int r1, int r2);
  /** mouse press event */
  void mousePressEvent (QMouseEvent *e);
  /** mouse release event */
  void mouseReleaseEvent (QMouseEvent *e);
  /** mouse move event */
  void mouseMoveEvent (QMouseEvent *e);
  /** mouse double-click event */
  void mouseDoubleClickEvent (QMouseEvent *e);
  /** disables scrolling aux.console using mouse wheel */
  void wheelEvent (QWheelEvent *e);
  /** add new text; called by addLine and addText */
  void addNewText (const QString &text, const QString &tc, const QString &bc,
      bool endTheLine);
  /** handles necessary updates when new line needs to be created */
  void handleBufferShift ();
  /** end current line */
  void endLine ();
  /** ensure that our current position is at beginning of a line */
  void forceBeginOfLine ();
  /** make the redrawing cache empty */
  void emptyCache ();
  /** delete one item from the cache */  
  void deleteCacheEntry (int which);

  /** are there blinking characters on that line? */
  bool hasBlink (int row);
    
  QPixmap *cacheBlinkOn[MAXCACHE], *cacheBlinkOff[MAXCACHE];
  int cacheEntry[MAXCACHE];
  /** ON = blinking text shown, OFF = blinking text hidden */
  bool blinkPhase;
  QTimer *blinkTimer;
  
  QColor mycolor[16];
  int echocolor;
  int systemcolor;
  int scriptoutputcolor;

  /** wordwrapping / indentation */
  bool wordWrapping;
  int indentValue;

  /** blinking ON/OFF */
  bool blinking;
  
  /** prompt label */
  bool promptlabel;
  
  /** show commands / messages ? */
  bool cmdecho, messages;
  /** currently used font */
  QFont myfont;
  /** descent of the currently used font, needed for drawing */
  int descent;
  /** width of one character */
  int charWidth;
  /** number of characters in one line */
  int charCount;

  /** number of added rows during current addNewText call */
  int pendingLineMove;

  /** current console size */
  int currows, curcols;
  
  /** our auxiliary console */
  cConsole *aconsole;

  /** are WE the auxiliary console? */
  bool aux;
  /** is the auxiliary console visible? */
  bool aconvisible;
  /** size of auxiliary console (in percents) */
  int aconsize;

  //selection
  int selrow1, selrow2, selrowpos1, selrowpos2;
  int previousr, previousc;
  bool selected, canselect;

  /** initial history size */
  static int initHistory;
  
  //THE DATA
  int usedrows;
  int currow;
  int currowpos;
  int historySize;

  //cHistoryBuffer is disabled for now due to crashes
  
  QChar **cell;    //characters, [row][col]
//  cHistoryBuffer<QChar *> cell;    //characters, [row][col]
  //TODO: attributes should have a separate array
  unsigned char **tcolor;  //color of letters, [row][col]
//  cHistoryBuffer<unsigned char *> tcolor;
  //higher bits in tcolor are attributes: 16 if underlined,
  //32 if italics, 64 if blinking, 128 if strikeout
  unsigned char **bcolor;  //color of background, [row][col]
//  cHistoryBuffer<unsigned char *> bcolor;
};

#endif
