//
// C++ Implementation: cpluginmanager
//
// Description: plug-in manager
//
//
// Author: Tomas Mecir <kmuddy@kmuddy.org>, (C) 2003
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "cpluginmanager.h"

#include <kglobal.h>
#include <klibloader.h>
#include <klibloader.h>
#include <kstddirs.h>

#include <list>
using namespace std;

#include "cconsole.h"
#include "csession.h"

cPluginManager::cPluginManager (QObject *parent, const char *name)
 : QObject(parent, name)
{
  findPlugins ();
}


cPluginManager::~cPluginManager ()
{
  //unload plug-ins
  unloadAll ();

  //free informtion structures
  map<QString, sPluginInfo *>::iterator it;
  for (it = pluginInfo.begin(); it != pluginInfo.end(); ++it)
    delete it->second;
  pluginInfo.clear();
}

void cPluginManager::findPlugins ()
{
  int pc = 0;
  QStringList plugins = KGlobal::dirs()->findAllResources ("module", "kmuddy/*.la");
  QStringList::iterator it;
  for (it = plugins.begin(); it != plugins.end(); ++it)
  {
    QString path = *it;
    QString pathNoLA = path.left (path.length() - 3);
    KLibrary *lib = KLibLoader::self()->library (pathNoLA);
    if (lib)
    {
      //check if all necessary symbols are there
      if (lib->hasSymbol ("pluginName") && lib->hasSymbol ("longName") && lib->hasSymbol ("desc")
          && lib->hasSymbol ("license") && lib->hasSymbol ("version")
          && lib->hasSymbol ("kmuddyPluginVersion") && lib->hasSymbol ("load"))
      {
        const char **pname = (const char **) lib->symbol ("pluginName");
        const char **lname = (const char **) lib->symbol ("longName");
        const char **desc = (const char **) lib->symbol ("desc");
        const char **author = (const char **) lib->symbol ("author");
        const char **license = (const char **) lib->symbol ("license");
        const char **version = (const char **) lib->symbol ("version");
        int *plugver = (int *) lib->symbol ("kmuddyPluginVersion");
        if (*plugver != KMUDDY_PLUGIN_VERSION)  //require compatible plug-in version
          qDebug ("Found incompatible plug-in " + QString (*pname) + ", ignoring...");
        else
        {
          if (pluginInfo.count (*pname))  //already have such plug-in!!!
            qDebug ("Found duplicate plug-in " + QString (*pname) + ", ignoring...");
          else
          {
            sPluginInfo *spi = new sPluginInfo;
            spi->name = *pname;
            spi->longname = *lname;
            spi->desc = *desc;
            spi->author = *author;
            spi->license = *license;
            spi->version = *version;
            spi->library = pathNoLA;
            qDebug ("Plugin " + spi->name + " (" + spi->longname + ") " + spi->version + " by " +
                spi->author + ".");
            //add the plug-in to our database
            pluginInfo[*pname] = spi;
            pc++;
          }
        }
      }
      lib->unload ();
    }
  }
  qDebug ("Found " + QString::number (pc) + " plugins.");
}

typedef cPlugin *(*loadFn) ();

bool cPluginManager::loadPlugin (const QString &name)
{
  if (isLoaded (name))    //plug-in is already loaded!
    return false;
  if (!(pluginInfo.count (name)))  //we know nothing of that plug-in
    return false;
  QString path = pluginInfo[name]->library;
  KLibrary *lib = KLibLoader::self()->library (path);
  if (!lib)
  {
    qDebug ("FAILED to load plug-in " + name + "!");
    return false;
  }
  loadFn fn = (loadFn) lib->symbol ("load");
  if (!fn)
  {
    lib->unload ();
    qDebug ("FAILED to load plug-in " + name + "!");
    return false;
  }
  cPlugin *plugin = (*fn) ();
  if (!plugin)
  {
    lib->unload ();
    qDebug ("FAILED to load plug-in " + name + "!");
    return false;
  }

  //OKay... Plug-in LOADED. Fill in necessary structures.
  libraries[name] = lib;
  loadedPlugins[name] = plugin;
  plugins.insert (pair<int, cPlugin *>(plugin->priority(), plugin));
  return true;
}

bool cPluginManager::unloadPlugin (const QString &name)
{
  if (!(isLoaded (name)))    //plug-in is not loaded!
    return false;
  //get necessary pointers
  cPlugin *p = loadedPlugins[name];
  KLibrary *lib = libraries[name];
  //remove plugin from normal lists
  loadedPlugins.erase (name);
  libraries.erase (name);
  //remove plugin from priority list
  for (it = plugins.begin(); it != plugins.end(); ++it)
    if (it->second == p)
    {
      plugins.erase (it);
      break;
    }
  //free plug-in
  delete p;
  //close the library
  lib->unload ();
  qDebug ("Unloaded plug-in " + name + ".");
  return true;
}

bool cPluginManager::isLoaded (const QString &name)
{
  return loadedPlugins.count (name);
}

void cPluginManager::setNoAutoLoadList (QStringList noAutoLoad)
{
  noload = noAutoLoad;
}

void cPluginManager::loadAll ()
{
  map<QString, sPluginInfo *>::iterator it;
  for (it = pluginInfo.begin(); it != pluginInfo.end(); ++it)
    if (!(noload.contains (it->first)))
      loadPlugin (it->first);
}

void cPluginManager::unloadUnwanted ()
{
  QStringList::iterator it;
  for (it = noload.begin(); it != noload.end(); ++it)
    if (isLoaded (*it))
      unloadPlugin (*it);
}

void cPluginManager::unloadAll ()
{
  //unloading a plug-in modifies the map, invalidating iterators...
  //I therefore have to create a list of loaded plug-in names and use that instead
  list<QString> loadedList;
  map<QString, cPlugin *>::iterator lpi;
  for (lpi = loadedPlugins.begin(); lpi != loadedPlugins.end(); ++lpi)
    loadedList.push_back (lpi->first);
  list<QString>::iterator it;
  for (it = loadedList.begin(); it != loadedList.end(); ++it)
    unloadPlugin (*it);
}

void cPluginManager::passRawData (cSession *sess, char *data)
{
  for (it = plugins.begin(); it != plugins.end(); ++it)
    it->second->rawData (sess, data);
}

void cPluginManager::passDecompressedData (cSession *sess, char *data)
{
  for (it = plugins.begin(); it != plugins.end(); ++it)
    it->second->decompressedData (sess, data);
}

void cPluginManager::preprocessInput (cSession *sess, QString &data)
{
  for (it = plugins.begin(); it != plugins.end(); ++it)
    it->second->preprocessInput (sess, data);
}

void cPluginManager::passInput (cSession *sess, int phase, QString &line, QString &fg,
    QString &bg, bool gagged)
{
  for (it = plugins.begin(); it != plugins.end(); ++it)
    it->second->processInput (sess, phase, line, fg, bg, gagged);
}

void cPluginManager::passPrompt (cSession *sess, QString &prompt, QString &fg, QString &bg)
{
  for (it = plugins.begin(); it != plugins.end(); ++it)
    it->second->processPrompt (sess, prompt, fg, bg);
}

void cPluginManager::passCommand (cSession *sess, QString &command, bool &dontSend)
{
  dontSend = false;
  for (it = plugins.begin(); it != plugins.end(); ++it)
  {
    it->second->processCommand (sess, command, dontSend);
    if (dontSend) break; //stop processing if plug-in doesn't want the command to be sent
  }
}

bool cPluginManager::passMacro (cSession *sess, const QString &macro, const QString &params)
{
  for (it = plugins.begin(); it != plugins.end(); ++it)
    if (it->second->processMacro (sess, macro, params))
      return true;
  //nobody has processed it
  return false;
}

void cPluginManager::event (cSession *sess, const QString &type, const QString &name,
    const QString &par1, const QString &par2)
{
  for (it = plugins.begin(); it != plugins.end(); ++it)
    it->second->event (sess, type, name, par1, par2);
}

void cPluginManager::systemMessage (cSession *sess, const QString &message)
{
  if (sess)
    sess->console()->systemMessage (message);
}

void cPluginManager::displayLine (cSession *sess, const QString &line, const QString &fg,
    const QString &bg)
{
  if (sess)
    sess->console()->addLine (line, fg, bg);
}

#include "cpluginmanager.moc"
