//
// C++ Interface: cpluginmanager
//
// Description: plug-in manager
//
//
// Author: Tomas Mecir <kmuddy@kmuddy.org>, (C) 2003
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef CPLUGINMANAGER_H
#define CPLUGINMANAGER_H

#include <qobject.h>
#include <qstringlist.h>
#include <map>

#include "cplugin.h"

using namespace std;

class cSession;

class KLibrary;

struct sPluginInfo {
  //name is internal name used when referring to the plug-in
  //longname is what the user will see
  //desc describes plug-in's functionality
  //author and license is obvious :)
  QString name, longname, desc, author, license, version;
  //file containing this plug-in (excluding the .la/.so suffix)
  QString library;
};

/**
Plug-in manager handles plug-in information, loads plug-ins, and passes data/events to them as needed.

@author Tomas Mecir
*/

class cPluginManager : public QObject
{
  Q_OBJECT
public:
  cPluginManager (QObject *parent = 0, const char *name = 0);
  ~cPluginManager ();

  const map<QString, sPluginInfo *> *getPluginInfo () { return &pluginInfo; };

  /** load a plug-in into memory, if possible */
  bool loadPlugin (const QString &name);
  /** unload a plug-in from memory, if possible */
  bool unloadPlugin (const QString &name);
  bool isLoaded (const QString &name);

  /** list of plug-ins that should not be loaded - from preferences */
  void setNoAutoLoadList (QStringList noAutoLoad);

  /** Load all available plug-ins expect those that should not be loaded.
  This ensures that all newly installed plug-ins get loaded initially. */
  void loadAll ();
  /** unload all plug-ins that are currently loaded, but are included in noload list */
  void unloadUnwanted ();
  /** unload all plug-ins */
  void unloadAll ();

public slots:
  void passRawData (cSession *sess, char *data);
  void passDecompressedData (cSession *sess, char *data);
  void preprocessInput (cSession *sess, QString &data);
  void passInput (cSession *sess, int phase, QString &line, QString &fg, QString &bg, bool gagged);
  void passPrompt (cSession *sess, QString &prompt, QString &fg, QString &bg);
  void passCommand (cSession *sess, QString &command, bool &dontSend);
  bool passMacro (cSession *sess, const QString &macro, const QString &params);
  void event (cSession *sess, const QString &type, const QString &name = QString::null,
      const QString &par1 = QString::null, const QString &par2 = QString::null);
protected slots:
  void systemMessage (cSession *sess, const QString &message);
  void displayLine (cSession *sess, const QString &line, const QString &fg, const QString &bg);
protected:
  void findPlugins ();
  map<QString, sPluginInfo *> pluginInfo;
  /** loaded libraries */
  map<QString, KLibrary *> libraries;
  /** loaded plug-ins */
  map<QString, cPlugin *> loadedPlugins;
  /** plug-ins sorted by priority */
  multimap<int, cPlugin *> plugins;
  /** iterator, it's not a local variable, so that passing data to plug-ins is a bit faster :) */
  multimap<int, cPlugin *>::iterator it;
  QStringList noload;
};

/** plug-in manager class */
extern cPluginManager *pluginmanager;

#endif
