/***************************************************************************
                          cprofiles.cpp  -  list of profiles
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : Ct Jul 25 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@host.sk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cprofiles.h"

#include "cconnprefs.h"

#include <karchive.h>
#include <kglobal.h>
#include <klocale.h>
#include <kprocess.h>
#include <kstandarddirs.h>
#include <ktar.h>
#include <qdir.h>
#include <qfile.h>

cProfiles::cProfiles()
{
  load ();
}

cProfiles::~cProfiles()
{
  //nothing (yet?)
}

void cProfiles::load ()
//loads data; save is not needed, as modifications are written immediately;
//this prevents some problems with multiple instances...
{
  QString dirname = locateLocal ("appdata", "profiles/");
  QDir *dir = new QDir (dirname);
  dir->setFilter (QDir::Dirs);
  profiles = dir->entryList ();
  profiles.remove (".");
  profiles.remove ("..");
  delete dir;
}

/** returns a list of profiles */
const QStringList & cProfiles::getprofiles ()
{
  return profiles;
}

/** adds a new profile */
bool cProfiles::addProfile (QString & name, cConnectionInfo *cci)
{
  //add = modify; cConnPrefs' constructor can handle this
  if (exists (name))
    return false;
  return modifyProfile (name, cci);
}

bool cProfiles::deleteProfile (QString & name)
//stuff like asking Really? must be handled by caller...
{
  //locate profile directory
  QString dir = locateLocal ("appdata", "profiles/" + name + "/");
  
  //DELETE IT NOW!!!
  //this could also be handled manually by deleting all the files in the directory,
  //then the directory itself, but why reinvent the wheel, when rm can do this
  //better (especially if there will be some sub-dirs; what's quite probable...)?
  KProcess *kp = new KProcess ();
  *kp << "rm" << "-rf" << dir;
  bool res = kp->start (KProcess::Block);
  
  delete kp;  //process has already finished...
  
  //update profile list
  load ();

  return res;
}

bool cProfiles::modifyProfile (QString & name, cConnectionInfo *cci)
{
  if (name.isNull() || (name.simplifyWhiteSpace() == ""))
    return false;
  cConnPrefs *cp = new cConnPrefs (name, NULL, false);

  cp->setName (cci->name());
  cp->setServer (cci->server());
  cp->setLogin (cci->login());
  cp->setPassword (cci->password());
  cp->setPort (cci->port());
  cp->setConnStr(cci->connStr());
  
  delete cp;  //this will save new information...

  //update profile list
  load ();

  return true;  //everything is okay
}

bool cProfiles::duplicateProfile (const QString &name, const QString &newname)
{
  QString oldname = name.simplifyWhiteSpace ();
  QString thenewname = newname.simplifyWhiteSpace ();
  
  //won't overwrite existing profile
  if (exists (thenewname))
    return false;
  //won't duplicate non-existing profile
  if (!exists (oldname))
    return false;
  //won't allow empty name
  if (oldname.isNull() || (oldname == ""))
    return false;
    
  QString dir = locateLocal ("appdata", "profiles/" + oldname + "/");
  QString newdir = locateLocal ("appdata", "profiles/" + thenewname + "/");

  //protect special characters in directory names
  dir = KShellProcess::quote (dir);
  newdir = KShellProcess::quote (newdir);

  //we have to use shell process, because we need to expand that '*' char
  KShellProcess *kp = new KShellProcess ();
  *kp << "cp" << "-R" << (dir + "/*") << newdir;
  bool res = kp->start (KProcess::Block);
  return res;
}

bool cProfiles::exists (QString & name)
{
  return (profiles.contains (name) == 0) ? false : true;
}

/** this CREATES A NEW OBJECT OF TYPE cConnectionInfo and returns a pointer
  it. CALLER MUST DESTROY THIS OBJECT when not needed. */
cConnectionInfo * cProfiles::getProfileConnInfo (QString & name)
{
  cConnPrefs *cp = new cConnPrefs (name, NULL, false);

  //this is the object that will be returned
  cConnectionInfo *ci = new cConnectionInfo;

  ci->setName (cp->name());
  ci->setServer (cp->server());
  ci->setLogin (cp->login());
  ci->setPassword (cp->password());
  ci->setPort (cp->port());
  ci->setConnStr(cp->connStr());
  
  delete cp;

  return ci;
}

bool cProfiles::importProfile (const QString &name,
    const QString &profileFile)
{
  //look if it exists
  QFile file (profileFile);
  if (!(file.exists()))
  {
    error = i18n ("Source file does not exist!");
    return false;
  }
  
  //bail out if target profile already exists
  if (exists ((QString&)name))
  {
    error = i18n ("Target profile already exists!");
    return false;
  }
  
  //create new profile directory
  QString directory = locateLocal ("appdata", "profiles/");
  QDir dir (directory);
  dir.mkdir (name, false);

  //extract the given file into that directory
  KTar archive (profileFile);
  archive.open (IO_ReadOnly);
  if (!archive.isOpened ())
  {
    error = i18n ("Could not open source file!");
    return false;
  }
  const KArchiveDirectory *ad = archive.directory ();
  QStringList entries = ad->entries ();
  QStringList::iterator it;
  for (it = entries.begin(); it != entries.end(); ++it)
  {
    const KArchiveEntry *entry = ad->entry (*it);
    if (entry)
    {
      if (entry->isDirectory ())    //no directories in the archive!
      {
        error = i18n ("This is not a valid exported profile!");
        load ();
        deleteProfile ((QString&)name); //delete all created files
        return false;       
      }
      //look if name is recognized
      if ((*it == "preferences") || (*it == "aliases") ||
          (*it == "triggers") || (*it == "actions") ||
          (*it == "aliasgroups") || (*it == "triggergroups") ||
          (*it == "scripts") || (*it == "timers") ||
          (*it == "macrokeys") || (*it == "vartriggers"))
      {
        //this code is copied from kdelibs sources, file karchive.cpp,
        //function KArchiveFile::copyTo (with some modifications)
        //I don't use that function directly, because it didn't exist in
        //KDE 3.0, and I want to stay compatible
        QFile f(directory + "/" + name + "/" + *it);
        f.open (IO_ReadWrite | IO_Truncate);
        f.writeBlock (((KArchiveFile *) entry)->data ());
        f.close ();
      }
      else
      {
        error = i18n ("This is not a valid exported profile!");
        load ();
        deleteProfile ((QString &) name); //delete all created files
        return false;
      }
    }
  }
  
  //good, everything's done, clean up
  archive.close ();
  return true;
}

bool cProfiles::exportProfile (const QString &name,
    const QString &profileFile)
//!!! This may not work properly if the profile is currently active  !!!
//!!! In that case, recent updates (not saved yet) won't be exported !!!
{
  //look if profile exists
  if (!exists ((QString &) name))
  {
    //this should NEVER happen, but just in case...
    error = i18n ("That profile does not exist!");
    return false;
  }
  
  //remove target archive if it exists
  QFile::remove (profileFile);

  //put all files into target file (it is a .tar file)
  KTar archive (profileFile);
  archive.open (IO_WriteOnly | IO_Truncate);
  if (!archive.isOpened ())
  {
    error = i18n ("Could not open target file!");
    return false;
  }
  QString directory = locateLocal ("appdata", "profiles/" + name + "/");
  QDir dir (directory);
  const QFileInfoList *entryList = dir.entryInfoList ();
  QFileInfoListIterator it (*entryList);
  QFileInfo *finfo;
  while ((finfo = it.current()) != 0)
  {
    //only files - this is to prevent inclusion of '.' and '..'
    if (finfo->isFile())
    {
      //write them there, file after file...
      QString fName = directory + finfo->fileName();
      QFile file (fName);
      if (!(file.open (IO_ReadOnly)))
      {
        //shouldn't happen, but...
        error = i18n ("I'm having problem reading the profile.");
        archive.close ();
        QFile::remove (profileFile);
        return false;
      }
      unsigned int size = file.size ();
      char *data = new char[size + 1];
      file.readBlock (data, size);
      archive.writeFile (finfo->fileName(), finfo->owner(), finfo->group(),
          size, data);
      delete data;
      file.close ();
    }
    
    //move to the next file
    ++it;
  }
  
  //good, everything's done, clean up
  archive.close ();
  return true;
}
