/***************************************************************************
                          crunninglist.cpp  -  list of running scripts
                             -------------------
    begin                : Ne dec 22 2002
    copyright            : (C) 2002-2003 by Tomas Mecir
    email                : kmuddy@kmuddy.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "crunninglist.h"

#include "ccmdprocessor.h"
#include "cconsole.h"
#include "crunningscript.h"
#include "csession.h"
#include "csendcmd.h"
#include "ctextprocessor.h"

#include <klocale.h>

#include <stdlib.h>

cRunningList::cRunningList (cSession *_sess)
{
  sess = _sess;
  scripts.clear ();
  siter = scripts.begin();
  lastid = 0;
  waitCounter = 0;
  waitLock    = false;
  textQueue.clear ();
  typeQueue.clear ();
  preprocessorScript = 0;
  //sending text to scripts...
  connect (sess->sendcmd(), SIGNAL (commandSent (const QString &)), this, SLOT (sendCommand (const QString &)));
  connect (sess->textproc(), SIGNAL (textLine (const QString &)), this, SLOT (sendServerOutput (const QString &)));
  connect (sess->textproc(), SIGNAL (receivedPrompt (const QString &)), this, SLOT (sendPrompt (const QString &)));
}

cRunningList::~cRunningList()
{
  killAll ();
}

void cRunningList::addScript (cRunningScript *script)
{
  //connect the script!
  connect (script, SIGNAL (sendText (const QString &)), sess->cmdproc(),
        SLOT (processInput (const QString &)));
  connect (script, SIGNAL (displayText (const QString &)), sess->console(),
        SLOT (addScriptOutput (const QString &)));
  connect (script, SIGNAL (scriptFinished (cRunningScript *, int)), this,
        SLOT (scriptFinished (cRunningScript *, int)));
  connect (script, SIGNAL (scriptKilled (cRunningScript *)), this,
        SLOT (scriptKilled (cRunningScript *)));
  connect (script, SIGNAL (scriptFailed (cRunningScript *)), this,
        SLOT (scriptFailed (cRunningScript *)));
  connect (script, SIGNAL (textSent ()), this, SLOT (scriptTextSent ()));
  connect (script, SIGNAL (textAccepted ()), this, SLOT (scriptTextAccepted ()));
        
  //set script ID
  //FIXME: things may break down if we run out of free IDs!
  //  (id is an 32-bit unsigned int)
  script->setId (++lastid);

  //and add it to my list
  scripts.push_back (script);

  //inform about the change
  emit stateChanged ();
}

bool cRunningList::addPreprocessorScript (cRunningScript *script)
{
  if (preprocessorScript)
    return false;

  //connect the script!
  connect (script, SIGNAL (sendText (const QString &)), this,
        SLOT (preprocessorOutput (const QString &)));
  connect (script, SIGNAL (scriptFinished (cRunningScript *, int)), this,
        SLOT (scriptFinished (cRunningScript *, int)));
  connect (script, SIGNAL (scriptKilled (cRunningScript *)), this,
        SLOT (scriptKilled (cRunningScript *)));
  connect (script, SIGNAL (scriptFailed (cRunningScript *)), this,
        SLOT (scriptFailed (cRunningScript *)));

  //set script ID
  //FIXME: things may break down if we run out of free IDs!
  //  (id is an 32-bit unsigned int)
  script->setId (++lastid);

  //and add it to my list
  scripts.push_back (script);

  //we now have a preprocessor script
  preprocessorScript = script;
  preprocessorBuffer = QString::null;
  
  //inform about the change
  emit stateChanged ();

  return true;
}

bool cRunningList::havePreprocessorScript ()
{
  return (preprocessorScript == 0) ? false : true;
}

void cRunningList::sendToPreprocessor (const QString &text)
{
  if (preprocessorScript)
    preprocessorScript->sendCommandToScript (text, SERVEROUTPUT);
}

void cRunningList::preprocessorOutput (const QString &text)
{
  sess->textproc()->preprocessorOutput (text);
}

void cRunningList::suspendAll ()
{
  list<cRunningScript *>::iterator it;
  for (it = scripts.begin(); it != scripts.end(); ++it)
    (*it)->suspend ();
}

void cRunningList::resumeAll ()
{
  list<cRunningScript *>::iterator it;
  for (it = scripts.begin(); it != scripts.end(); ++it)
    (*it)->resume ();
}

void cRunningList::killAll ()
{
  list<cRunningScript *>::iterator it;
  it = scripts.begin();
  while (it != scripts.end())
  {
    cRunningScript *script = (*it);
    it = scripts.erase (it);
    delete script;
  }
}

unsigned int cRunningList::getFirstId ()
{
  siter = scripts.begin();
  if (siter == scripts.end())  //empty list
    return 0;
  cRunningScript *script = (*siter);
  return script->getId();
}

unsigned int cRunningList::getNextId ()
{
  //advance the iterator, return script ID (if any)
  ++siter;
  if (siter == scripts.end())
    return 0;  
  cRunningScript *script = (*siter);
  return script->getId();
}

const QString cRunningList::name (int id)
{
  cRunningScript *rs = getRunningScript (id);
  return rs->name();
}

void cRunningList::suspend (int id)
{
  cRunningScript *rs = getRunningScript (id);
  rs->suspend();
  //inform about the change
  emit stateChanged ();
}

void cRunningList::resume (int id)
{
  cRunningScript *rs = getRunningScript (id);
  rs->resume();
  //inform about the change
  emit stateChanged ();
}

bool cRunningList::isSuspended (int id)
{
  cRunningScript *rs = getRunningScript (id);
  return rs->isSuspended();
  //inform about the change
  emit stateChanged ();
}

bool cRunningList::isRunning (int id)
{
  cRunningScript *rs = getRunningScript (id);
  return rs->isRunning();
  //inform about the change
  emit stateChanged ();
}

void cRunningList::terminate (int id)
{
  cRunningScript *rs = getRunningScript (id);
  rs->terminate();
  //inform about the change
  emit stateChanged ();
}

void cRunningList::kill (int id)
{
  cRunningScript *rs = getRunningScript (id);
  rs->kill();
  //inform about the change
  emit stateChanged ();
}

const QString cRunningList::getStatus (int id)
{
  cRunningScript *rs = getRunningScript (id);
  //currently only these two states; maybe I'll add more later?
  if (rs->isSuspended())
    return i18n ("Suspended");
  else
    return i18n ("Running");
}

bool cRunningList::requestLock (cRunningScript *script, const QString &varname)
{
  if (locks.count (varname))
  {
    //look who has the lock
    cRunningScript *rs = locks[varname];
    //report success if the caller has the lock :)
    if (script == rs)
      return true;
    //report failure otherwise
    return false;
  }
  else
  {
    //success
    locks[varname] = script;
    return true;
  }
}

void cRunningList::releaseLock (cRunningScript *script, const QString &varname)
{
  if (locks.count (varname))
    if (locks[varname] == script)
      locks.erase (varname);
}

bool cRunningList::canModify (cRunningScript *script, const QString &varname)
{
  if (locks.count (varname))
    return (locks[varname] == script);
  return true;
}

cRunningScript *cRunningList::getRunningScript (int id)
{
  //first, this is usually called during getFirstID/getNextID, so we optimize
  //things a bit by looking if last returned ID isn't what he wants
  if (siter != scripts.end())
    //FIXME: what happens if that iterator has been invalidated?
    if ((*siter)->getId() == id)
      return *siter;
  //this is O(n); maybe I'll change list to map, achieving O(log n)?
  list<cRunningScript *>::iterator it;
  for (it = scripts.begin(); it != scripts.end(); ++it)
    if ((*it)->getId() == id)  //this is our script!
      return *it;
  //script not found - return 0
  return 0;
}

void cRunningList::removeScript (cRunningScript *script)
{
  //release all locks held by this script
  //we have to do this in two stages, because deleting from a map
  //immediately would invalidate our iterator (yes we could have two
  //iterators, but why bother...)
   //stage 1: obtain a list of all such locks
  list<QString> ourLocks;            
  map<QString, cRunningScript *>::iterator it;
  for (it = locks.begin(); it != locks.end(); ++it)
    if (it->second == script)
      ourLocks.push_back (it->first);
   //stage 2: release all the locks
  list<QString>::iterator it2;
  for (it2 = ourLocks.begin(); it2 != ourLocks.end(); ++it2)
    locks.erase (*it2);
  
  //deleting preprocessor script...
  if (script == preprocessorScript)
    preprocessorScript = 0;
  
  //delete the script
  list<cRunningScript *>::iterator it3;
  for (it3 = scripts.begin(); it3 != scripts.end(); ++it3)
    if ((*it3) == script)  //this is our script!
    {
      scripts.erase (it3);
      delete script;
      return;
    }
  //script not found - VERY strange...
  //now there's a problem - to delete or not to delete? We delete, because
  //reaching this means a bug somewhere, which will be caught sooner if we
  //crash here...
  delete script;
}

void cRunningList::scriptFinished (cRunningScript *script, int returnValue)
{
  sess->console()->systemMessage ("Script " + script->name() +
        " has finished with return code " + QString::number (returnValue) + ".");
  removeScript (script);
  
  emit stateChanged ();
}

void cRunningList::scriptKilled (cRunningScript *script)
{
  sess->console()->systemMessage ("Script " + script->name() + " has been terminated!");
  removeScript (script);

  emit stateChanged ();
}

void cRunningList::scriptFailed (cRunningScript *script)
{
  sess->console()->systemMessage ("Script " + script->name() + " could not be started!");

  removeScript (script);

  emit stateChanged ();

}

void cRunningList::sendThisNow (const QString &text, int type, bool noFC,
    bool toPreprocessor)
{
  list<cRunningScript *>::iterator it;

  if (noFC == true)
  {
     for (it = scripts.begin(); it != scripts.end(); ++it)
       //only scripts with the same no-flow-control setting
       if (((*it)->noFlowControl() == noFC) &&
           (toPreprocessor == (*it)->isPreprocessor ()))
         (*it)->sendCommandToScript (text, type);
  } // endif non flow-controlled script
  else
  {
     // Assuming preprocessor script cannot be flow-controlled

     waitLock = true;  // Lock scriptTextSent() from sending data to scripts
     
     sendToFlowControlled(text,type);

     QString q_text;
     int     q_type;

     while ((waitCounter  == 0) &&
            (!textQueue.empty())  )
     {
       q_text = textQueue.front ();
       q_type = typeQueue.front ();
       textQueue.pop_front ();
       typeQueue.pop_front ();
       sendToFlowControlled(q_text, q_type);

     } // endwhile last script in list didn't send command and things in textQueue

     waitLock = false; // Allow scriptTextSent() to send data to scripts
  } // endelse flow-controlled script
  
} // cRunningList::sendThisNow



void cRunningList::sendToFlowControlled(const QString &text, int type)
{
  list<cRunningScript *>::iterator it;

  waitCounter = 0;

  // Count number of flow-controlled scripts
  for (it = scripts.begin(); it != scripts.end(); ++it)
  {
    if ((*it)->noFlowControl() == false)
    {
      waitCounter++;
    } // endif flow-controlled script

  } // endfor more flow-controlled scripts to count


  // Send data to flow-controlled scripts
  for (it = scripts.begin(); it != scripts.end(); ++it)
  {
    if ((*it)->noFlowControl() == false)
    {
      (*it)->sendCommandToScript (text, type);

      if (!(*it)->actuallySentCommand())
      {
        waitCounter--;
      } // endif no command actually sent to script

    } // endif flow-controlled script

  } // endfor more flow-controlled scripts to count

} // cRunningList::sendToFlowControlled


void cRunningList::sendCommand (const QString &text)
{
  //manage no-flow-control scripts first
  sendThisNow (text + "\n", USERCOMMAND, true);

  //if we're not waiting for anything
  if ((waitCounter == 0) && (textQueue.empty()))
    sendThisNow (text + "\n", USERCOMMAND);
  else
  {
    textQueue.push_back (text + "\n");
    typeQueue.push_back (USERCOMMAND);
  }
}

void cRunningList::sendPrompt (const QString &text)
{
  //manage no-flow-control scripts first
  sendThisNow (text + "\n", PROMPT, true);
  
  //if we're not waiting for anything
  if ((waitCounter == 0) && (textQueue.empty()))
    sendThisNow (text + "\n", PROMPT);
  else
  {
    textQueue.push_back (text + "\n");
    typeQueue.push_back (PROMPT);
  }
}

void cRunningList::sendServerOutput (const QString &text)
{
  //manage no-flow-control scripts first
  sendThisNow (text + "\n", SERVEROUTPUT, true);

  //if we're not waiting for anything
  if ((waitCounter == 0) && (textQueue.empty()))
    sendThisNow (text + "\n", SERVEROUTPUT);
  else
  {
    textQueue.push_back (text + "\n");
    typeQueue.push_back (SERVEROUTPUT);
  }
}

void cRunningList::scriptTextAccepted ()
{
//  waitCounter++;
}

void cRunningList::scriptTextSent ()
{
  waitCounter--;

  if (waitLock == true)
  {
    return;
  } // endif sending from textQueue locked off
  else if ((waitCounter == 0) && (!(textQueue.empty())))
  {
    //we're no longer waiting - send next line, if any
    
    QString text = textQueue.front ();
    int     type = typeQueue.front ();
    textQueue.pop_front ();
    typeQueue.pop_front ();
    sendThisNow (text, type);
    
  } // endelse allowed to send from textQueue
  
} // scriptTextSent

