/***************************************************************************
                          csaveablefield.h  -  alias/trigger baseclass
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : So sep 7 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@kmuddy.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CSAVEABLEFIELD_H
#define CSAVEABLEFIELD_H

/*
Current status of cSaveableField class (Dec 05, 2002):
 - requested features implemented
 - new features may still be added
 - abstract class
 - code has been tested and it seems to work
*/

#include <qobject.h>
#include <qregexp.h>
#include <qstring.h>
#include <qstringlist.h>

class KConfig;

class cGroupList;
class cGroup;

/**
Abstract base-class for alias / trigger objects (and more).
Knows how to save itself.

This class is arranged in a double-linked list.

TODO: lots of stuff declared here is not used in most derived classes (only in cAlias/cTrigger).
Therefore, this class needs to be simplified a bit.

  *@author Tomas Mecir
  */

//object types - NEED TO BE CHANGED WHEN REDESIGNING FOR 2.0
#define TYPE_ALIAS 1
#define TYPE_TRIGGER 2
#define TYPE_ACTION 3
#define TYPE_GROUP 4
#define TYPE_SCRIPT 5
#define TYPE_TIMER 6
#define TYPE_SHORTCUT 7
#define TYPE_VARIABLE 8
#define TYPE_VARTRIG 9

class cSaveableField : public QObject {
  Q_OBJECT
public:
  friend class cSaveableList;

  /** constructor */
  cSaveableField ();
  /** destructor */
  virtual ~cSaveableField ();

  /** creates a new instance of the class; this is needed because I need to
  create instances of childclasses from within this class, but I don't know
  the exact type of that instance... */
  virtual cSaveableField *newInstance () = 0;

  /** does the text match this alias/trigger? */
  virtual bool match (const QString &what);
  /** Abstract; returns a command that is to be executed (expanded alias,
triggered command, macro call, ...). To be called after match() returns true.*/
  virtual QString performAction () = 0;
  /** Expands pseudo-variables in the string. Pseudo-variables are the following:
<br><b>$prefix</b> - part of string BEFORE the matched part
<br><b>$suffix</b> - part of string AFTER the matched part
<br><b>$prefixtrim</b> - prefix without leading/trailing spaces
<br><b>$suffixtrim</b> - suffix without leading/trailing spaces
<br><b>$matched</b> - matched part of the string
<br><b>$1</b>, <b>$2</b>, ... - back-references; regexp expansion only

Real variables will have similar syntax (when they will be implemented).*/
  void expandPseudoVariables (QString &string);
  /** If match is a result of match(), should we continue searching for
appropriate alias/trigger handler? */
  virtual bool shouldContinue (bool match) = 0;

  /** abstract; load data from a config file*/
  virtual void load (KConfig *config, const QString &group) = 0;
  /** abstract; save data to a config file*/
  virtual void save (KConfig *config, const QString &group) = 0;

  /** abstract; returns type of item (light-weight RTTI) */
  virtual int itemType () = 0;

  /** mark/unmark this object - used to prevent infinite recursion in alias
expansion; may have other uses as well */
  void setMarked (bool mark) { _marked = mark; };
  /** is this object marked? */
  bool marked () { return _marked; };

  /** set alias/trigger text */
  void setText (const QString &_text);
  /** returns current alias/trigger text */
  QString getText () { return text; };

  /** set alias/trigger replacement(s)/command(s) */
  void setNewText (const QStringList &_text) { newtext = _text; } ;
  /** returns current alias/trigger replacement/command */
  QStringList getNewText () { return newtext; };

  const QStringList &commandsToExec () { return commandstoexec; };

  void setCaseSensitive (bool cs);
  bool caseSensitive () { return _cs; };

  /** set comparison type */
  void setType (int t);
  /** returns comparison type */
  int getType () { return type; };
  /** possible types of comparison */
  enum compType { exact, substring, begin, end, regexp };
  #define TYPES 5

  /** returns last text being successfully compared */
  QString getLastText () { return lasttext; };
  /** returns last prefix */
  QString getPrefix () { return prefix; };
  /** returns last suffix */
  QString getSuffix () { return suffix; };
  int getLastPos () { return lastlen; };
  int getLastLength () { return lastpos; };
  /** returns a list of back-references (if doing regexp compare) */
  QStringList getBackRefList () { return backreflist; };

  /** Expands variable; used by expandVariables() and in test areas */
  QString processVariable (const QString &varname);

  /** returns group to which we belong - only used with aliases/triggers */
  int group () { return _group; };
  /** sets group to which we belong - only used with aliases/triggers */
  void setGroup (int val) { _group = val; };
  void updateGroupPointer (cGroupList *list);
  /** sets a no-group mode - used in testarea, which must work even if all
  the groups are disabled */
  void setNoGroupMode () { nogroup = true; };
protected:
  /** fields form a double-linked list. These are some other entries... */
  cSaveableField *prev, *next;
  /** is this field marked? */
  bool _marked;

  /** alias/trigger text */
  QString text;
  /** alias replacement(s) or triggered command(s) */
  QStringList newtext;

  /** commands that are to be executed NOW */
  QStringList commandstoexec;

  /** case sensitive comparison */
  bool _cs;

  /** compare type */
  int type;

  /** last text that has been successfully compared */
  QString lasttext;
  /** prefix/suffix parts of that text (before/after matched part) */
  QString prefix, suffix;
  /** position/length of matching part */
  int lastpos, lastlen;
  /** list of back-references (if doing regexp compare) */
  QStringList backreflist;
  /** positions of backreferences */
  int *backrefpos;
  
  /** creating regexp objects on each trigger matching is SLOW - pre-parse
  them */
  QRegExp t_regexp;
  
  /** group ID to which we belong - only used with aliases/triggers */
  int _group;
  /** pointer to our group */
  cGroup *ourgroup;
  bool nogroup;
};

#endif
