/***************************************************************************
                          cscript.cpp  -  one script
                             -------------------
    begin                : So dec 7 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@kmuddy.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cscript.h"

#include "cconnection.h"
#include "cconnprefs.h"
#include "cconsole.h"
#include "crunninglist.h"
#include "crunningscript.h"
#include "csession.h"

#include <qdir.h>
#include <kconfig.h>
#include <klocale.h>
#include <kprocess.h>

cScript::cScript (cSession *s)
{
  sess = s;
  runningCount = 0;

  //set some defaults...
  name = ""; //name must be set before script can be executed...
  comment = "";
  command = "";
  workdir = QDir::homeDirPath ();
  prefix = "";
  suffix = "";
  enableinput = true;
  enablestdout = true;
  sendstdout = true;
  enablestderr = false;
  sendstderr = false;
  sendusercommands = false;
  useadvcomm = false;
  allowparams = true;
  singleinstance = false;
  shellexpansion = true;
  noflowcontrol = false;
  allowvars = false;
  preprocessor = false;
  text = "";
  type = substring;
  onlyifmatch = true;
}

cScript::~cScript()
{
  //nothing here
}

cSaveableField *cScript::newInstance ()
{
  return new cScript (sess);
}

void cScript::load (KConfig *config, const QString &group)
{
  config->setGroup (group);
  name = config->readEntry ("Name", "");
  comment = config->readEntry ("Comment", "");
  command = config->readEntry ("Command", "");
  workdir = config->readEntry ("Workdir", QDir::homeDirPath ());
  prefix = config->readEntry ("Prefix", "");
  suffix = config->readEntry ("Suffix", "");
  enableinput = config->readBoolEntry ("Enable input", true);
  enablestdout = config->readBoolEntry ("Enable output", true);
  sendstdout = config->readBoolEntry ("Send output", true);
  enablestderr = config->readBoolEntry ("Include stderr", false);
  sendstderr = config->readBoolEntry ("Send stderr", false);
  sendusercommands = config->readBoolEntry ("Send user commands", false);
  useadvcomm = config->readBoolEntry ("Use advanced communication", false);
  allowparams = config->readBoolEntry ("Allow parameters", true);
  singleinstance = config->readBoolEntry ("Single instance", false);
  shellexpansion = config->readBoolEntry ("Shell expansion", true);
  noflowcontrol = config->readBoolEntry ("No flow control", false);
  allowvars = config->readBoolEntry ("Communicate variables", false);
  preprocessor = config->readBoolEntry ("Preprocessor script", false);

  text = config->readEntry ("Text", "");
  type = config->readNumEntry ("Type", substring);
  onlyifmatch = config->readBoolEntry ("Only if match", true);
}

void cScript::save (KConfig *config, const QString &group)
{
  config->setGroup (group);
  config->writeEntry ("Name", name);
  config->writeEntry ("Comment", comment);
  config->writeEntry ("Command", command);
  config->writeEntry ("Workdir", workdir);
  config->writeEntry ("Prefix", prefix);
  config->writeEntry ("Suffix", suffix);
  config->writeEntry ("Enable input", enableinput);
  config->writeEntry ("Enable output", enablestdout );
  config->writeEntry ("Send output", sendstdout);
  config->writeEntry ("Include stderr", enablestderr);
  config->writeEntry ("Send stderr", sendstderr);
  config->writeEntry ("Send user commands", sendusercommands);
  config->writeEntry ("Use advanced communication", useadvcomm);
  config->writeEntry ("Allow parameters", allowparams);
  config->writeEntry ("Single instance", singleinstance);
  config->writeEntry ("Shell expansion", shellexpansion);
  config->writeEntry ("No flow control", noflowcontrol);
  config->writeEntry ("Communicate variables", allowvars);
  config->writeEntry ("Preprocessor script", preprocessor);

  config->writeEntry ("Text", text);
  config->writeEntry ("Type", type);
  config->writeEntry ("Only if match", onlyifmatch);
}

cRunningScript *cScript::prepareToLaunch (const QString &params)
{
  //do nothing if we're a single-instance script that is already running
  if (singleinstance && (runningCount != 0))
  {
    sess->console()->systemMessage (i18n ("Sorry, single-instance script."));
    return 0;
  }

  cRunningList *rl = sess->connection()->getConnPrefs()->runningList();
  if (preprocessor && (rl->havePreprocessorScript()))
  {
    sess->console()->systemMessage
        ("Another preprocessor script is already running.");
    return 0;
  }
  
  //create a new instance of cRunningScript
  cRunningScript *script = new cRunningScript (this);
  
  //setup its parameters:

  //prepare some variables
  if (preprocessor)
  {
    script->prefix = script->suffix = QString::null;
    script->onlyifmatch = false;
    script->sendusercommands = false;
    script->sendstdout = true;
    script->sendstderr = false;
    script->useadvcomm = false;
    script->noflowcontrol = true;
  }
  else
  {
    script->prefix = prefix;
    script->suffix = suffix;
    script->onlyifmatch = onlyifmatch;
    if (enableinput)                //only if input to script is enabled
      script->sendusercommands = sendusercommands;
    else
      script->sendusercommands = false;
    if (enablestdout)
      script->sendstdout = sendstdout;
    else
      script->sendstdout = false;
    if (enablestderr)
      script->sendstderr = sendstderr;
    else
      script->sendstderr = false;

    script->useadvcomm = useadvcomm;

    script->noflowcontrol = noflowcontrol;
  }

  script->preprocessor = preprocessor;

  //establish socket if needed
  if (allowvars)
    script->establishSocket (sess);
  
  //create a KProcess or KShellProcess instance
  if (shellexpansion)
    script->process = new KShellProcess;
  else
    script->process = new KProcess;
  
  script->process->setWorkingDirectory (workdir);

  //split the command into parts...
  QString cmd = command;
  QString pars = params.stripWhiteSpace ();
  if (allowparams && (pars.length() > 0))
    cmd += QString(" ") + pars;
  QString oneparam;
  bool inquotedparam = false;
  QChar quotechar;
  bool wasbackslash = false;
  for (unsigned int i = 0; i < cmd.length(); i++)
  {
    if (wasbackslash)
    {
      oneparam += cmd[i];
      wasbackslash = false;
    }
    else
      if (cmd[i].isSpace())
      {
        if (inquotedparam)
          oneparam += cmd[i];
        else
          if (oneparam.length() > 0)
          {
            if (shellexpansion)
              oneparam = KShellProcess::quote (oneparam);
            *(script->process) << oneparam;  //add this parameter!
            oneparam = QString::null;
          }
      }
      else
        if (cmd[i].latin1() == '\\')    //backslash
          wasbackslash = true;
        else
          if (inquotedparam && (cmd[i] == quotechar)) //quote ends
            inquotedparam = false;
          else
            if ((cmd[i].latin1() == '\'') || (cmd[i].latin1() == '"'))
            {     //quote starts
              inquotedparam = true;
              quotechar = cmd[i];
            }
            else    //normal character
              oneparam += cmd[i];
  }
  if (oneparam.length() > 0)
  {
    if (shellexpansion)
      oneparam = KShellProcess::quote (oneparam);
    *(script->process) << oneparam;  //add last parameter!
  }

  //Great! Command & parameters are set!

  //now we set up communication mode and everything is ready for a launch
  int comm = 0;
  if (preprocessor)
  {
    comm |= KProcess::Stdin;
    comm |= KProcess::Stdout;
    script->comm = comm;
    script->inputenabled = true;
    connect (script->process, SIGNAL (receivedStdout (KProcess *, char *, int)),
            script, SLOT (processScriptStdOutput (KProcess *, char *, int)));
    connect(script->process, SIGNAL(wroteStdin (KProcess *)),
            script, SLOT(stdinReady()));
  }
  else
  {
    if (enableinput)
      comm |= KProcess::Stdin;
    if (enablestdout)
      comm |= KProcess::Stdout;
    if (enablestderr)
      comm |= KProcess::Stderr;  
    script->comm = comm;
    script->inputenabled = enableinput;
    //some connections...
    if (enablestdout)
    {
      connect (script->process, SIGNAL (receivedStdout (KProcess *, char *, int)),
              script, SLOT (processScriptStdOutput (KProcess *, char *, int)));
    }
    if (enablestderr)
    {
      connect (script->process, SIGNAL (receivedStderr (KProcess *, char *, int)),
              script, SLOT (processScriptStdError (KProcess *, char *, int)));
    }
    if (enableinput)
    {
      connect(script->process, SIGNAL(wroteStdin (KProcess *)),
              script, SLOT(stdinReady()));
    } // endif script input enabled
  }
  
  
  //return that instance
  return script;
}

void cScript::scriptIsStarting ()
{
  runningCount++;
}

void cScript::scriptIsTerminating ()
{
  runningCount--;
}
