/***************************************************************************
                          ctelnet.cpp  -  telnet...
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : P Jun 14 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@kmuddy.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#define CTELNET_CPP

#include "ctelnet.h"

#include "cconnection.h"
//we only need cConnPrefs for MSP stuff
#include "cconnprefs.h"

#ifdef HAVE_SOUND
#include "cmsp.h"
#endif

#include "cpluginmanager.h"
#include "csession.h"
#include "kmuddy.h"

#include <unistd.h>
#include <errno.h>
#include <klocale.h>
#include <ksock.h>
#include <kmessagebox.h>
#include <kapplication.h>

#define COMMAND_INTERVAL 50

extern KMuddy *kmuddy;

cTelnet::cTelnet()
{
  socket = NULL;
  timer = new QTimer;
  connect (timer, SIGNAL(timeout()), this, SIGNAL(cmdRequest()));
  //I'm not sure if setting this to KMuddy instead of s.t. well-known
  //(vt100, xterm, ...) won't cause problems. Let's hope for the best :)
  termType = "KMuddy";
  datalen = 0;
  sentbytes = 0;
  LastError = "";
  offLineMode ();
  _connected = false;
  curX = 0;
  curY = 0;
  setQuitCommand (QUIT);
  MCCP = new cMCCP (this);

#ifdef HAVE_SOUND
  MSP = new cMSP (this);
  usingmsp = false;
#endif
}

cTelnet::~cTelnet()
{
  timer->stop ();
  
  if (isConnected ())
    disconnect ();

#ifdef HAVE_SOUND
  usingmsp = false;
  delete MSP;
  MSP = 0;
#endif
  
  delete MCCP;
  MCCP = 0;
  timer->stop ();
  delete timer;
}

/** establishes a new connection */
bool cTelnet::connectIt (const QString &address, int port)
{
  //close existing connection first (if any)
  if (isConnected ())
    disconnect ();

  socket = new KSocket (address, port, 10);
  
  int n = socket->socket ();
  if (n < 0)   //connection was not successful
  {
    //int error = errno;
    //LastError = QString::fromLocal8Bit (strerror (errno));

    //error reporting disabled; will enable it after I switch to
    //KExtendedSocket; the above code was always giving the same reply
    //about operation being now in progress, so it wasn't of much use either
    LastError = i18n ("Sorry, error reporting is disabled for now.");

    delete socket;
    socket = NULL;
    return false;
  }

  _connected = true;
  LastError = i18n ("Everything works perfectly.");

  //reset MCCP, MSP and byte counters
  MCCP->reset ();
#ifdef HAVE_SOUND
  MSP->reset (address);
#endif
  
  sentbytes = 0;
  
  //now we should be connected
  emit connected ();

  //prepare option variables
  for (int i = 0; i < 256; i++)
  {
    myOptionState[i] = false;
    hisOptionState[i] = false;
    announcedState[i] = false;
    heAnnouncedState[i] = false;
    triedToEnable[i] = false;
  }
  //reset buffer
  datalen = 0;
  
  //activate event handlers
  connect (socket, SIGNAL(readEvent (KSocket *)), this, SLOT (socketRead (KSocket *)));
  connect (socket, SIGNAL(closeEvent (KSocket *)), this, SLOT (socketClose (KSocket *)));
  socket->enableRead (true);

  //commands will be sent in 50ms intervals
  timer->start (COMMAND_INTERVAL);

  QString s;
  
  /*  --- OPTIONS DISABLED ---
  //and as we want to enable SUPPRESS-GO-AHEAD option, we send a request to do so...
  QString s = (QChar) TN_IAC;
  s += (QChar) TN_DO;
  s += (QChar) OPT_SUPPRESS_GA;
  triedToEnable[OPT_SUPPRESS_GA] = true;
  //try to negotiate TIMING_MARKs as well...
  s += (QChar) TN_IAC;
  s += (QChar) TN_DO;
  s += (QChar) OPT_TIMING_MARK;
  triedToEnable[OPT_TIMING_MARK] = true;
  */
  //and don't forget the NAWS (used to send info about window size)
  s += (QChar) TN_IAC;
  s += (QChar) TN_WILL;
  s += (QChar) OPT_NAWS;
  //do not allow server to echo our text!
  s += (QChar) TN_IAC;
  s += (QChar) TN_DONT;
  s += (QChar) OPT_ECHO;

  doSendData (s);

  return true;
}

/** closes connection */
void cTelnet::disconnect (bool callConnectionDisconnect)
{
  timer->stop ();
  datalen = 0;

  //send command quit first! - otherwise MUD would say you're net-dead...
  if (isConnected ())
    sendData (quitcommand + "\n");

  //reset quit-command back to default value
  quitcommand = QUIT;

  if (socket != NULL)
  {
    close (socket->socket());
    delete socket;
  }
  socket = NULL;

  //alright - we're disconnected
  emit disconnected ();

  _connected = false;
  
  //call cConnection::disconnect - saves triggers and so on...
  if ((sess->connection() != NULL) && callConnectionDisconnect)
    sess->connection()->disconnect ();

  //remove the session - should now be safe
  kmuddy->removeSession (kmuddy->sessionId(sess), true);
}

void cTelnet::offLineMode ()
{
  timer->start (50);
}

/** are we connected ? */
bool cTelnet::isConnected ()
{
  return ((socket != NULL) && _connected);
}

#ifdef HAVE_SOUND

void cTelnet::setMSPGlobalPaths (const QStringList &paths)
{
  if (MSP != 0)
    MSP->setGlobalPaths (paths);
}

void cTelnet::setMSPAllowed (bool allow)
{
  if (MSP != 0)
    MSP->setMSPAllowed (allow);
}

void cTelnet::setDownloadAllowed (bool allow)
{
  if (MSP != 0)
    MSP->setDownloadAllowed (allow);
}

#endif  //HAVE_SOUND

/** sends data, used by cSendCmd - !!! use cSendCmd class if possible !!! */
bool cTelnet::sendData (const QString &data, bool addTimingRequest, bool use8Bit)
{
  if (!(isConnected()))
    return false;
  if (socket->socket () < 0)
  {
    LastError = QString::fromLocal8Bit (strerror (errno));
    KMessageBox::error (kmuddy, i18n ("Connection has been closed.\nReason: ") + LastError);
    disconnect ();  //something is wrong -- better disconnect...
    return false;
  }

  //IAC byte must be doubled
  QString d;
  int len = data.length();
  for (int i = 0; i < len; i++)
  {
    QChar ch = data[i];
    d += ch;
    if ((unsigned char) ch.latin1() == 255)
      d += ch;  //double IAC
  }

  if (addTimingRequest && hisOptionState[OPT_TIMING_MARK])
  {
    d += (QChar) TN_IAC;
    d += (QChar) TN_DO;
    d += (QChar) OPT_TIMING_MARK;
  }

  if(t_cmdEcho == true && t_lpmudstyle)
    prependGANewLine = false;

  //data ready, send it!
  return doSendData (d, use8Bit);
}

bool cTelnet::doSendData (const QString &data, bool use8Bit)
{
  if (!(isConnected()))
    return false;
  QString d = data;
  //write data to socket - it's so complicated because sometimes only a part of data
  //is accepted at a time
  int dataLength = data.length ();
  int remlen = dataLength;
  do {
    int written;
    if (use8Bit)
      written = write (socket->socket (), (void*)((const char*)d.local8Bit()), remlen);
    else
      written = write (socket->socket (), (void*)(d.latin1()), remlen);
    if (written == -1)  //an error has occured
      return false;
    remlen -= written;
    d.remove (0, written);
  } while (remlen > 0);

  //update counter
  sentbytes += dataLength;
  return true;
}

void cTelnet::windowSizeChanged (int x, int y)
{
  //remember the size - we'll need it if NAWS is currently disabled but will
  //be enabled. Also remember it if no connection exists at the moment;
  //we won't be called again when connecting
  curX = x;
  curY = y;
  if (!(isConnected()))
    return;
  if (myOptionState[OPT_NAWS])   //only if we have negotiated this option
  {
    QString s;
    s = (QChar) TN_IAC;
    s += (QChar) TN_SB;
    s += (QChar) OPT_NAWS;
    unsigned char x1, x2, y1, y2;
    x1 = (unsigned char) x / 256;
    x2 = (unsigned char) x % 256;
    y1 = (unsigned char) y / 256;
    y2 = (unsigned char) y % 256;
    //IAC must be doubled
    s += (QChar) x1;
    if (x1 == TN_IAC)
      s += (QChar) TN_IAC;
    s += (QChar) x2; 
    if (x2 == TN_IAC)
      s += (QChar) TN_IAC;
    s += (QChar) y1;
    if (y1 == TN_IAC)
      s += (QChar) TN_IAC;
    s += (QChar) y2;
    if (y2 == TN_IAC)
      s += (QChar) TN_IAC;
    
    s += (QChar) TN_IAC;
    s += (QChar) TN_SE;
    doSendData (s);
  }
}

void cTelnet::processTelnetCommand (const QString &command)
{
  unsigned char ch = command[1];
    unsigned char option;
    switch (ch) {
    case TN_AYT:
      doSendData ("I'm here! Please be more patient!\r\n");
          //well, this should never be executed, as the response would probably
          //be treated as a command. But that's server's problem, not ours...
          //If the server wasn't capable of handling this, it wouldn't have
          //sent us the AYT command, would it? Impatient server = bad server.
          //Let it suffer! ;-)
      break;
    case TN_GA:
      recvdGA = true;
      //signal will be emitted later
      break;
    case TN_WILL:
      //server wants to enable some option (or he sends a timing-mark)...
      option = command[2];
      heAnnouncedState[option] = true;
      if (triedToEnable[option])
      {
        hisOptionState[option] = true;
        triedToEnable[option] = false;
      }
      else
      {
        if (option == OPT_TIMING_MARK)
          receivedTimingMark = true;
        else
        if (!hisOptionState[option])
            //only if this is not set; if it's set, something's wrong wth the server
            //(according to telnet specification, option announcement may not be
            //unless explicitly requested)
        {
          if ((option == OPT_SUPPRESS_GA) || (option == OPT_STATUS) ||
              (option == OPT_TERMINAL_TYPE) || (option == OPT_NAWS))
                 //these options are supported; compression is handled
                 //separately
          {
            QString s;
            s = (QChar) TN_IAC;
            s += (QChar) TN_DO;
            s += (QChar) option;
            doSendData (s);
            hisOptionState[option] = true;
          }
#ifdef HAVE_SOUND
          else
          if (option == OPT_MSP)
          {
            QString s;
            s = (QChar) TN_IAC;
            s += (QChar) TN_DO;
            s += (QChar) option;
            doSendData (s);
            hisOptionState[option] = true;

            //MSP is now enabled
            usingmsp = true;
            MSP->enableMSP ();
          }
#endif  //HAVE_SOUND
          else
          if ((option == OPT_COMPRESS) || (option == OPT_COMPRESS2))
          //these are handled separately, as they're a bit special
          {
            if ((option == OPT_COMPRESS) && (hisOptionState[OPT_COMPRESS2]))
            {
              //protocol says: reject MCCP v1 if you have previously accepted
              //MCCP v2...
              QString s;
              s = (QChar) TN_IAC;
              s += (QChar) TN_DONT;
              s += (QChar) option;
              doSendData (s);
              hisOptionState[option] = false;
            }
            else
            {
              QString s;
              s = (QChar) TN_IAC;
              s += (QChar) TN_DO;
              s += (QChar) option;
              doSendData (s);
              hisOptionState[option] = true;
              //inform MCCP object about the change
              if ((option == OPT_COMPRESS))
                MCCP->setMCCP1 (true);
              else
                MCCP->setMCCP2 (true);
            }
          }
          else
          {
            QString s;
            s = (QChar) TN_IAC;
            s += (QChar) TN_DONT;
            s += (QChar) option;
            doSendData (s);
            hisOptionState[option] = false;
          }
        }
      }
      break;
    case TN_WONT:
      //server refuses to enable some option...
      option = command[2];
      if (triedToEnable[option])
      {
        hisOptionState[option] = false;
        triedToEnable[option] = false;
        heAnnouncedState[option] = true;
      }
      else
      {
        //send DONT if needed (see RFC 854 for details)
        if (hisOptionState[option] || (!heAnnouncedState[option]))
        {
          QString s = (QChar) TN_IAC;
          s += (QChar) TN_DONT;
          s += (QChar) option;
          doSendData (s);
          hisOptionState[option] = false;
          //inform MCCP object about the change - won't cause problems in
          //cMCCP - see cmccp.cpp for more info
          if ((option == OPT_COMPRESS))
            MCCP->setMCCP1 (false);
          if ((option == OPT_COMPRESS2))
            MCCP->setMCCP2 (false);
        }
        heAnnouncedState[option] = true;
      }
#ifdef HAVE_SOUND
      if (option == OPT_MSP)
      {
        //MSP is now disabled
        usingmsp = false;
        MSP->disableMSP ();
      }
#endif
      break;
    case TN_DO:
      //server wants us to enable some option
      option = command[2];
      if (option == OPT_TIMING_MARK)
      {
        //send WILL TIMING_MARK
        QString s;
        s = (QChar) TN_IAC;
        s += (QChar) TN_WILL;
        s += (QChar) OPT_TIMING_MARK;
        doSendData (s);
      }
      else
      if (!myOptionState[256])
      //only if the option is currently disabled
      {
        if ((option == OPT_SUPPRESS_GA) || (option == OPT_STATUS) ||
            (option == OPT_TERMINAL_TYPE) || (option == OPT_NAWS))
        {

          QString s;
          s = (QChar) TN_IAC;
          s += (QChar) TN_WILL;
          s += (QChar) option;
          doSendData (s);
          myOptionState[option] = true;
          announcedState[option] = true;
        }
        else
        {
          QString s;
          s = (QChar) TN_IAC;
          s += TN_WONT;
          s += option;
          doSendData (s);
          myOptionState[option] = false;
          announcedState[option] = true;
        }
      }
      if (option == OPT_NAWS)  //NAWS here - window size info must be sent
        windowSizeChanged (curX, curY);
      break;
    case TN_DONT:
      //only respond if value changed or if this option has not been announced yet
      option = command[2];
      if (myOptionState[option] || (!announcedState[option]))
      {
        QString s;
        s = (QChar) TN_IAC;
        s += (QChar) TN_WONT;
        s += (QChar) option;
        doSendData (s);
        announcedState[option] = true;
      }
      myOptionState[option] = false;
      break;
    case TN_SB:
      //subcommand - we analyze and respond...
      option = command[2];
      switch (option) {
        case OPT_STATUS:
          //see OPT_TERMINAL_TYPE for explanation why I'm doing this
          if (true /*myOptionState[OPT_STATUS]*/)
          {
            if (command[3].latin1() == TNSB_SEND)
            //request to send all enabled commands; if server sends his
            //own list of commands, we just ignore it (well, he shouldn't
            //send anything, as we do not request anything, but there are
            //so many servers out there, that you can never be sure...)
            {
              QString s;
              s = (QChar) TN_IAC;
              s += (QChar) TN_SB;
              s += (QChar) OPT_STATUS;
              s += (QChar) TNSB_IS;
              for (int i = 0; i < 256; i++)
              {
                if (myOptionState[i])
                {
                  s += (QChar) TN_WILL;
                      //nice line, isn't it? :)
                  s += (QChar) ((unsigned char) i);
                }
                if (hisOptionState[i])
                {
                  s += (QChar) TN_DO;
                  s += (QChar) ((unsigned char) i);
                }
              }
              s += (QChar) TN_IAC;
              s += (QChar) TN_SE;
              doSendData (s);
            }
          }
        break;
        case OPT_TERMINAL_TYPE:
          if (myOptionState[OPT_TERMINAL_TYPE])
          {
            if (command[3].latin1() == TNSB_SEND)
              //server wants us to send terminal type; he can send his own type
              //too, but we just ignore it, as we have no use for it...
            {
              QString s;
              s = (QChar) TN_IAC;
              s += (QChar) TN_SB;
              s += (QChar) OPT_TERMINAL_TYPE;
              s += (QChar) TNSB_IS;
              s += termType;
              s += (QChar) TN_IAC;
              s += (QChar) TN_SE;
              doSendData (s);
            }
          }
        break;
        //other cmds should not arrive, as they were not negotiated.
        //if they do, they are merely ignored
      };
      break;
    //other commands are simply ignored (NOP and such, see .h file for list)
  };
}

void cTelnet::socketRead (KSocket *)
{
  /*
  This function makes heavy use of char* instead of relying on QString,
  because otherwise there are problems with locale-based data and/or
  telnet commands.
  */

  char buffer[32769];  //we should never receive such a big amount of data,
        //but we want to be sure that we read all we can...

  char buffer2[32769]; //decompression buffer

  int amount = read (socket->socket(), (void*)buffer, 32768);
  if (amount == -1)
    return;   //something is wrong (no data?)
  if (amount == 0)
    return;   //0 means socket has been closed; maybe I'll do something more?
  buffer[amount] = '\0';    //just to be sure - mark end of string

  //inform plug-ins about newly received text
  pluginmanager->passRawData (sess, buffer);

  //MCCP will be handled first. This is done in a separate class, which
  //detects all MCCP-related codes itself. I know that this involves some
  //code duplication (two telnet option parsers), but it's needed because
  //telnet sequences may (and will) appear inside the compressed streams,
  //so it's best to get plain uncompressed data first. Another reason is
  //improper design of MCCP v1, which uses unterminated telnet subsequences.
  //This has been fixed in MCCP v2, but some MUDs may still use the old
  //version.

  MCCP->prepareDecompression (buffer, buffer2, amount, 32768);
  int size;
  while ((size = MCCP->uncompressNext ()) != -1)
  {
    //inform plug-ins about uncompressed data (can be equal to raw data if MCCP isn't used
    pluginmanager->passDecompressedData (sess, buffer2);

    //we add buffer to existing data (if any)... I do this, because previous
    //data could end in the middle of a telnet command, which was then left
    //in the buffer for subsequent reading (= this one ;)
    for (int i = 0; i < size; i++)
      data[datalen + i] = buffer2[i];
    datalen += size;
    data[datalen] = 0;

    //we have some data...
    newdata = true;

    //clear the GO-AHEAD flag
    recvdGA = false;
    
    //now we have the data, but we cannot forward it to next stage of processing,
    //because the data contains telnet commands
    //so we parse the text and process all telnet commands:

    QString cleandata, command;
    bool iac, iac2, insb;
      //iac: last char was IAC
      //iac2: last char was DO, DONT, WILL or WONT
      //insb: we're in IAC SB, waiting for IAC SE
    int processed = -1;
    iac = iac2 = insb = false;
    cleandata = "";
    command = "";
    receivedTimingMark = false;

    for (unsigned int i = 0; i < datalen; i++)
    {
      unsigned char ch = data[i];
      if (iac || iac2 || insb || (ch == TN_IAC))
      {
        //there are many possibilities here:
        //1. this is IAC, previous character was regular data
        if (! (iac || iac2 || insb) && (ch == TN_IAC))
        {
          iac = true;
          command += ch;
        }
        else
        //2. seq. of two IACs
        if (iac && (ch == TN_IAC) && (!insb))
        {
          iac = false;
          char str[2];
          str[0] = ch;
          str[1] = 0;
          cleandata += QString::fromLocal8Bit (str);
          processed = i;
          command = "";
        }
        else
        //3. IAC DO/DONT/WILL/WONT
        if (iac && (!insb) &&
            ((ch == TN_WILL) || (ch == TN_WONT) || (ch == TN_DO) || (ch == TN_DONT)))
        {
          iac = false;
          iac2 = true;
          command += ch;
        }
        else
        //4. IAC DO/DONT/WILL/WONT <command code>
        if (iac2)
        {
          iac2 = false;
          command += ch;
          processTelnetCommand (command);
          processed = i;
          command = "";
          //TIMING-MARK may have been received - if it was, process it!
          if (receivedTimingMark)
          {
            //process data that came before the timing-mark first...
            if(prependGANewLine && t_lpmudstyle)
            {
              cleandata = "\n" + cleandata;
              emit dataReceived (cleandata);
              prependGANewLine = false;
            }
            else
            {
              emit dataReceived (cleandata);
            }
            //now that the data is processed, clean the buffer and inform others
            //about the TIMING-MARK. Then continue processing...
            cleandata = "";
            emit timingMark ();
            receivedTimingMark = false;
          }
        }
        else
        //5. IAC SB
        if (iac && (!insb) && (ch == TN_SB))
        {
          iac = false;
          insb = true;
          command += ch;
        }
        else
        //6. IAC SE without IAC SB - error - ignored
        if (iac && (!insb) && (ch == TN_SE))
        {
          command = "";
          iac = false;
        }
        else
        //7. inside IAC SB
        if (insb)
        {
          command += ch;
          if (iac && (ch == TN_SE))  //IAC SE - end of subcommand
          {
            processTelnetCommand (command);
            processed = i;
            command = "";
            iac = false;
            insb = false;
          }
          if (iac)
            iac = false;
          else
          if (ch == TN_IAC)
            iac = true;
        }
        else
        //8. IAC fol. by something else than IAC, SB, SE, DO, DONT, WILL, WONT
        {
          iac = false;
          command += ch;
          processTelnetCommand (command);
          //this could have set receivedGA to true; we'll handle that later
          // (at the end of this function)
          processed = i;
          command = "";
        }
      }
      else   //plaintext
      {
        //everything except CRLF is okay; CRLF is replaced by LF(\n) (CR ignored)
        if (ch != 13)
        {
          char str[2];
          str[0] = ch;
          str[1] = 0;
          cleandata += QString::fromLocal8Bit (str);
          processed = i;
        }
      }
    }

    //shift data (so that telnet commands broken into multiple packets
    // are handled correctly)
    for (unsigned int i = 0; i < datalen - (processed + 1); i++)
      data[i] = data[(processed + 1) + i];
    datalen -= (processed + 1);

    cConnPrefs *cp = sess->connection()->getConnPrefs();
    //ask MSP parser to look for MSP tags if MSP is enabled
#ifdef HAVE_SOUND
    //look if the user want to have MSP all the time
    bool alwaysmsp = false;
    if (cp)
      alwaysmsp = cp->alwaysMSP();
    //hand data to MSP parser if desired
    if (usingmsp || alwaysmsp)
      cleandata = MSP->parseServerOutput (cleandata);
#endif
    
    //forward data for further processing
    if(prependGANewLine && t_lpmudstyle)
    {
      cleandata = "\n" + cleandata;
      emit dataReceived (cleandata);
      prependGANewLine = false;
    }
    else
    {
      emit dataReceived (cleandata);
    }

    //if we have received GO-AHEAD signal, let others know about this!
    if (recvdGA)
    {
      if(cp && cp->consolePrompt())
	      prependGANewLine = true;
      emit receivedGA ();
    }
    //no need to clear that flag, it will be done in next iteration
  }

  emit moreTextHere ();
}

void cTelnet::socketClose (KSocket *)
{
  _connected = false;
  disconnect ();
  (KApplication::kApplication())->processEvents ();
}

void cTelnet::setQuitCommand (QString cmd)
{
  quitcommand = cmd;
}

void cTelnet::setCommandEcho (bool cmdEcho)
{
  t_cmdEcho = cmdEcho;
}

void cTelnet::setLPMudStyle (bool lpmudstyle)
{
  t_lpmudstyle = lpmudstyle;
}
