/***************************************************************************
                          ctelnet.h  -  handles telnet connection
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : P Jun 14 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@host.sk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
Current status of cTelnet class (Oct 18, 2002):
 - all required functionality has been implemented
 - new features may still be added
 - everything seems to work well...
 - uses KSocket again, because of some bugs in KExtendedSocket :(((
*/

#ifndef CTELNET_H
#define CTELNET_H

#include <config.h>

#include <qobject.h>
#include <qtimer.h>

#include "csessionbase.h"
#include "cmccp.h"

#ifdef HAVE_SOUND
class cMSP;
#endif

class KSocket;

/**
cTelnet handles the connection and telnet commands
However, as this is a MUD client and not a telnet, only a small subset of
telnet commands is handled.

List of direct telnet commands and their support in this class:
=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
240 - SE (subcommand end) - supported
241 - NOP (no operation) - ignored ;-)
242 - DM (data mark) - NOT USED - Synch is not implemented (well,
  maybe I'll implement this later if reall necessary...)
243 - B (break) - not used (only client may use this, so the server won't bother with this)
244 - IP (interrupt process) - not applicable (only useful for shell processes)
245 - AO (about output) - not applicable (see IP)
246 - AYT (are you there) - I do not send this; "I'm here" is sent
  in an unlikely event that AYT is received
247 - EC (erase char) - not applicable (line editing fully handled by client)
248 - EL (erase line) - not applicable (line editing fully handled by client)
249 - GA (go ahead) - ignored; suppress-go-ahead is negotiated if possible
250 - SB (subcommand begin) - supported
251 - WILL - supported
252 - WON'T - supported
253 - DO - supported
254 - DON'T - supported
255 - IAC - supported

Extended telnet commands:
=-=-=-=-=-=-=-=-=-=-=-=-=
  these are handled with the IAC DO/DON'T/WILL/WON'T and SB/SE mechanism;
  see RFC 854 and RFCs listed here for further information.

a) FULLY SUPPORTED COMMANDS:
- STATUS (cmd 5, RFC 859) - sending/receiving option status
- TIMING-MARK (cmd 6, RFC 860) - timing marks, used by auto-mapper
- TERMINAL-TYPE (cmd 24, RFC 1091) - terminal type sending
- NAWS (cmd 31, RFC 1073) - negotiate about window size - used to inform
  the server about window size (cols x rows)
- MCCP v1 (cmd 85) - compression protocol version 1; handled by cMCCP class
- MCCP v2 (cmd 86) - compression protocol version 2; handled by cMCCP class
  
b) PARTIALLY SUPPORTED COMMANDS
- SUPPRESS-GO-AHEAD (cmd 3, RFC 858) - we try to suppress GA's if possible.
  If we fail, we just ignore all GA's, hoping for no problems...

c) COMMANDS THAT ARE NOT SUPPORTED
The following commands are not supported at all - they are all disabled,
if the server asks us to enable it, we respond with DON'T or WON'T (this
is a standard behaviour described in RFC 854)
- TRANSMIT-BINARY (cmd 0, RFC 856) (see note below)
- ECHO (cmd 1, RFC 857) (see note below)
- EXTENDED-OPTIONS-LIST (cmd 255, RFC 861) (see note below)
- RCTE (cmd 7, RFC 726)
- NAOCRD (cmd 10, RFC 652)
- NAOHTS (cmd 11, RFC 653)
- NAOHTD (cmd 12, RFC 654)
- NAOFFD (cmd 13, RFC 655)
- NAOVTS (cmd 14, RFC 656)
- NAOVTD (cmd 15, RFC 657)
- NAOLFD (cmd 16, RFC 658)
- EXTEND-ASCII (cmd 17, RFC 698)
- LOGOUT (cmd 18, RFC 727)
- BM (cmd 19, RFC 735)
- DET (cmd 20, RFC 1043)
- SUPDUP (cmd 21, RFC 736)
- SUPDUP-OUTPUT (cmd 22, RFC 749)
- SEND-LOCATION (cmd 23, RFC 779) (well, maybe one day I'll support this)
- END-OF-RECORD (cmd 25, RFC 885)
- TUID (cmd 26, RFC 927)
- OUTMARK (cmd 27, RFC 933)
- TTYLOC (cmd 28, RFC 946)
- 3270-REGIME (cmd 29, RFC 1041)
- X.3-PAD (cmd 30, RFC 1053)
- TERMINAL-SPEED (cmd 32, RFC 1079)  (not needed, let the server decide; we can
  accept data at (almost) any speed, only limited by connection speed)
- TOGGLE-FLOW-CONTROL (cmd 33, RFC 1372) (maybe I'll support it, if it
  proves to be necessary/useful)
- LINEMODE (cmd 34, RFC 1184) - linemode is not needed, as we send the entire
  line either...
- NEW-ENVIRON (cmd 39, RFC 1572)
- TN3270E (cmd 40, RFC 1647)

Note: the first three commands are internet standards and each telnet application
should support them. However, this is NOT a telnet application and those
commands have no use for a MUD client, so I leave those commands unsupported.
Who knows, maybe one day I'll change my opinion and implement them (or somebody
else does it ;))

  *@author Tomas Mecir
  */


//telnet command codes (prefixed with TN_ to prevent duplicit #defines
#define TN_SE 240
#define TN_NOP 241
#define TN_DM 242
#define TN_B 243
#define TN_IP 244
#define TN_AO 245
#define TN_AYT 246
#define TN_EC 247
#define TN_EL 248
#define TN_GA 249
#define TN_SB 250
#define TN_WILL 251
#define TN_WONT 252
#define TN_DO 253
#define TN_DONT 254
#define TN_IAC 255

//telnet option codes (supported options only)
#define OPT_ECHO 1
#define OPT_SUPPRESS_GA 3
#define OPT_STATUS 5
#define OPT_TIMING_MARK 6
#define OPT_TERMINAL_TYPE 24
#define OPT_NAWS 31
#define OPT_COMPRESS 85
#define OPT_COMPRESS2 86
#define OPT_MSP 90
#define OPT_MXP 91

//telnet SB suboption types
#define TNSB_IS 0
#define TNSB_SEND 1

//default quit command
#define QUIT "quit"

class cTelnet : public QObject, public cSessionBase {
Q_OBJECT
public: 
  cTelnet();
  ~cTelnet();
  /** establishes a new connection */
  bool connectIt (const QString &address, int port);
  /** closes connection */
  void disconnect (bool callConnectionDisconnect = true);
  
  /** returns whether we're connected to some other host */
  bool isConnected ();
  void setTermType (const QString & ttype) { termType = ttype; };
  const QString &lastError () { return LastError; };
  /** these two functions control waiting for data from server */
  void waitingForData () { newdata = false; };
  bool newData () { return newdata; };
  void setQuitCommand (QString cmd);
  /** enters off-line mode - commands are accepted but not sent */
  void offLineMode ();
  int compressedBytes () { return MCCP->compressedBytes (); };
  int uncompressedBytes () { return MCCP->uncompressedBytes (); };
  int sentBytes () { return sentbytes; };
  bool usingMCCP () { return MCCP->usingMCCP(); };
  int MCCPVer () { return MCCP->MCCPVer(); };

  /* Command echo setting */
  void setCommandEcho(bool cmdEcho);

  /* LPMud prompt style */
  void setLPMudStyle (bool lpmustyle);
  #ifdef HAVE_SOUND
  /** are we currently using MSP? */
  bool usingMSP () { return usingmsp; };
  void setMSPAllowed (bool allow);
  void setDownloadAllowed (bool allow);
  void setMSPGlobalPaths (const QStringList &paths);
  #endif
signals:
  /** new command requested - cSendCmd is connected here */
  void cmdRequest ();
  /** something has arrived from the server - cANSIParser is connected here... */
  void dataReceived (const QString &data);
  /** we're now connected to the server */
  void connected ();
  /** we're now disconnected from the server */
  void disconnected ();
  /** TIMING-MARK received */
  void timingMark ();
  /** GO-AHEAD received */
  void receivedGA ();
  /** used to display partial line in status bar */
  void moreTextHere ();
public slots:
  /** sends data, used by cSendCmd - !!! use cSendCmd class if possible !!!
  This function doubles IAC bytes and therefore isn't suitable for sending
  telnet sequences; use doSendData (protected) for that purpose. */
  bool sendData (const QString &data, bool addTimingRequest = false, bool use8Bit = false);
  /** window size has changed */
  void windowSizeChanged (int x, int y);
protected:
  /** Send out the data. Does not double IACs, this must be done by caller
  if needed. This function is suitable for sending telnet sequences. */
  bool doSendData (const QString &data, bool use8Bit = false);
  /** socket - one of the most important things in the program :) */
  KSocket *socket;
  /** object that handles MCCP */
  cMCCP *MCCP;

  #ifdef HAVE_SOUND
  /** object that handles MSP */
  cMSP *MSP;
  /** are we currently using MSP? */
  bool usingmsp;
  #endif
  
  QTimer *timer;
  bool receivedTimingMark;
  /** type of terminal - sent to server if needed */
  QString termType;
  /** buffer for received data */
  char data[65536];
  unsigned int datalen;
  /** description of last error */
  QString LastError;
  /** quit command */
  QString quitcommand;
  /** has any data arrived since last call to waitingForData() */
  bool newdata;
  /** are we connected?*/
  bool _connected;  
  /** current state of options on our side and on server side */
  bool myOptionState[256], hisOptionState[256];
    /** whether we have announced WILL/WON'T for that option (if we have, we don't
    respond to DO/DON'T sent by the server -- see implementation and RFC 854
    for more information... */
  bool announcedState[256];
    /** whether the server has already announced his WILL/WON'T */
  bool heAnnouncedState[256];;
    /** whether we have tried to enable this option */
  bool triedToEnable[256];
    /** amount of bytes sent up to now */
  int sentbytes;
  /** have we received the GA signal? */
  bool recvdGA;
  /** should we prepend newline after receving a GA */
  bool prependGANewLine;
  bool t_cmdEcho;
  bool t_lpmudstyle;
  /** current dimensions */
  int curX, curY;
  
    /** protected function that processes a telnet command (IAC ...) */
  void processTelnetCommand (const QString &command);
protected slots:
  void socketRead (KSocket *);
  void socketClose (KSocket *);
};

#endif
