/***************************************************************************
                          ctextprocessor.cpp  -  text processor
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : Po oct 14 2002
    copyright            : (C) 2002-2003 by Tomas Mecir
    email                : kmuddy@kmuddy.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ctextprocessor.h"

#include <knotifyclient.h>

#include <qtimer.h>

#include "cconnection.h"
#include "cconnprefs.h"
#include "cconsole.h"
#include "cpluginmanager.h"
#include "crunninglist.h"
#include "ctriggerlist.h"
#include "cwindowlist.h"
#include "csession.h"

cTextProcessor::cTextProcessor ()
{
  defcolor = CL_WHITE;
  defbkcolor = CL_BLACK;
  flush ();
  gotprompt = false;

  elapsedticks = 0;
  pdtimer = new QTimer;
  connect (pdtimer, SIGNAL (timeout ()), this, SLOT (timeout ()));
  pdtimer->start (250);
}

cTextProcessor::~cTextProcessor ()
{
  pdtimer->stop ();
  delete pdtimer;
}

void cTextProcessor::preprocess (const QString &text)
{
  //let plug-ins preprocess the text first
  QString txt = text;  //drop the const flag
  pluginmanager->preprocessInput (sess, txt);

  cConnPrefs *cp = sess->connection()->getConnPrefs();
  if (cp)
  {
    cRunningList *rl = cp->runningList();
    if (rl->havePreprocessorScript ())
    {
      //send the text to preprocessor, if we have one
      rl->sendToPreprocessor (txt);
      return;
    }
  }
  //else we do nothing
  emit preprocessed (txt);
}

void cTextProcessor::preprocessorOutput (const QString &text)
{
  //text should be in format <number> <text>
  //where number is 1 or 2, 1 means standard text (one line), 2 means prompt
  if (text.isEmpty()) return;
  char ch = text[0].latin1();
  QString t = text;
  t.remove (0, 2);
  if (ch == '1')
  {
    //if a prompt is the last thing displayed, we send a newline, so that
    //the text coming from preprocessor gets to the new line. Note that
    //gotprompt is set to false if the user sends some command, so we needn't
    //worry about that
    if (gotprompt)
      emit preprocessed ("\n");
    
    //send the line
    emit preprocessed (t + "\n");
  }
  if (ch == '2')
  {
    emit preprocessed (t);
    //... processing takes place ...
    //... we regain control when everything is set for the following stuff:
    //part of receivedGA():
    if (!gotprompt)
    {
      pluginmanager->passPrompt (sess, line, tcolor, bcolor);
      emit receivedPrompt (line);
      emit displayPrompt (line, tcolor, bcolor);
      line = tcolor = bcolor = QString::null;
      gotprompt = true;
    }
  }
}

void cTextProcessor::newText (const QString &text)
{
  //reset tick-counter
  elapsedticks = 0;

  QChar f, b;
  int fg, bg;
  fg = negactive ? curbkcolor : curcolor;
  bg = negactive ? curcolor : curbkcolor;

  if (underline) fg += 16;
  if (italics) fg += 32;
  if (blinkactive) fg += 64;
  if (strikeout) fg += 128;

  f = QChar (fg);
  b = QChar (bg);

  int len = text.length();
  int curpos = 0;
  for (int i = 0; i < len; i++)
  {
    if (text[i] == '\n')    //end-of-line
    {
      newLine();
      curpos = 0;
    }
    else if (text[i] == '\t')  //TAB character
    {
      //add spaces till next position divisible by 8
      do {
        line += QChar (' ');
        tcolor += f;
        bcolor += b;
        curpos++;
      } while (curpos % 8 != 0);
    }
    else if (text[i] == '\a')  //beep
    {
      KNotifyClient::beep ();
    }
    else
    {
      line += invisible ? QChar (' ') : text[i];
      tcolor += f;
      bcolor += b;
      curpos++;
    }
  }

  //prompt detection (only if no prompt has been recvd since end of line)
  if ((!line.isEmpty()) && (!gotprompt))
  {
    haveprompt = false;

    cConnPrefs *cp = sess->connection()->getConnPrefs ();
    //if cp is NULL, we're not connected using a profile => no triggers
    if (cp != NULL)
    {
      cTriggerList *triggers = cp->triggerList();
      if (triggers != NULL)
      {
        //only prompt-detecting triggers will be executed here
        triggers->setDetectingPrompt (true);
        triggers->matchString (line);
      }
      if (haveprompt)
        //if triggers say that this is a prompt, then we behave as if
        //telnet-GA was received
        receivedGA ();
    }
  }
}

void cTextProcessor::sendingCommand ()
{
  //we can receive a prompt again, even if no newline arrives
  //this is to support things like 'Name:', then 'Password:', but
  //no newline between them (as that comes from the user)
  gotprompt = false;
}

void cTextProcessor::newLine ()
{
  //inform plug-ins - phase 1 (before triggers)
  pluginmanager->passInput (sess, 1, line, tcolor, bcolor, false);

  //is a gagging trigger gets activated, it will call gagLine(), which will
  //set this flag
  gag = false;
  //the same for primary gag, which is set by window-redirection triggers
  primarygag = false;
  //also clear window name
  wname = QString::null;

  //first call trigger comparison/execution...
  cConnPrefs *cp = sess->connection()->getConnPrefs ();
  //if cp is NULL, we're not connected using a profile => no triggers
  if (cp != NULL)
  {
    cTriggerList *triggers = cp->triggerList();
    if (triggers != NULL)
    {
      triggers->setDetectingPrompt (false);
      triggers->matchString (line);
    }
  }

  //inform plug-ins - phase 2 (after triggers)
  pluginmanager->passInput (sess, 2, line, tcolor, bcolor, gag);

  //now emit some signals, if the line shouldn't be gagged...
  //the cConsole class is connected here, amongst others
  if (!gag)
  {
    if (wname)
    {
      if (cp)
      {
        cWindowList *wlist = cp->windowList();
        //send the text to the output window, if it exists
        if (wlist->exists (wname))
          wlist->textToWindow (wname, line, tcolor, bcolor);
        else
          //if the window doesn't exist, display the text in the main window, even if it would
          //normally be gagged
          //this is to prevent loss of information in case that the user removes some window, but
          //keeps triggers that use that window
          primarygag = false;
      }
    }
    if (!primarygag)
      emit displayLine (line, tcolor, bcolor);
    else
      emit transcriptLine (line, tcolor, bcolor);
  }
  else
    emit transcriptLine (line, tcolor, bcolor);
  //textLine is emitted even if line is gagged - this means that scripts will receive the line
  //even if it is gagged... Transcript uses displayLine so gagged lines won't be written into
  //session transcript
  emit textLine (line);

  //clear old line data
  bcolor = tcolor = line = QString::null;

  //increase line counter (even if line was gagged)
  lines++;

  //we can receive prompt again
  gotprompt = false;
}

void cTextProcessor::changeColor (int color)
{
  //text color
  if ((color >= 30) && (color <= 37))
  {
    curcolor = color - 30;
    if (boldactive)
      curcolor |= CL_BRIGHT;
  }
  else
  //text background
  if ((color >= 40) && (color <= 47))
  {
    curbkcolor = color - 40;
  }
  else
  {
    //color code definitions from Ecma-048, page 61
    switch (color) {
      case 0:  //default color
        curbkcolor = defbkcolor;
        curcolor = defcolor;
        blinkactive = false;
        boldactive = false;
        italics = false;
        underline = false;
        strikeout = false;
        invisible = false;
        negactive = false;
        break;
      case 1:   //BOLD
        curcolor |= CL_BRIGHT;
        boldactive = true;
        break;
      case 2:   //FAINT (DECREASED INTENSITY)
        //this is handled as BOLD OFF, I think it's sufficient
        boldactive = false;
        curcolor &= ~CL_BRIGHT;
        break;
      case 3:   //ITALICS
        italics = true;
        break;
      case 4:   //UNDERLINE
        underline = true;
        break;
      case 5:   //BLINK
        blinkactive = true;
        break;
      case 6:   //RAPID BLINK
        //will blink, but NOT rapidly - rapid blinking is pure evil
        //both 5 and 6 are cancelled by 25, so we don't have to manage two
        //separate flags for this
        blinkactive = true;
        break;
      case 7:   //NEGATIVE IMAGE
        negactive = true;
        break;
      case 8:   //INVISIBLE
        invisible = true;
        break;
      case 9:   //STRIKE OUT
        strikeout = true;
        break;
      //10-19 are font changes. These are not supported.
      case 20:  //FRAKTUR (GOTHIC)
        //gothic is treated as italics
        //both italics and gothic are turned off with 23 so I don't need to
        //maintain two separate flags
        italics = true;
        break;
      case 21:  //DOUBLE UNDERLINE
        //this is handled as single underline
        //one flag is enough, as both options are disabled with 24
        underline = true;
        break;
      case 22:  //BOLD OFF
        boldactive = false;
        curcolor &= ~CL_BRIGHT;
        break;
      case 23:  //ITALICS OFF
        italics = false;
        break;
      case 24:  //UNDERLINE OFF
        underline = false;
        break;
      case 25:  //BLINK OFF
        blinkactive = false;
      //26 is proportional spacing - NOT SUPPORTED
      case 27:  //POSITIVE IMAGE
        negactive = false;
        break;
      case 28:  //VISIBLE
        invisible = false;
        break;
      case 29:  //NOT STRIKED OUT
        strikeout = false;
        break;
      //30-37 are colors, see above
      //38 is some special color, not supported
      case 39:  //default color, keep background
        curcolor = defcolor;
        if (boldactive)
          curcolor |= CL_BRIGHT;
        break;
      //40-47 are colors, see above
      //48 is some special color, not supported
      case 49:  //default background, keep text color
        curbkcolor = defbkcolor;
      //50 means proportional spacing off - not supported
      //codes >50 are not supported (these include special effects like
      // framed, encircled, overlined and so...
    };
  }
}

void cTextProcessor::flush ()
{
  line = "";
  lines = 0;

  boldactive = false;
  blinkactive = false;
  italics = false;
  underline = false;
  strikeout = false;
  negactive = false;
  invisible = false;

  curcolor = defcolor;
  curbkcolor = defbkcolor;
}

void cTextProcessor::receivedGA ()
{
  //ignore GA signals if a preprocessor script exists
  //just inform the running script that GA is here
  cConnPrefs *cp = sess->connection()->getConnPrefs();
  if (cp)
  {
    cRunningList *rl = cp->runningList();
    if (rl->havePreprocessorScript ())
    {
      rl->sendToPreprocessor ("\x01\n");
      return;
    }
  }
  
  if (!gotprompt)  //avoid 2+ prompts on one line
  {
    pluginmanager->passPrompt (sess, line, tcolor, bcolor);
    emit receivedPrompt (line);
    emit displayPrompt (line, tcolor, bcolor);
    line = tcolor = bcolor = QString::null;
    gotprompt = true;
  }
}

void cTextProcessor::moreTextHere ()
{
  if (!line.isEmpty())
    emit showPartialLine (line);
}

int cTextProcessor::defaultTextColor ()
{
  return defcolor;
}

void cTextProcessor::setDefaultTextColor (int index)
{
  if (curcolor == defcolor)
    curcolor = index;
  defcolor = index;
}

int cTextProcessor::defaultBkColor ()
{
  return defbkcolor;
}

void cTextProcessor::setDefaultBkColor (int index)
{
  if (curbkcolor == defbkcolor)
    curbkcolor = index;
  defbkcolor = index;
  sess->console()->setDefaultBkColor (index);
}

void cTextProcessor::recolorize (int *newfg, int *newbg)
{
  int len = line.length();
  for (int i = 0; i < len; i++)
  {
    if ((newfg[i] >= 0) && (newfg[i] <= 255))
      //set color but keep attributes
      tcolor[i] = QChar
        (newfg[i] % 16 +(int(tcolor[i].latin1()) & (128+64+32+16)));
    if ((newbg[i] >= 0) && (newbg[i] <= 255))
      bcolor[i] = QChar (newbg[i]);
  }
}

void cTextProcessor::timeout ()
{
  //If some part of the line is waiting, and no prompt has been received
  //so far, we increase our counter. If the counter reaches 4 and no more
  //text arrives, we assume that the waiting text is a prompt and we
  //treat it as such.
  
  if ((!gotprompt) && (line.length() > 0))
    elapsedticks++;
  if (elapsedticks == 4)  //it's probably a prompt
    receivedGA ();  
}
