/***************************************************************************
                          ctranscript.cpp  -  session transcript
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : Ct oct 31 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@host.sk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ctranscript.h"

#include "cconnection.h"
#include "cconnectioninfo.h"
#include "cconnprefs.h"
#include "cconsole.h"
#include "csession.h"
#include "ctelnet.h"
#include "ctextprocessor.h"
#include "kmuddy.h"
#include "dialogs/dlgtranscript.h"
#include "dialogs/dlgdumpbuffer.h"

#include <errno.h>

#include <qdir.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kconfig.h>
#include <kstandarddirs.h>

extern KMuddy *kmuddy;

cTranscript::cTranscript ()
{
  running = false;
  advrunning = false;
  file = NULL;
  advfile = NULL;
  overwrite = true;
  allowansi = false;
  includedump = false;
}

cTranscript::~cTranscript ()
{
  stopTranscript ();
  stopAdvTranscript();
}

void cTranscript::addLineToTranscript (const QString &what,
    const QString &tcolor, const QString &bcolor)
{
  //add the text
  addTextToTranscript (what, tcolor, bcolor);
  
  //append newline
  addToTranscript ("\n");

  fflush (file);
}

void cTranscript::addLineToAdvTranscript (const QString &what)
{
  QString timestamp;

  //add the text
  if (advrunning)
  {
    const char *ch = what.latin1();
    if (ch)
    {
      if(includetimestamp)
      {
        QTime time = QTime::currentTime();
        timestamp.sprintf("[%02d:%02d:%02d] ", time.hour(), time.minute(), time.second());
        fputs (timestamp.latin1(), advfile);
      }
      fputs (ch, advfile);
    }
    fputs("\n", advfile);

    fflush (advfile);
  }
 
}

void cTranscript::addTextToTranscript (const QString &what,
    const QString &tcolor, const QString &bcolor)
{
  /*
    This saves a line of text, optionally including ANSI sequences;
    it only supports color codes, not attributes like bold/italics/...
  */
  if (what.isEmpty())
    return;
  
  if (running)
    if (allowansi)  //saving ANSI sequences
    {
      QString text;
      int fg = -1;
      int bg = -1;
      int len = what.length();
      for (int i = 0; i < len; i++)
      {
        int newfg = ((int) tcolor[i].latin1()) % 16;
        int newbg = (int) bcolor[i].latin1();
        if ((newfg != fg) || (newbg != bg))
        {
          addToTranscript (text);
          text = QString::null;
          addANSIToTranscript (constructANSI (newfg, newbg));
          fg = newfg;
          bg = newbg;
        }
        text += what[i];
      }
      if (!(text.isEmpty()))
        addToTranscript (text);
    }
    else
      addToTranscript (what);
  fflush (file);
}

void cTranscript::addToTranscript (const QString &what)
{
  if (running)
  {
    const char *ch = what.latin1();
    if (ch)
      fputs (ch, file);
  }
}

void cTranscript::addANSIToTranscript (const QString &what)
{
  if (allowansi)
    addToTranscript (what);
}

QString cTranscript::constructANSI (int fg, int bg)
{
  char command[13]; //max sequence length is 10, +1 for \0

  //no change - nothing to do
  if ((curfg == fg) && (curbg == bg))
    return QString::null;
  
  //construct some ANSI command...
  int idx = 0;                        
  command[idx++] = 27;
  command[idx++] = '[';
  //backgrounds 8-15 not supported by ANSI sequences :(
  bg = bg % 8;
  if (curfg != fg)
  {
    if (fg > 7)
    {
      command[idx++] = '1';
      command[idx++] = ';';
    }
    fg = fg % 8;
    command[idx++] = '3';
    command[idx++] = '0' + fg;

    curfg = fg;
    if (curbg != bg)    //will also include background color
      command[idx++] = ';';
  }
  if (curbg != bg)
  {
    command[idx++] = '4';
    command[idx++] = '0' + bg;
  }
  command[idx++] = 'm';
  command[idx++] = 0; //end that string!

  //and return the result
  return QString (command);
}

void cTranscript::addCommandToTranscript (const QString &what)
{
  if (!running)
    return;

  if (allowansi)
  {
    int defbkcolor = sess->textproc()->defaultBkColor();
    int echocolor;
    int systemcolor;
    sess->console()->getAllColors (&echocolor, &systemcolor);
    
    addToTranscript (constructANSI (echocolor, defbkcolor));
  }
  addToTranscript (what);
  addToTranscript ("\n");
}

void cTranscript::stopTranscript ()
{
  if (running)
  {
    fputs ("\n\n", file);
    if (allowansi)
    {
      //set output color to default (usually gray on black)
      char defcolor[5];
      defcolor[0] = 27;
      defcolor[1] = '[';
      defcolor[2] = '0';
      defcolor[3] = 'm';
      defcolor[4] = 0;
      fputs (defcolor, file);
    }
    QString s = i18n ("Session transcript has been stopped.");
    fputs (s.latin1(), file);
    fputs ("\n\n", file);
    running = false;
    fclose (file);
    file = NULL;
    sess->console()->systemMessage (i18n ("Session transcript has been stopped."));
  }
}

void cTranscript::stopAdvTranscript ()
{
  if (advrunning)
  {
    fputs ("\n\n", advfile);
/*
  Not supported yet
  
    if (allowansi)
    {
      //set output color to default (usually gray on black)
      char defcolor[5];
      defcolor[0] = 27;
      defcolor[1] = '[';
      defcolor[2] = '0';
      defcolor[3] = 'm';
      defcolor[4] = 0;
      fputs (defcolor, advfile);
    }
*/
    QString s = i18n ("Advanced session transcript has been stopped.");
    fputs (s.latin1(), advfile);
    fputs ("\n\n", advfile);
    advrunning = false;
    fclose (advfile);
    advfile = NULL;
    if(transtimer->isActive())
    {
      transtimer->stop();
      delete transtimer;
    }
    sess->console()->systemMessage (i18n ("Advanced session transcript has been stopped."));
  }
}

void cTranscript::configure ()
{
  //so first we have to create the dialog...
  tdlg = new dlgTranscript (kmuddy);

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (tdlg, SIGNAL (okClicked()), this, SLOT (applyTranscript ()));
  connect (tdlg, SIGNAL (applyClicked()), this, SLOT (applyTranscript ()));

  //next we fill in its data
  tdlg->setEnabled (running);
  tdlg->setAllowANSI (allowansi);

  cConnPrefs *cp = sess->connection()->getConnPrefs();
  if (cp)
  {
    advfname = cp->transcriptDir();
    fname = cp->transcriptDir();
  }
  else
  {
    fname = QDir::homeDirPath();
    advfname = QDir::homeDirPath();
  }
  
  fname += "/kmuddy_transcript.txt";

  tdlg->setFName (fname);
  tdlg->setOverwrite (overwrite);
  tdlg->setIncludeDump (includedump);
  tdlg->setAdvTranscript(advrunning);
  tdlg->setRotateDaily (rotatedaily);
  tdlg->setIncludeTimestamp(includetimestamp);  
  //dialog is ready - show it!
  tdlg->exec ();

  //finally, destroy that dialog!
  delete tdlg;
}

void cTranscript::startTranscript ()
{
  if (running)
    stopTranscript ();
  if (!(sess->telnet()->isConnected()))  //no transcript if we aren't connected
    return;
  errno = 0;
  
  file = fopen (fname.latin1(), overwrite ? "w" : "a");
  if (file == NULL)
  {
    KMessageBox::detailedSorry (kmuddy,
        i18n ("Transcript file could not be opened."), strerror (errno));
    errno = 0;
    sess->console()->systemMessage (i18n ("Session transcript could not be started."));
  }
  else
  {
    running = true;
    fputs ("\n\n", file);
    QString s = i18n ("Session transcript has just started.");
    fputs (s.latin1(), file);
    fputs ("\n\n", file);

    //add buffer dump if requested
    if (includedump)
      sess->console()->dumpBuffer (false, file);
        
    sess->console()->systemMessage (i18n ("Session transcript has been started."));

    //initialize FG/BG variables
    curfg = curbg = -1;
  }
}

void cTranscript::load()
{
  name = sess->connection()->getConnPrefs()->name();
  directory = locateLocal ("appdata", "profiles/" + name + "/");
  qDebug("load called "+ directory +" "+name);
  KConfig *config = new KConfig ( directory + "transcript", true);
  
  config->setGroup("Advanced transcript");
  setEnableRotate(config->readBoolEntry("Daily rotation", false));
  setPrependTimestamp(config->readBoolEntry("Prepend timestamp", false));
  
  delete config;
}

void cTranscript::save()
{
  qDebug("save called "+ directory +" "+name);
  KConfig *config = new KConfig ( directory + "transcript");
  
  config->setGroup("Advanced transcript");
  config->writeEntry("Daily rotation", rotatedaily);
  config->writeEntry("Prepend timestamp", includetimestamp);
  
  delete config;
}

void cTranscript::startAdvTranscript ()
{
  int day, month, year;
  QString sessname;
  
  if (advrunning)
    stopAdvTranscript ();
  if (!(sess->telnet()->isConnected()))  //no transcript if we aren't connected
    return;
  errno = 0;
 
  QDate date = QDate::currentDate();
  transtimer = new QTimer();
  connect (transtimer, SIGNAL (timeout ()), this, SLOT (timeout ()));

  day = date.day();
  month = date.month();
  year = date.year();
  
  if(sess)
    sessname = sess->connection()->getConnPrefs()->name();
  
  if(sessname)
    advfname.sprintf("/%s-%02d-%02d-%04d.log", 
		     sessname.latin1(), day, month, year);
  else
      advfname.sprintf("/unnamed-%02d-%02d-%04d.log", 
		       day, month, year);
  
  cConnPrefs *cp = sess->connection()->getConnPrefs();
  if (cp)
  {
    advfname = cp->transcriptDir() + advfname;
  }
  else
  {
    advfname = QDir::homeDirPath() + advfname;
  }

  advfile = fopen (advfname.latin1(), "a");
  if (advfile == NULL)
  {
    KMessageBox::detailedSorry (kmuddy,
        i18n ("Advanced transcript file could not be opened."), strerror (errno));
    errno = 0;
    sess->console()->systemMessage (i18n ("Advanced session transcript could not be started."));
  }
  else
  {
    advrunning = true;
    fputs ("\n\n", advfile);
    QString s = i18n ("Advanced session transcript has just started.");
    fputs (s.latin1(), advfile);
    fputs ("\n\n", advfile);

    sess->console()->systemMessage (i18n ("Advanced session transcript has been started."));
    
    if(rotatedaily)
    {
      sess->console()->systemMessage(i18n ("Session transcript will be rotated at midnight."));
      transtimer->start(60000);
    }

    //initialize FG/BG variables
    curfg = curbg = -1;
  }
}

void cTranscript::applyTranscript ()
//handler of OK/Apply in transcript dialog
{
  stopTranscript ();
  stopAdvTranscript();
  overwrite = tdlg->isOverwrite ();
  allowansi = tdlg->isAllowANSI ();
  includedump = tdlg->includeDump ();
  fname = tdlg->getFName ();  
  if (tdlg->isEnabled ())
    startTranscript ();  
  
  rotatedaily = tdlg->rotateDaily();
  includetimestamp = tdlg->includeTimestamp();
  if(tdlg->advTranscript())
    startAdvTranscript();
  
  save();
}

void cTranscript::dumpBuffer ()
{
  //so first we have to create the dialog...
  bdlg = new dlgDumpBuffer (kmuddy);

  QString fName;
  cConnPrefs *cp = sess->connection ()->getConnPrefs ();
  if (cp)
    fName = cp->transcriptDir ();
  else
    fName = QDir::homeDirPath ();
  fName += "/buffer_dump.txt";
  bdlg->setFileName (fName);

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (bdlg, SIGNAL (okClicked()), this, SLOT (doDumpBuffer ()));

  //dialog is ready - show it!
  bdlg->exec ();

  //finally, destroy that dialog!
  delete bdlg;
}

void cTranscript::doDumpBuffer ()
{
  QString fName = bdlg->fileName();
  FILE *f = fopen (fName.latin1(), "w");
  if (f)
    sess->console()->dumpBuffer (bdlg->curPos(), f);
  else
    KMessageBox::detailedSorry (kmuddy,
        i18n ("Dump file could not be opened."), strerror (errno));
  fclose (f);
}

void cTranscript::timeout()
{
  QTime time = QTime::currentTime();
  
  if((time.hour() == 0) && (time.minute() == 0))
  {
    stopAdvTranscript();
    startAdvTranscript();
  }
}

