/***************************************************************************
                          ctrigger.cpp  -  trigger
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : Ne okt 13 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@kmuddy.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ctrigger.h"

#ifdef HAVE_SOUND
#include "csoundplayer.h"
#endif

#include "cconnection.h"
#include "cconnprefs.h"
#include "cconsole.h"
#include "cgroup.h"
#include "ctextprocessor.h"
#include "ctriggerlist.h"
#include "kmuddy.h"

#include <kconfig.h>

extern KMuddy *kmuddy;

cTrigger::cTrigger (cSession *s)
{
  sess = s;
  
  text = "";

  //default compare type is substring
  type = substring;

  continueifmatch = false;
  continueifnomatch = true;
  dontsend = false;
  
  colortrigger = false;
  clearColorizations ();

  gagtrigger = false;
  notifytrigger = false;
  prompttrigger = false;
#ifdef HAVE_SOUND
	soundtrigger = false;
#endif
}

cTrigger::~cTrigger ()
{
  //nothing here
}

cSaveableField *cTrigger::newInstance ()
{
  return new cTrigger (sess);
}

bool cTrigger::match (const QString &what)
{
  bool detecting = sess->connection()->getConnPrefs()->triggerList()->
      detectingPrompt ();
  if (detecting != prompttrigger)
    return false;
  return cSaveableField::match (what);
}

//always returns QString::null
QString cTrigger::performAction ()
{
  commandstoexec.clear();

  //colorize the line if needed!
  if (colortrigger && ((!gagtrigger) || prompttrigger))
  {
    int len = lasttext.length();

    int *newfg = new int[len];
    int *newbg = new int[len];
    for (int i = 0; i < len; i++)
    {
      newfg[i] = -1;
      newbg[i] = -1;
    }
    prepareColorizationArrays (newfg, newbg, len);
    sess->textproc()->recolorize (newfg, newbg);
    delete[] newfg;
    delete[] newbg;
  }

  //prompt-detection triggers are special - no gagging, sounds,
  //notification or cmd sending...
  if (prompttrigger)
  {
    sess->textproc()->setHavePrompt ();
    return QString::null;
  }
  else
  {
    //gag the line if needed
    if (gagtrigger)
      sess->textproc()->gagLine ();

    //notify the user if needed
    if (notifytrigger)
      kmuddy->requestNotify(sess);

    if(outputwindowtrigger)
    {
      sess->textproc()->setOutputWindow(outputwindowname);
      if(gagoutputwindow)
        sess->textproc()->gagLineInPrimary ();
    }

#ifdef HAVE_SOUND
    //attempt to play the sound if needed
    if (soundtrigger && (!(soundFName.isEmpty())))
    {
      cSoundPlayer *player = kmuddy->soundPlayer();
      if (player)
      {
        player->setFileName (soundFName);
        player->setPriority (50);
        player->setRepeatsCount (1);
        player->setVolume (50);
        player->play ();
      }
    }
#endif

    int commands = 0;
    bool haveNonPlain = false;
    QStringList::iterator it;
    for (it = newtext.begin(); it != newtext.end(); ++it)
    {
      QString nt = *it;
      expandPseudoVariables (nt);
      commandstoexec.push_back (nt);
      commands++;
      if (! (nt.isEmpty())) haveNonPlain = true;
    }

    //empty to-sent text
    if ((commands == 1) && (!haveNonPlain))
    {
      //send nothing
      commandstoexec.clear ();
      return QString::null;
    }

    if (dontsend)        
    {
      for (it = commandstoexec.begin(); it != commandstoexec.end(); ++it)
      {
        QString tcols, bcols;
        QChar t = QChar (sess->textproc()->defaultTextColor());
        QChar b = QChar (sess->textproc()->defaultBkColor());
        int len = (*it).length();
        for (int i = 0; i < len; i++)
        {
          tcols += t;
          bcols += b;
        }
        sess->console()->addLine (*it, tcols, bcols);
      }
      
      commandstoexec.clear ();    //don't send anything!
      return QString::null;
    }
    return QString::null;
  }
}

bool cTrigger::shouldContinue (bool match)
{
  //should we continue?

  //if curent detection mode differs from our prompt-detect setting, then
  //we always continue, as matching was not done
  bool detecting = sess->connection()->getConnPrefs()->triggerList()->
      detectingPrompt ();
  if (detecting != prompttrigger)
    return true;
  
  if ((!nogroup) && (ourgroup != NULL) && (!ourgroup->active()))
    return true;    //if our group is not activated, then we should always
                    //continue (as if we hadn't existed)
  //if the group is active or doesn't exist, use trigger's settings
  return match ? continueifmatch : continueifnomatch;
}

void cTrigger::load (KConfig *config, const QString &group)
{
  config->setGroup (group);
  setText (config->readEntry ("Text", ""));

  //ntxt: for compatibility with KMuddy <= 0.6pre1
  QString ntxt = config->readEntry ("Replacement text", QString::null);

  newtext.clear();
  int repcount = config->readNumEntry ("Replacement count", -1);
  if (repcount == -1)   //use old-style replacement
    newtext.push_back (ntxt);
  else    //new-style replacement
    for (int i = 1; i <= repcount; i++)
    {
      QString repline = config->readEntry ("Replacement line " +
            QString::number (i), "");
      newtext.push_back (repline);
    }

  setType (config->readNumEntry ("Type", substring));
  continueifmatch = config->readBoolEntry ("Continue if match", false);
  continueifnomatch = config->readBoolEntry ("Continue if no match", true);
  _group = config->readNumEntry ("Group ID", 0);

  clearColorizations ();
  colortrigger = config->readBoolEntry ("Color trigger", false);
  if (colortrigger)
  {
    int cnum = config->readNumEntry ("Colorizations", 0);
    for (int i = 0; i < cnum; i++)
    {
      QString cname = config->readEntry ("Colorization variable " +
            QString::number(i + 1), "");
      int colors = config->readNumEntry ("Colorization colors " +
            QString::number(i + 1), 255 * 256 + 255);
      addColorization (cname, colors);
    }
  }

  gagtrigger = config->readBoolEntry ("Gag trigger", false);
  notifytrigger = config->readBoolEntry ("Notify trigger", false);
  prompttrigger = config->readBoolEntry ("Prompt trigger", false);
#ifdef HAVE_SOUND
  soundtrigger = config->readBoolEntry ("Sound trigger", false);
  soundFName = config->readEntry ("Sound file name", QString::null);
#endif

  outputwindowtrigger = config->readBoolEntry ("Send output to separate window", false);
  gagoutputwindow = config->readBoolEntry ("Gag main window in favour of output", false);
  outputwindowname = config->readEntry ("Output window name", QString::null);
}

void cTrigger::save (KConfig *config, const QString &group)
{
  config->setGroup (group);
  config->writeEntry ("Text", text);
  //remove obsolete entry
  config->deleteEntry ("Replacement text");

  int repcount = newtext.count();
  config->writeEntry ("Replacement count", repcount);
  QStringList::iterator it = newtext.begin();
  for (int i = 1; i <= repcount; i++)
  {
    config->writeEntry ("Replacement line " + QString::number (i), *it);
    it++;
  }

  config->writeEntry ("Type", type);
  config->writeEntry ("Continue if match", continueifmatch);
  config->writeEntry ("Continue if no match", continueifnomatch);
  config->writeEntry ("Group ID", _group);

  config->writeEntry ("Color trigger", colortrigger);
  if (colortrigger)
  {
    for (int i = 0; i < colorizationCount; i++)
    {
      config->writeEntry ("Colorization variable " +
            QString::number(i + 1), colorizationVar[i]);
      config->writeEntry ("Colorization colors " +
            QString::number(i + 1), colorization[i]);
    }
    config->writeEntry ("Colorizations", colorizationCount);
  }

  config->writeEntry ("Gag trigger", gagtrigger);
  config->writeEntry ("Notify trigger", notifytrigger);
  config->writeEntry ("Prompt trigger", prompttrigger);
#ifdef HAVE_SOUND
  config->writeEntry ("Sound trigger", soundtrigger);
  config->writeEntry ("Sound file name", soundFName);
#endif
  config->writeEntry ("Send output to separate window", outputwindowtrigger);
  config->writeEntry ("Gag main window in favour of output", gagoutputwindow);
  config->writeEntry ("Output window name", outputwindowname);
}

void cTrigger::clearColorizations ()
{
  colorizationCount = 0;
  for (int i = 0; i < MAX_COLORIZATIONS; i++)
  {
    colorizationVar[i] = QString::null;
    colorization[i] = 255 * 256 + 255;
  }
}

bool cTrigger::addColorization (QString pseudovar, int newcolors)
{
  if (colorizationCount == MAX_COLORIZATIONS)
    return false;   //NO MORE FREE SPACE!!!
  colorizationVar[colorizationCount] = pseudovar;
  colorization[colorizationCount] = newcolors;
  colorizationCount++;
  return true;
}

QString cTrigger::getColorizationVariable (int num)
{
  if ((num < 0) || (num >= MAX_COLORIZATIONS))
    return QString::null;
  return colorizationVar[num];
}

int cTrigger::getColorizationColor (int num)
{
  if ((num < 0) || (num >= MAX_COLORIZATIONS))
    return 255 * 256 + 255;
  return colorization[num];
}

void cTrigger::prepareColorizationArrays (int *newfg, int *newbg, int cols)
{
  //now go entry by entry :)
  for (int clr = 0; clr < colorizationCount; clr++)
  {
    QString varname = colorizationVar[clr];
    varname = varname.stripWhiteSpace();
    //varname must start in $
    if (varname[0] != '$')
      continue;
    varname = varname.mid (1);    
    int color = colorization[clr];
    int fgcol = color % 256;
    int bgcol = color / 256;
    if (fgcol == 255) fgcol = -1;
    if (bgcol == 255) bgcol = -1;

    int fromIndex, toIndex;

    //look if we know the variable name...
    bool ok;
    int number = varname.toInt (&ok);
    if (ok)   //it was a number
    {
      if (type == regexp)
      {
        if (number >= (int) backreflist.count())
          continue; //wrong number
        int len = backreflist[number].length();
        fromIndex = backrefpos[number];
        toIndex = fromIndex + len;
      }
      else
        continue;
    }
    else
    {
      if (varname == "line")
      {
        fromIndex = 0;
        toIndex = cols;
      }
      else
      if (varname == "prefix")
      {
        fromIndex = 0;
        toIndex = prefix.length() - 1;
      }
      else                                      
      if (varname == "suffix")                  
      {                                         
        toIndex = lasttext.length() - 1;
        fromIndex = toIndex - suffix.length() + 1;
      }
      else
      if (varname == "prefixtrim")
      {
        fromIndex = 0;
        toIndex = prefix.stripWhiteSpace().length() - 1;
      }
      else
      if (varname == "suffixtrim")
      {
        toIndex = lasttext.length() - 1;
        fromIndex = toIndex - suffix.stripWhiteSpace().length() + 1;
      }
      else
      if (varname == "matched")
      {
        fromIndex = lastpos;
        toIndex = fromIndex + lastlen;
      }
      else
        continue;
    }
    //if we are here, colorization can occur!
    for (int i = fromIndex; i < toIndex; i++)
    {
      newfg[i] = fgcol;
      newbg[i] = bgcol;
    }
  }
}
