/***************************************************************************
                          cvariablelist.cpp  -  list of variables
                             -------------------
    begin                : Po sep 8 2003
    copyright            : (C) 2003 by Tomas Mecir
    email                : kmuddy@kmuddy.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cvariablelist.h"

#include "cpluginmanager.h"
#include "cvariable.h"

#include <kconfig.h>

using namespace std;

//maximum number of passes when doing recursive variable expansion

#define MAXEXPANDPASSES 10

cVariableList::cVariableList (QString filename, cSession *s)
      : cSaveableList (s)
{
  prototype = new cVariable;
  objName = "Variable";
  config = new KConfig (filename);
  load ();
  initMapping ();
}

cVariableList::~cVariableList ()
{
  save ();
  config->sync ();
  vars.clear ();
}

void cVariableList::initMapping ()
{
  //traverse the list of variables, adding each of them to the mapping
  for (reset(); *this; (*this)++)
  {
    cVariable *var = (cVariable *) cur;
    //no duplicity check - assumes that the savefile is sane
    //consequence: don't mess up the variable savefile!
    vars[var->name()] = var;
  }
}

QString cVariableList::getValue (const QString &varname)
{
  QString vn = varname;
  if (varname[0] == '$')
    vn = vn.mid(1);

  //no such variable -> nothing to do
  if (!exists (vn))
    return QString::null;

  QString val = vars[vn]->value();
  //we mustn't return QString::null if the variable contains it, because
  //that would be treated as if the variable hadn't existed; return empty
  //string instead
  if (val == QString::null)
    return QString ("");
  else
    return val;
}

int cVariableList::getIntValue (const QString &varname)
{
  //retrieve value
  QString varvalue = getValue (varname);
  //convert it to number
  bool ok;
  int n = varvalue.toInt (&ok);
  //return that number if successfull, 0 otherwise
  return ok ? n : 0;
}

bool cVariableList::exists (const QString &varname)
{
  QString vn = varname;
  if (varname[0] == '$')
    vn = vn.mid(1);
  return (vars.count (vn) != 0);
}

void cVariableList::set (const QString &varname, const QString &value)
{
  QString vn = varname;
  if (varname[0] == '$')
    vn = vn.mid(1);

  QString oldval = QString::null;
  if (exists (vn))
  {
    oldval = vars[vn]->value();
    //variable already exists - we just update its value
    vars[vn]->setValue (value);
  }
  else
  {
    //good, add new variable to the list
    cVariable *var = new cVariable;
    var->setName (vn);
    var->setValue (value);
    addToEnd (var);
    vars[vn] = var;
  }

  if (oldval != value)
  {
    //inform plug-ins
    pluginmanager->event (sess, "var-changed", vn, value, oldval);
    //also inform others about the change... variable triggers will be connected here
    emit variableChanged (vn, value);
  }
}

void cVariableList::unset (const QString &varname)
{
  QString vn = varname;
  if (varname[0] == '$')
    vn = vn.mid(1);

  //no such variable -> nothing to do
  if (!exists (vn))
    return;

  //get the variable
  cVariable *var = vars[vn];
  //update the mapping
  vars.erase (vn);
  //update the list (also frees memory)
  cur = var;
  removeCurrent ();
}

void cVariableList::inc (const QString &varname, int delta)
{
  //delta must be positive
  if (delta <= 0) return;
  int n = getIntValue (varname);
  set (varname, QString::number (n + delta));
}

void cVariableList::dec (const QString &varname, int delta)
{
  //delta must be positive
  if (delta <= 0) return;
  int n = getIntValue (varname);
  set (varname, QString::number (n - delta));
}

void cVariableList::provideResource (const QString &varname)
{
  //returns a resource to available ones
  inc (varname, 1);
}

bool cVariableList::requestResource (const QString &varname)
{
  //is a resource available?
  if (getIntValue (varname) <= 0)
    //no :(
    return false;
  //yes, "allocate" one
  dec (varname, 1);
  return true;
}

QStringList cVariableList::getList ()
{
  map<QString, cVariable *>::iterator it;
  QStringList list;
  for (it = vars.begin(); it != vars.end(); ++it)
    list.push_back (it->first);
  return list;
}

QString cVariableList::expandVariables (const QString &string, bool recursive)
{
  return doExpandVariables (string, recursive ? MAXEXPANDPASSES : 1);
}

QString cVariableList::doExpandVariables (const QString &string, int recursionCounter)
{
  //almost identical with cSaveableField::expandPseudoVariables

  QString newstring = "";
  int len = string.length ();
  bool hadvar = false; //whether we had some variable - used to prevent
                       //unneeded recursion
  bool invar = false;
  bool inpar = false; //variable name is in parentheses '(' and ')'
  QString varname;
  for (int i = 0; i < len; i++)
  {
    QChar ch = string[i];

    //support things like $$$a correctly
    if (invar && (!inpar) && (ch == '$') && (varname.isEmpty()))
    {
      invar = false;
      newstring += ch;
    }

    if (!invar)
    {
      if (ch == '$')
      {
        invar = true;
        inpar = false;
        varname = "";
      }
      else
        newstring += ch;
    }
    else
    {
      if ((varname == "") && (ch == '(') && (!inpar))
        inpar = true;
      else
      if (!(ch.isLetterOrNumber ()))  //end of variable name
      {
        invar = false;

        if (inpar)
        {
          inpar = false;
          if (ch == ')')
          {
            QString newstr = processVariable (varname);
            if (newstr == QString::null)
              newstring += "$(" + varname + ")";
            else
            {
              newstring += newstr;
              hadvar = true;
            }
          }
          else
            newstring += "$(" + varname + ch;
        }
        else
        {
          QString newstr = processVariable (varname);
          if (newstr == QString::null)
            newstring += "$" + varname;
          else
          {
            newstring += newstr;
            hadvar = true;  
          }
          newstring += ch;
        }
      }
      else
        varname += ch;
    }
  }

  if (invar)    //if a variable ends the string
    if (inpar)
      newstring += "$(" + varname;  //no ending parenthesis - no variable
    else
    {
      QString newstr = processVariable (varname);
      if (newstr == QString::null)
        newstring += "$" + varname;
      else
      {
        newstring += newstr;
        hadvar = true;
      }
    }

  //another pass if needed, then return the result
  recursionCounter--;
  if (hadvar && (recursionCounter > 0))
    return doExpandVariables (newstring, recursionCounter);
  return newstring;
}

QString cVariableList::processVariable (const QString &varname)
{
  //pretty simple, isn't it? :)
  return getValue (varname);
}

