/***************************************************************************
                          dlggroups.cpp  -  manages a list of groups
                             -------------------
    begin                : St dec 5 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@host.sk
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "../dialogs/dlggroups.h"

#include "../dialogs/cnumberedchecklistitem.h"

#include "../cgroup.h"
#include "../cgrouplist.h"
#include "../kmuddy.h"

#include <qlabel.h>
#include <qlayout.h>
#include <qpushbutton.h>

#include <klineeditdlg.h>
#include <klistview.h>
#include <klocale.h>
#include <kmessagebox.h>

extern KMuddy *kmuddy;

dlgGroups::dlgGroups (cGroupList *groupList, const QString &caption,
    QWidget *parent, const char *name) :
    KDialogBase (parent, name, false, caption, 0)
{
  glist = groupList;
  if (glist != 0)
    connect (glist, SIGNAL (destroyed ()), this, SLOT (listDestroyed ()));
  
  //initial size
  setInitialSize (QSize (100, 300));

  //main widget
  QWidget *page = new QWidget (this);
  QVBoxLayout *layout = new QVBoxLayout (page);

  setMainWidget (page);

  setFocusPolicy (QWidget::StrongFocus);

  QLabel *label = new QLabel (i18n (caption), page);
  view = new KListView (page);
  view->addColumn (i18n ("Name"));
  view->setColumnWidthMode (0, QListView::Maximum);
  //unfortunately there is no signal that informs us about item state change
  connect (view, SIGNAL (clicked (QListViewItem *)), this, SLOT (itemClicked (QListViewItem *)));
  connect (view, SIGNAL (currentChanged (QListViewItem *)), this, SLOT (itemClicked (QListViewItem *)));
  connect (view, SIGNAL (selectionChanged (QListViewItem *)), this, SLOT (itemClicked (QListViewItem *)));
  connect (view, SIGNAL (spacePressed (QListViewItem *)), this, SLOT (itemClicked (QListViewItem *)));
  connect (view, SIGNAL (mouseButtonClicked (int, QListViewItem *, const QPoint &, int)), this, SLOT (mouseClick (int, QListViewItem *, const QPoint &, int)));

  //buttons
  //I create them on my own instead of relying on KDialogBase, because I
  //need them in vertical order at the bottom of the dialog, and this layout
  //is not supported by KDialogBase...
  QPushButton *add, *rename, *remove;
  add = new QPushButton (i18n ("&Add"), page);
  rename = new QPushButton (i18n ("&Rename"), page);
  remove = new QPushButton (i18n ("&Delete"), page);

  layout->setSpacing (5);
  layout->addWidget (label);
  layout->addWidget (view);
  layout->addWidget (add);
  layout->addWidget (rename);
  layout->addWidget (remove);
  
  //activate the buttons!
  connect (add, SIGNAL (clicked()), this, SLOT (addItem ()));
  connect (rename, SIGNAL (clicked()), this, SLOT (renameItem ()));
  connect (remove, SIGNAL (clicked()), this, SLOT (deleteItem ()));

  //now update the listview
  updateView ();
}

dlgGroups::~dlgGroups()
{
  //nothing here
}

void dlgGroups::switchGroupList (cGroupList *newlist)
{
  if (glist != 0)
    disconnect (glist, SIGNAL (destroyed ()), this, SLOT (listDestroyed ()));
  glist = newlist;
  if (glist != 0)
    connect (glist, SIGNAL (destroyed ()), this, SLOT (listDestroyed ()));
  updateView ();  
}

void dlgGroups::itemClicked (QListViewItem *item)
{
  if (item == 0)  //if he was clicked outside of items, there's nothing to do
    return;
  if (glist == 0) //if there is no grouplist, there's nothing to do
    return;

  cNumberedCheckListItem *i = (cNumberedCheckListItem *) item;
  //we don't know if the click has changed the checkbox, assuming it did...
  glist->activateGroup(i->number(), i->isOn());
}

void dlgGroups::addItem ()
{
  if (glist == 0) //if there is no grouplist, there's nothing to do
    return;

  QString text;
  bool ok;
  text = KLineEditDlg::getText (i18n ("Add new group"), i18n ("Enter group name:"), QString::null, &ok, this);
  if (ok)  //He pressed OK, add the group!!!
  {
    text = text.stripWhiteSpace ();
    if (text.length() == 0)  //do nothing if no group name
      return;
    glist->addGroup(text);
    updateView ();
  }  
}

void dlgGroups::renameItem ()
{
  cNumberedCheckListItem *item = (cNumberedCheckListItem *) view->selectedItem ();
  if (item == 0)  //do nothing if there is no selected item
    return;
  if (glist == 0) //if there is no grouplist, there's nothing to do
    return;

  //BASE group can never be renamed
  if (item->number() == 0) //BASE group
  {
    KMessageBox::sorry (this, i18n ("Sorry, but the BASE group can not be renamed."));
    return;
  }

  QString text;
  bool ok;
  text = KLineEditDlg::getText (i18n ("Rename group"), i18n ("Enter group name:"), item->text(), &ok, this);
  if (ok)  //He pressed OK, add the group!!!
  {
    text = text.stripWhiteSpace ();
    if (text.length() == 0)  //do nothing if no group name entered
      return;
    glist->renameGroup(item->number(), text);
    //update item's text... There's no need to call updateView()...
    item->setText (0, text);
  }
}

void dlgGroups::deleteItem ()
{
  cNumberedCheckListItem *item = (cNumberedCheckListItem *) view->selectedItem ();
  if (item == 0)  //do nothing if there is no selected item
    return;
  if (glist == 0) //if there is no grouplist, there's nothing to do
    return;

  //BASE group can never be deleted
  if (item->number() == 0) //BASE group
  {
    KMessageBox::sorry (this, i18n ("BASE group is the main group and can never be deleted!"));
    return;
  }
  
  if (KMessageBox::questionYesNoCancel (this, i18n ("Do you really want to "
        "delete this group?\n\nItems belonging to this group will be moved "
        "to the BASE group."), i18n ("Delete group")) == KMessageBox::Yes)
  {
    //OK, proceed to deleting!
    bool b = glist->removeGroup (item->number ());
    if (!b)
    {
      //report failure :(
      KMessageBox::sorry (this, i18n ("This group could not be deleted."));
      return;
    }
    updateView ();
  }
}

void dlgGroups::updateView ()
{
  view->clear ();

  if (glist == 0) //if there is no grouplist, there's nothing more do
    return;

  view->setSelectionMode (QListView::Single);
  for (glist->reset(); *glist; (*glist)++)
  {
    cGroup *g = (cGroup *) **glist;
    cNumberedCheckListItem *item = new cNumberedCheckListItem (view, g->getText(), QCheckListItem::CheckBox);
    item->setOn (g->active());
    item->setNumber (g->number()); //set group number
  }
}

void dlgGroups::mouseClick (int, QListViewItem *i, const QPoint &, int)
{
  itemClicked (i);
}

void dlgGroups::listDestroyed ()
{
  glist = 0;
  updateView ();
}
