/***************************************************************************
                          kmuddy.cpp  -  main class that handles interface
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : P Jun 14 12:37:51 CEST 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@kmuddy.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "kmuddy.h"

#include "cactionlist.h"
#include "ccmdprocessor.h"
#include "cconnection.h"
#include "cconnectioninfo.h"
#include "cconsole.h"
#include "cgrouplist.h"
#include "cinputline.h"
#include "cpluginmanager.h"
#include "cprofiles.h"
#include "csendcmd.h"
#include "csession.h"
#include "cshortcutlist.h"
#include "ctranscript.h"
#include "ctelnet.h"
#include "ctabbar.h"

#include "cglobalsettings.h"
#include "cmultilineinput.h"
#include "csoundplayer.h"

#include "dialogs/dlgconnect.h"
#include "dialogs/dlgquickconnect.h"
#include "dialogs/dlgeditprofile.h"
#include "dialogs/dlggrabkey.h"
#include "dialogs/dlggroups.h"
#include "dialogs/dlgimportexportprofile.h"
#include "dialogs/dlgrunninglist.h"
#include "dialogs/dlgstatistics.h"

#include <config.h>

#include <stdlib.h>
#include <ctime>

#ifdef HAVE_SDL
#include <SDL/SDL.h>
#include <SDL/SDL_mixer.h>
#endif

#include <qdesktopwidget.h>
#include <qdockarea.h>
#include <qlayout.h>
#include <qtabbar.h>
#include <qtimer.h>
#include <qwidgetstack.h>
#include <qwidget.h>
#include <qpopupmenu.h>
#include <qevent.h>

#include <kaboutapplication.h>
#include <kaction.h>
#include <kapp.h>
#include <khelpmenu.h>
#include <kiconloader.h>
#include <klineeditdlg.h>
#include <klocale.h>
#include <kmenubar.h>
#include <kmessagebox.h>
#include <kpopupmenu.h>
#include <kshortcut.h>
#include <kstatusbar.h>
#include <ktoolbar.h>
#include <kdebug.h>

//all the objects; everybody who needs them should declare them as externs
cMultiLineInput *multiline;
cGlobalSettings *globalsettings;
cPluginManager *pluginmanager;

#define ID_CONNECTION 1000
#define ID_EDIT 1001
#define ID_VIEW 1002
#define ID_PROFILE 1003
#define ID_TOOLS 1004
#define ID_SETTINGS 1005
#define ID_HELP 1006

//are the actions ready?
bool actionsReady;

bool KMuddy::goingDown = false;

KMuddy::KMuddy(QWidget *parent, const char *name) : KMainWindow(parent, name)
{
  //initialize random number generator (needed by cMSP)
  srand (time (0));

  count = 0;
  for (int i = 0; i < MAXSESSIONS; i++)
    session[i] = 0;
  wantToolbar = false;
  wantTabBar = true;
  tabbar = 0;
  actionsReady = false;
  globalnotify = localnotify = alwaysnotify = false;
  globalFlashingNOW = false;
  notifyTimer = new QTimer (this);
  connect (notifyTimer, SIGNAL (timeout ()), this, SLOT (handleFlashing ()));
  agroups = 0;
  tgroups = 0;
  rdlg = 0;
  statdlg = 0;
  grabdlg = 0;

#ifdef HAVE_SDL
  bool failedInit = false;
  if (SDL_Init (SDL_INIT_AUDIO) < 0)
  {
    failedInit = true;
    qDebug ("Could not initialize SDL library!");
  }
  if (!failedInit)
    if (Mix_OpenAudio (44100, AUDIO_S16, 2, 4096) < 0)
    {
      failedInit = true;
      qDebug ("Couldn't open SDL_Mixer!");
    }
#endif

#ifdef HAVE_SOUND
  splayer = new cSoundPlayer (true);
  mplayer = new cSoundPlayer (false);
#else
  splayer = 0;
  mplayer = 0;
#endif

#ifdef HAVE_SDL
  if (failedInit)
  {
    splayer->disableSound ();
    mplayer->disableSound ();
  }
#endif
}

KMuddy::~KMuddy()
{
  goingDown = true;
  killObjects ();
#ifdef HAVE_SDL
  Mix_CloseAudio ();
  SDL_Quit ();
#endif
}

void KMuddy::prepareObjects ()
{
  //only run this once!!!
  static bool ready = false;
  if (ready)
    return;
  ready = true;

  //plug-in manager
  pluginmanager = new cPluginManager;

  globalsettings = new cGlobalSettings;
  multiline = new cMultiLineInput;

  atoolbar = new KToolBar (this, QMainWindow::Top);
  atoolbar->setCaption ("Action toolbar");
  atoolbar->hide ();

  //Let's create the central widget and friends first!

  QFrame *central = new QFrame (this);
  QVBoxLayout *layout = new QVBoxLayout (central);

  //widget stack
  stack = new QWidgetStack (central);
  central->setMinimumSize (400, 300);

  //icons for tabs
  KIconLoader *loader = (KApplication::kApplication())->iconLoader();
  QPixmap i1 = loader->loadIcon ("ok", KIcon::Small, 16);
  QPixmap i2 = loader->loadIcon ("cancel", KIcon::Small, 16);
  QPixmap i3 = loader->loadIcon ("flag", KIcon::Small, 16);
  iconOk.setPixmap (i1, QIconSet::Automatic);
  iconNo.setPixmap (i2, QIconSet::Automatic);
  iconFlag.setPixmap (i3, QIconSet::Automatic);
  
  //tab bar
//  tabbar = new QTabBar (central);
  tabbar = new cTabBar (central);
  connect (tabbar, SIGNAL (selected (int)), this, SLOT (changeSession (int)));
  connect (tabbar, SIGNAL (closeTab()), this, SLOT (closeTab()));
  connect (tabbar, SIGNAL (reConnect()), this, SLOT (doReconnect()));
  connect (tabbar, SIGNAL (disConnect()), this, SLOT (doDisconnect()));

  //prepare the new session
  addSession ();

  //this addWidget stuff MUST go AFTER the addSession() command!!!
  //otherwise the tabbar is not shown (it contains no tabs at the
  //time of adding to the layout)
  layout->addWidget (tabbar, 1);
  layout->addWidget (stack, 1);

  //set up the window
  setCentralWidget (central);

  //--- establish connections between objects ---
  //---  most of stuff has moved to cSession  ---

  //multi-line input
  connect (multiline, SIGNAL (oneLine (const QString &)), this, SLOT (sendCommandToActiveSession (const QString &)));
  connect (multiline, SIGNAL (dialogClosed ()), this, SLOT (uncheckMultiLine ()));

  KApplication::kApplication()->processEvents ();

  //now we create the menubar

  //get action collection
  KActionCollection *acol = globalsettings->getACol ();

  //Connection
  stdconnect = new KAction (i18n ("&Connect..."), "connect_no", CTRL+Key_N, this,
      SLOT(showAndHandleConnectDialog()), acol, "Connect");
  qconnect = new KAction (i18n ("&QuickConnect..."), "connect_creating", CTRL+Key_O, this,
      SLOT (showAndHandleQuickConnectDialog ()), acol, "QuickConnect");
  disconnect = new KAction (i18n ("&Disconnect"), "cancel", CTRL+Key_D, this,
      SLOT (doDisconnect ()), acol, "Disconnect");
  reconnect = new KAction (i18n ("&Reconnect"), CTRL+Key_R, this,
      SLOT (doReconnect ()), acol, "Reconnect");
  closetab = new KAction (i18n ("Close &tab"), KShortcut::null(), this,
      SLOT (closeTab ()), acol, "CloseTab");
  closetab->setEnabled (false);
  reconnect->setEnabled (false);

  //Edit
  clipcopy = new KAction (i18n ("&Copy to clipboard"), "editcopy", KShortcut::null(), this,
      SLOT (addSelectionToClipboard ()), acol, "ClipboardCopy");
  parsing = new KToggleAction (i18n ("Enable command &parsing"), "goto", CTRL+SHIFT+Key_P,
      acol, "EnableParsing");
  connect (parsing, SIGNAL (toggled (bool)), this, SLOT (setParsing (bool)));
  parsing->setChecked (true);  //enabled by default
  cancelpending = new KAction (i18n ("Ca&ncel pending commands"), "remove", KShortcut::null(),
      this, SLOT (clearCommandQueue ()), acol, "CancelPending");

  //View
  showmultiline = new KToggleAction (i18n ("Show &multi-line input"), "leftjust", CTRL+Key_M,
      multiline, SLOT (toggle ()), acol, "ShowMultiLine");
  showatoolbar = new KToggleAction (i18n ("Show a&ction toolbar"), KShortcut::null(),
      acol, "ShowAToolbar");
  connect (showatoolbar, SIGNAL (toggled (bool)), this, SLOT (showAToolbar (bool)));
  connect (showatoolbar, SIGNAL (toggled (bool)), this, SLOT (setSettingsDirty()));
  showatoolbar->setChecked (false);
  connect (atoolbar, SIGNAL (visibilityChanged (bool)), showatoolbar, SLOT (setChecked (bool)));
  showaliasgroups = new KToggleAction (i18n ("Show &alias groups"), KShortcut::null(),
      acol, "ShowAliasGroups");
  connect (showaliasgroups, SIGNAL (toggled (bool)), this, SLOT (showAliasGroups (bool)));
  showaliasgroups->setChecked (false);
  showtriggergroups = new KToggleAction (i18n ("Show &trigger groups"), KShortcut::null(),
      acol, "ShowTriggerGroups");
  connect (showtriggergroups, SIGNAL (toggled (bool)), this, SLOT (showTriggerGroups (bool)));
  showtriggergroups->setChecked (false);
  showrunningscripts = new KToggleAction (i18n ("Show &running scripts"), KShortcut::null(),
      acol, "ShowRunningScripts");
  connect (showrunningscripts, SIGNAL (toggled (bool)), this, SLOT (showRunningScripts (bool)));
  showrunningscripts->setChecked (false);
  showstatistics = new KToggleAction (i18n ("Show connection &statistics"), KShortcut::null(),
      acol, "ShowStatistics");
  connect (showstatistics, SIGNAL (toggled (bool)), this, SLOT (showStatistics (bool)));
  showstatistics->setChecked (false);

  //Profile
  aliases = new KAction (i18n ("&Aliases..."), CTRL+ALT+Key_A, this,
      SLOT (handleAliasesDialog()), acol, "Aliases");
  triggers = new KAction (i18n ("&Triggers..."), CTRL+ALT+Key_T, this,
      SLOT (handleTriggersDialog()), acol, "Triggers");
  actions = new KAction (i18n ("A&ctions..."), KShortcut::null(), this,
      SLOT (handleActionsDialog()), acol, "Actions");
  scripts = new KAction (i18n ("&Scripts..."), KShortcut::null(), this,
      SLOT (handleScriptsDialog()), acol, "Scripts");
  timers = new KAction (i18n ("Time&rs..."), KShortcut::null(), this,
      SLOT (handleTimersDialog()), acol, "Timers");
  shortcuts = new KAction (i18n ("&Macro keys..."), KShortcut::null(), this,
      SLOT (handleShortcutsDialog()), acol, "MacroKeys");
  variables = new KAction (i18n ("&Variables..."), KShortcut::null(), this,
      SLOT (handleVariablesDialog()), acol, "Variables");
  vartrigs = new KAction (i18n ("Varia&ble triggers..."), KShortcut::null(), this,
      SLOT (handleVarTrigsDialog()), acol, "VarTrigs");
  windows = new KAction (i18n ("&Output windows..."), KShortcut::null(), this,
      SLOT (handleWindowsDialog()), acol, "OutputWindows");
  switchaliases = new KToggleAction (i18n ("A&liases enabled"), "aliases", CTRL+SHIFT+Key_A,
      acol, "EnableAliases");
  switchtriggers = new KToggleAction (i18n ("Tri&ggers enabled"), "text", CTRL+SHIFT+Key_T,
      acol, "EnableTriggers");
  switchtimers = new KToggleAction (i18n ("Tim&ers enabled"), KShortcut::null(),
      acol, "EnableTimers");
  switchshortcuts = new KToggleAction (i18n ("Macro &keys enabled"), KShortcut::null(),
      acol, "EnableMacroKeys");
  connect (switchaliases, SIGNAL (toggled (bool)), this, SLOT (switchAliases (bool)));
  connect (switchtriggers, SIGNAL (toggled (bool)), this, SLOT (switchTriggers (bool)));
  connect (switchtimers, SIGNAL (toggled (bool)), this, SLOT (switchTimers (bool)));
  connect (switchshortcuts, SIGNAL (toggled (bool)), this, SLOT (switchShortcuts (bool)));
  switchaliases->setChecked (true);
  switchtriggers->setChecked (true);
  switchtimers->setChecked (true);
  switchshortcuts->setChecked (true);
  connprefs = new KAction (i18n ("&MUD Preferences..."), "identity", KShortcut::null(), this,
      SLOT (showConnPrefsDialog()), acol, "ConnPrefs");

  //Tools
  sesstranscript = new KAction (i18n ("&Session transcript..."), "toggle_log", KShortcut::null(),
      this, SLOT (configureTranscript()), acol, "Transcript");
  dumpbuffer = new KAction (i18n ("Dump output &buffer..."), KShortcut::null(),
      this, SLOT (dumpBuffer()), acol, "DumpBuffer");
  decide = new KAction (i18n ("&Decision assistant"), KShortcut::null(),
      this, SLOT (makeDecision()), acol, "Decision");
  importprofile = new KAction (i18n ("&Import profile..."),
      KShortcut::null(), this, SLOT (importProfile()), acol, "ImportProf");
  exportprofile = new KAction (i18n ("&Export profile..."),
      KShortcut::null(), this, SLOT (exportProfile()), acol, "ExportProf");

  //Settings
  appconfig = new KAction (i18n ("&Global settings..."), "configure", KShortcut::null(), globalsettings,
      SLOT(showSettingsDialog()), acol, "Global settings");

  //Help

  //other actions
  lineup = new KAction (i18n ("Shift line up"), SHIFT+Key_Up, this,
      SLOT (lineUp ()), acol, "LineUp");
  linedown = new KAction (i18n ("Shift line down"), SHIFT+Key_Down, this,
      SLOT (lineDown ()), acol, "LineDown");
  pageup = new KAction (i18n ("Shift page up"), SHIFT+Key_Prior, this,
      SLOT (pageUp ()), acol, "PageUp");
  pagedown = new KAction (i18n ("Shift page down"), SHIFT+Key_Next, this,
      SLOT (pageDown ()), acol, "PageDown");
  aconup = new KAction (i18n ("Split-screen up"), CTRL+ALT+Key_Up, this,
      SLOT (aconUp ()), acol, "AconUp");
  acondown = new KAction (i18n ("Split-screen down"), CTRL+ALT+Key_Down, this,
      SLOT (aconDown ()), acol, "AconDown");
  prevtab = new KAction (i18n ("Previous tab"), ALT+Key_Prior, this,
      SLOT (prevTab ()), acol, "PreviousTab");
  nexttab = new KAction (i18n ("Next tab"), ALT+Key_Next, this,
      SLOT (nextTab ()), acol, "NextTab");
  tabnum[0] = new KAction (i18n ("Switch to tab 1"), ALT+Key_1, this,
      SLOT (switchTab1 ()), acol, "SwitchTab1");
  tabnum[1] = new KAction (i18n ("Switch to tab 2"), ALT+Key_2, this,
      SLOT (switchTab2 ()), acol, "SwitchTab2");
  tabnum[2] = new KAction (i18n ("Switch to tab 3"), ALT+Key_3, this,
      SLOT (switchTab3 ()), acol, "SwitchTab3");
  tabnum[3] = new KAction (i18n ("Switch to tab 4"), ALT+Key_4, this,
      SLOT (switchTab4 ()), acol, "SwitchTab4");
  tabnum[4] = new KAction (i18n ("Switch to tab 5"), ALT+Key_5, this,
      SLOT (switchTab5 ()), acol, "SwitchTab5");
  tabnum[5] = new KAction (i18n ("Switch to tab 6"), ALT+Key_6, this,
      SLOT (switchTab6 ()), acol, "SwitchTab6");
  tabnum[6] = new KAction (i18n ("Switch to tab 7"), ALT+Key_7, this,
      SLOT (switchTab7 ()), acol, "SwitchTab7");
  tabnum[7] = new KAction (i18n ("Switch to tab 8"), ALT+Key_8, this,
      SLOT (switchTab8 ()), acol, "SwitchTab8");
  tabnum[8] = new KAction (i18n ("Switch to tab 9"), ALT+Key_9, this,
      SLOT (switchTab9 ()), acol, "SwitchTab9");
  tabnum[9] = new KAction (i18n ("Switch to tab 10"), ALT+Key_0, this,
      SLOT (switchTab10 ()), acol, "SwitchTab10");
  
  actionsReady = true;
  
  //create all the menus

  //menu Connection
  connectionMenu = new QPopupMenu (this);
  stdconnect->plug (connectionMenu);
  qconnect->plug (connectionMenu);
  disconnect->plug (connectionMenu);
  connectionMenu->insertSeparator ();
  reconnect->plug (connectionMenu);
  closetab->plug (connectionMenu);
  connectionMenu->insertSeparator ();
  KStdAction::quit(this, SLOT(quit()), 0)->plug (connectionMenu);

  //menu Edit
  editMenu = new QPopupMenu (this);
  clipcopy->plug (editMenu);
  editMenu->insertSeparator ();
  parsing->plug (editMenu);
  editMenu->insertSeparator ();
  cancelpending->plug (editMenu);

  //menu View
  viewMenu = new QPopupMenu (this);
  showmultiline->plug (viewMenu);
  showatoolbar->plug (viewMenu);
  showaliasgroups->plug (viewMenu);
  showtriggergroups->plug (viewMenu);
  showrunningscripts->plug (viewMenu);
  showstatistics->plug (viewMenu);

  //menu Profile
  profileMenu = new QPopupMenu (this);
  aliases->plug (profileMenu);
  triggers->plug (profileMenu);
  actions->plug (profileMenu);
  scripts->plug (profileMenu);
  timers->plug (profileMenu);
  shortcuts->plug (profileMenu);
  variables->plug (profileMenu);
  vartrigs->plug (profileMenu);
  windows->plug (profileMenu);
  profileMenu->insertSeparator ();
  connprefs->plug (profileMenu);
  profileMenu->insertSeparator ();
  switchaliases->plug (profileMenu);
  switchtriggers->plug (profileMenu);
  switchtimers->plug (profileMenu);
  switchshortcuts->plug (profileMenu);

  //menu Tools
  toolsMenu = new QPopupMenu (this);
  sesstranscript->plug (toolsMenu);
  dumpbuffer->plug (toolsMenu);
  toolsMenu->insertSeparator ();
  decide->plug (toolsMenu);
  toolsMenu->insertSeparator ();
  importprofile->plug (toolsMenu);
  exportprofile->plug (toolsMenu);

  //menu Settings
  settingsMenu = new QPopupMenu (this);
  appconfig->plug (settingsMenu);

  //menu Help
  helpMenu = new KHelpMenu (this, QString::null, false);
      //make About Kmuddy work (I have no idea why it doesn't work immediately)
  connect (helpMenu, SIGNAL (showAboutApplication ()), this, SLOT (aboutApp ()));

  //and the main menubar
  menuBar()->insertItem (i18n ("&Connection"), connectionMenu, ID_CONNECTION);
  menuBar()->insertItem (i18n ("&Edit"), editMenu, ID_EDIT);
  menuBar()->insertItem (i18n ("&View"), viewMenu, ID_VIEW);
  menuBar()->insertItem (i18n ("&Profile"), profileMenu, ID_PROFILE);
  menuBar()->insertItem (i18n ("&Tools"), toolsMenu, ID_TOOLS);
  menuBar()->insertItem (i18n ("&Settings"), settingsMenu, ID_SETTINGS);
  menuBar()->insertItem (i18n ("&Help"), helpMenu->menu(), ID_HELP);

  //read saved keys for acol object
  acol->readShortcutSettings ("Shortcuts");

  //we're not connected on startup, so let's disable some options...
  disableConnectedOptions ();
  disableAdvancedOptions ();

  //apply global settings
  globalsettings->apply ();

  //look if there's some auto-connect profile set
  if (autoconnect.length() > 0)
    doProfileConnect (autoconnect, false);

  //let the window remember its position!
  setAutoSaveSettings ();
}

void KMuddy::killObjects ()
{                 
  //destroy objects and set pointers to NULL - needed so that we can prevent
  //some crashes if some object calls a function of an object that has already
  //been destroyed; this gives him a chance to verify its state...

  //menubar and KActions are not deleted...

  //session objects are NOT deleted here - doing so causes kmuddy to crash

#ifdef HAVE_SOUND
  delete splayer;
  splayer = 0;
  delete mplayer;
  mplayer = 0;
#endif
  delete multiline;
  multiline = 0;
  delete globalsettings;
  globalsettings = 0;

  delete pluginmanager;
  pluginmanager = 0;

  //also delete some dialogs...
  if (agroups != 0)
    delete agroups;
  agroups = 0;
  if (tgroups != 0)
    delete tgroups;
  tgroups = 0;
  if (rdlg != 0)
    delete rdlg;
  rdlg = 0;
  if (statdlg != 0)
    delete statdlg;
  statdlg = 0;
}

void KMuddy::saveProperties (KConfig */*config*/)
{
  //nothing here for now, maybe I'll include re-connecting later?

}

void KMuddy::readProperties (KConfig */*config*/)
{
  //nothing here now now, maybe I'll include re-connecting later?

}

void KMuddy::quit ()
{
  for ( int i = 0; i < MAXSESSIONS; i++ ) {
    if ( session[i] ) {
      if ( !session[i]->connection()->handleDisconnect() ) {
        return;
      }
    }
  }

  kapp->quit();
}

void KMuddy::disableConnectedOptions ()
//may contain further commands...
{
  disconnect->setEnabled (false);
  sesstranscript->setEnabled (false);
  dumpbuffer->setEnabled (false);
  
  disableAdvancedOptions ();
}

void KMuddy::enableConnectedOptions ()
//may contain further commands...
{
  if (!(disconnect->isEnabled()))
    disconnect->setEnabled (true);
  if (!(sesstranscript->isEnabled()))
    sesstranscript->setEnabled (true);
  if (!(dumpbuffer->isEnabled()))
    dumpbuffer->setEnabled (true);
  reconnect->setEnabled (false);
}

void KMuddy::disableAdvancedOptions ()
//may contain further commands...
{
  doShowAToolbar (false);
  showatoolbar->setEnabled (false);
  menuBar()->setItemEnabled (ID_PROFILE, false);

  //!!! must also disable all profile-related actions!!! (otherwise
  //they're still accessible with hotkeys)
  aliases->setEnabled (false);
  triggers->setEnabled (false);
  actions->setEnabled (false);
  scripts->setEnabled (false);
  timers->setEnabled (false);
  shortcuts->setEnabled (false);
  variables->setEnabled (false);
  vartrigs->setEnabled (false);
  windows->setEnabled (false);
  switchaliases->setEnabled (false);
  switchtriggers->setEnabled (false);
  switchtimers->setEnabled (false);
  switchshortcuts->setEnabled (false);
  connprefs->setEnabled (false);
}

void KMuddy::enableAdvancedOptions ()
//may contain further commands...
{
  showatoolbar->setEnabled (true);
  menuBar()->setItemEnabled (ID_PROFILE, true);

  //show the toolbar if we want it
  if (wantToolbar)
  {
    doShowAToolbar (true);
    cConnPrefs *cp = getSession (activeSession())->connection()->getConnPrefs();
    if (cp != 0)
    {
      cActionList *al = cp->actionList();
      if (al != 0)
        al->rebuildToolbar();
    }
  }

  aliases->setEnabled (true);
  triggers->setEnabled (true);
  actions->setEnabled (true);
  scripts->setEnabled (true);
  timers->setEnabled (true);
  shortcuts->setEnabled (true);
  variables->setEnabled (true);
  vartrigs->setEnabled (true);
  windows->setEnabled (true);
  switchaliases->setEnabled (true);
  switchtriggers->setEnabled (true);
  switchtimers->setEnabled (true);
  switchshortcuts->setEnabled (true);
  connprefs->setEnabled (true);
}

void KMuddy::setClosedConn (bool canCT)
{
  closetab->setEnabled (canCT);
  reconnect->setEnabled (true);
}

void KMuddy::unsetClosedConn ()
{
  closetab->setEnabled (false);
  reconnect->setEnabled (false);
}

void KMuddy::closeTab ()
{
  //close that session...
	if(getSession(activeSession())->connection()->isConnected())
	  doDisconnect();
	if(!(getSession(activeSession())->connection()->isConnected()))
	  removeSession (activeSession(), false);
}

void KMuddy::doShowAToolbar (bool val)
//show/hide action toolbar
//does not modify wantToolbar variable, so that toolbar can be
//shown/hidden as needed
{
  bool shown = atoolbar->isVisible ();
  if (val)
    atoolbar->show ();
  else
    atoolbar->hide ();

  if (val != shown)
    //showing/hiding toolbar shifts view - fix it!
    //!!! THIS IS FAR FROM BEING IDEAL !!!
    for (int i = 0; i < MAXSESSIONS; i++)
      if (session[i] != NULL)
        session[i]->console()->pageDown ();
}

void KMuddy::showAToolbar (bool val)
//Remember this value then call doShowAToolbar
//This is a SLOT. Do NOT call this one. Use doShowAToolbar instead.
{
  if (val)
    wantToolbar = true;
  else
    wantToolbar = false;
  
  doShowAToolbar (val);

  //keep action in sync (this function can be explicitly called)
  showatoolbar->setChecked (val);
}

void KMuddy::showAliasGroups (bool val)
{
  cConnPrefs *cp = getSession (activeSession())->connection()->getConnPrefs();
  cGroupList *list = 0;
  if (cp != 0)
    list = cp->aliasGroupList();

  if (agroups == 0)
  {
    agroups = new dlgGroups (list, i18n ("Alias groups"), this);
    connect (agroups, SIGNAL (finished()), this, SLOT (aliasGroupsFinished ()));
  }
  if (val)
    agroups->show ();
  else
    agroups->hide ();

  //keep action in sync (this function can be explicitly called)
  showaliasgroups->setChecked (val);
}

void KMuddy::showTriggerGroups (bool val)
{
  cConnPrefs *cp = getSession (activeSession())->connection()->getConnPrefs();
  cGroupList *list = 0;
  if (cp != 0)
    list = cp->triggerGroupList();

  if (tgroups == 0)
  {
    tgroups = new dlgGroups (list, i18n ("Trigger groups"), this);
    connect (tgroups, SIGNAL (finished()), this, SLOT (triggerGroupsFinished ()));
  }
  if (val)
    tgroups->show ();
  else
    tgroups->hide ();

  //keep action in sync (this function can be explicitly called)
  showtriggergroups->setChecked (val);
}

void KMuddy::showRunningScripts (bool val)
{
  cConnPrefs *cp = getSession (activeSession())->connection()->getConnPrefs();
  cRunningList *list = 0;
  if (cp != 0)
    list = cp->runningList();

  if (rdlg == 0)
  {
    rdlg = new dlgRunningList (list, this);
    connect (rdlg, SIGNAL (finished()), this, SLOT (runningScriptsFinished ()));
  }
  if (val)
    rdlg->show ();
  else
    rdlg->hide ();

  //keep action in sync (this function can be explicitly called)
  showrunningscripts->setChecked (val);
}

void KMuddy::showStatistics (bool val)
{
  if (statdlg == 0)
  {
    statdlg = new dlgStatistics (this);
    connect (statdlg, SIGNAL (finished()), this, SLOT (statisticsFinished ()));
  }
  if (val)
    statdlg->show ();
  else
    statdlg->hide ();

  //keep action in sync (this function can be explicitly called)
  showstatistics->setChecked (val);
}

void KMuddy::aliasGroupsFinished ()
{
  showAliasGroups (false);
}

void KMuddy::triggerGroupsFinished ()
{
  showTriggerGroups (false);
}

void KMuddy::runningScriptsFinished ()
{
  showRunningScripts (false);
}

void KMuddy::statisticsFinished ()
{
  showStatistics (false);
}

void KMuddy::uncheckMultiLine ()
{
  showmultiline->setChecked (false);
}

void KMuddy::aboutApp ()
{
  KAboutApplication *dlg = new KAboutApplication (this);
  dlg->exec ();
  delete dlg;
}

//session management functions start here:

int KMuddy::sessCount ()
{
  return count;
}

cSession *KMuddy::getSession (int which)
{
  if ((which >= 0) && (which < MAXSESSIONS))
    return session[which];
  else
    return 0;
}

int KMuddy::getSessionNumber (cSession *sess)
{
  for (int i = 0; i < MAXSESSIONS; i++)
    if (session[i] == sess)
      return i;
  return -1;
}

cSession *KMuddy::addSession ()
{
  if (goingDown)
    return 0;
    
  if (count == MAXSESSIONS)
    return 0;

  //first, one very special case - if we only have one session
  //and that one is not connected, we return that session
  //and don't add any new one. This session will be used for
  //the connection that is being established.
  if (count == 1)
  {
    cSession *sess = 0;
    
    int which;
    for (int i = 0; i < MAXSESSIONS; i++)
      if (session[i] != 0)
      {
        sess = session[i];
        which = i;
        break;
      }
    if (!(sess->telnet()->isConnected()))
    {
      if (actionsReady)
        globalsettings->apply ();
      return sess;
    }
    //else: continue
  }

  //now for the standard operation

  //find a first free place for the connection
  int which = -1;
  for (int i = 0; i < MAXSESSIONS; i++)
    if (session[i] == 0)
    {
      which = i;
      break;
    }
  if (which == -1)  //this means a BUG in program, hope it'll never happen
    return 0;
  session[which] = new cSession (stack);
  stack->addWidget (session[which], which);
  count++;

  //update tab bar
  QTab *tab = new QTab (iconNo, i18n ("No connection"));
  tabID[which] = tabbar->addTab (tab);
  tabbar->update ();
  setSession (which);
  //just to be sure...
  activesess = which;

  (KApplication::kApplication())->processEvents();
  
  //apply global settings to the newly created session
  //this also shows/hides the tabbar if needed
  if (actionsReady)
    globalsettings->apply ();
  
  return session[which];
}

bool KMuddy::removeSession (int which, bool dontClose)
{
  if (goingDown)
    return false;

  //reconnect shortcut
  QString reconnectText = i18n (" To reconnect, press");
  reconnectText += " " + reconnect->shortcut().toString() + " ";
  reconnectText += i18n ("or use Connection / Reconnect.");

  if ((which < 0) || (which >= MAXSESSIONS))
    return false;
  if (session[which] == 0)
    return false;

  //now the special case - only one session; we do almost nothing!
  if (count == 1)
  {
    int id, _which;
    for (int i = 0; i < MAXSESSIONS; i++)
      if (session[i] != 0)
      {
        _which = i;
        id = tabID[i];
        break;
      }
    setSessionName (which, i18n ("No connection"));
    //no connection...
    tabbar->tab (tabID[which])->setIconSet (iconNo);
    tabbar->update ();
    session[which]->telnet()->offLineMode();
    if (session[which]->connection()->getConnectionInfo())
    {
      session[which]->connection()->setConnectionClosed (false);
      session[which]->connection()->updateMenus ();
      session[which]->console()->systemMessage (reconnectText);
    }
    if (!wantTabBar)
      tabbar->hide ();
    return true;
  }

  if (!dontClose)
  {
    //inform plug-ins
    pluginmanager->event (session[which], "tab-closed");
    //remove widget
    stack->removeWidget (stack->widget (which));
    delete session[which];
    session[which] = 0;
    count--;

    //update tab bar
    tabbar->removeTab (tabbar->tab (tabID[which]));

    //removeTab switches to some other tab but doesn't emit a signal (no idea why)
    //so I have to do this myself :(((
    int id = tabbar->currentTab ();
    changeSession (id);

    //hide the tabbar if needed
    if ((!wantTabBar) && (count <= 1))
      tabbar->hide ();
  }
  else
  {
    setSessionName (which, i18n ("Connection closed."));
    tabbar->tab (tabID[which])->setIconSet (iconNo);
    tabbar->update ();
    session[which]->telnet()->offLineMode();
    session[which]->connection()->setConnectionClosed (true);
    session[which]->connection()->updateMenus ();
    session[which]->console()->systemMessage (i18n ("Press ENTER to close this tab."));
    session[which]->console()->systemMessage (reconnectText);
  }

  return true;
}

void KMuddy::setSession (int which)
{
  if ((which < 0) || (which >= MAXSESSIONS))
    return;
  if (session[which] == 0)
    return;
  //this also invokes changeSession, which updates activesess
  tabbar->setCurrentTab (tabID[which]);

  //do not remove this line! It's usually useless, but it's required
  //for the program to work correctly on startup with hidden tabbar
  stack->raiseWidget (which);
}

int KMuddy::sessionId (QWidget *widget)
{
  //prevent crash on app close
  if (goingDown) return -1;

  return stack->id (widget);
}

int KMuddy::activeSession ()
{
  return activesess;
  //return stack->id (stack->visibleWidget ());
}

void KMuddy::changeSession (int id)
{
  //find session ID; a bit ineffective, but MAXSESSIONS is only 50,
  //so nobody will even notice the slowdown...
  int which = -1;
  for (int i = 0; i < MAXSESSIONS; i++)
    if (tabID[i] == id)
    {
      which = i;
      break;
    }
  if (which == -1)  //such ID does not exist
    return;

  //inform plug-ins
  pluginmanager->event (getSession (activesess), "deactivated");

  stack->raiseWidget (which);
  activesess = which;
  QTab *tab = tabbar->tab (tabID[which]);

  //disable flashing (if enabled)
  if (getSession (activeSession())->flashing())
  {
    getSession (activeSession())->setFlashing (false);

    //if flashing was on, icon could be wrong - FIX IT!
    if ((session[which] == 0) || (!session[which]->telnet()->isConnected()))
      tab->setIconSet (iconNo);
    else
      tab->setIconSet (iconOk);
    tabbar->update();
  }

  //inform plug-ins
  pluginmanager->event (getSession (activesess), "activated");

  //global caption
  /*QWidget::*/setPlainCaption (kapp->makeStdCaption (tab->text ()));

  //update menus (active items and such)
  getSession (activeSession())->connection()->updateMenus ();

  //update item in alias/trigger groups dialog
  cConnPrefs *cp = getSession (activeSession())->connection()->getConnPrefs();
  if (agroups != 0)
    agroups->switchGroupList ((cp == 0) ? 0 : cp->aliasGroupList());
  if (tgroups != 0)
    tgroups->switchGroupList ((cp == 0) ? 0 : cp->triggerGroupList());
  //the same for running scripts dialog
  if (rdlg != 0)
    rdlg->switchRunningList((cp == 0) ? 0 : cp->runningList());

  //still not enough; we also need to update the statistics window
  if (statdlg != 0)
    statdlg->update ();
  
  //give focus to the input line
  //cInputLine::focus[In/Out]Event also restores selections :)
  cInputLine *inputline = getSession (activeSession())->inputline();
  inputline->setFocus ();
}

void KMuddy::setSessionName (int which, QString name)
{
  getSession (which)->setName (name);
  QTab *tab = tabbar->tab (tabID[which]);
  tab->setText (name);
  //change icon to Ok (we'll change it later if it's not correct)
  tab->setIconSet (iconOk);
  tabbar->update ();
  if (which == activeSession ())
    /*QWidget::*/setPlainCaption (kapp->makeStdCaption (name));
}

void KMuddy::setAlwaysTabBar (bool value)
{
  if (tabbar == 0)
    return;
  wantTabBar = value;
  if (value)
    tabbar->show ();
  else
    if (count > 1)
      tabbar->show ();
    else
      tabbar->hide ();
}

void KMuddy::setAutoConnect (const QString &ac)
{
  autoconnect = ac;
}

void KMuddy::prevTab ()
{
  int idx = tabbar->indexOf (tabbar->currentTab ());
  idx--;
  if (idx < 0)
    idx = tabbar->count() - 1;
  QTab *tab = tabbar->tabAt (idx);
  if (tab != 0)
    tabbar->setCurrentTab (tab);
}

void KMuddy::nextTab ()
{
  int idx = tabbar->indexOf (tabbar->currentTab ());
  idx++;
  if (idx >= tabbar->count())
    idx = 0;
  QTab *tab = tabbar->tabAt (idx);
  if (tab != 0)
    tabbar->setCurrentTab (tab);
}

void KMuddy::switchTab (int index)
{
  QTab *tab = tabbar->tabAt (index);
  if (tab != 0)   //tab exists
    tabbar->setCurrentTab (tab);
}

void KMuddy::setAllowGlobalNotify (bool notify)
{
  globalnotify = notify;
}

void KMuddy::setAllowLocalNotify (bool notify)
{
  localnotify = notify;
}

void KMuddy::setAlwaysNotify (bool notify)
{
  alwaysnotify = notify;
}

void KMuddy::displayedSomeText (cSession *sess)
{
  //only if alwaysnotify is on...
  if (!alwaysnotify)
    return;

  requestNotify (sess);
}

void KMuddy::requestNotify (cSession *sess)
{
  //global notify
  if (globalnotify && (!isActiveWindow ()))
  {
    globalFlashingNOW = true;
    if (!(notifyTimer->isActive()))
      notifyTimer->start (FLASHTIME);
  }

  //local notify
  if (localnotify && (getSession (activeSession()) != sess))
  {
    sess->setFlashing (true);
    if (!(notifyTimer->isActive()))
      notifyTimer->start (FLASHTIME);
  }
}

void KMuddy::setMenuAliasesEnabled (bool val)
{
  switchaliases->setChecked (val);
}

void KMuddy::setMenuTriggersEnabled (bool val)
{
  switchtriggers->setChecked (val);
}

void KMuddy::setMenuTimersEnabled (bool val)
{
  switchtimers->setChecked (val);
}

void KMuddy::setMenuShortcutsEnabled (bool val)
{
  switchshortcuts->setChecked (val);
}

void KMuddy::doDisconnect ()
{
  getSession (activeSession())->connection()->handleDisconnect ();
}

void KMuddy::doReconnect ()
{
  getSession (activeSession())->connection()->reconnect ();
}

void KMuddy::sendCommandToActiveSession (const QString &command)
{
  getSession (activeSession())->sendcmd()->addCommand (command);
}

//THESE METHODS WERE ORIGINALLY IN cConnection; now they're here because
//of multi-tab support

void KMuddy::doQuickConnect ()
{
  //hide the dialog so that it doesn't look like we've crashed
  qdlg->hide ();

  //this object will be deleted in cConnection::disconnect()
  cConnectionInfo *i = new cConnectionInfo;
  i->setServer (qdlg->host());
  i->setPort (qdlg->port());

  //create cSession object and related stuff
  cSession *s = addSession ();

  //now establish that connection!
  s->connection()->establishQuickConnection (i);

  //if it failed
  if (!(s->connection()->isConnected()))
  {
    //remove that session!
    removeSession (activeSession());
    return;
  }
  
  //update item in alias/trigger groups dialog
  if (agroups != 0)
    agroups->switchGroupList (0);
  if (tgroups != 0)
    tgroups->switchGroupList (0);
  //the same for running scripts dialog
  if (rdlg != 0)
    rdlg->switchRunningList (0);
}

void KMuddy::doConnect ()
{
  //hide the dialog, so that it doesn't look like we've crashed
  cdlg->hide ();
  
  QString name = cdlg->selName();
  bool sendNothing = cdlg->sendNothing();

  doProfileConnect (name, sendNothing);
}

void KMuddy::doProfileConnect (const QString &name, bool sendNothing)  
{
  cProfiles *prof = new cProfiles;
    //this object will be deleted in cConnection::disconnect()
  cConnectionInfo *i;
  if (prof->exists ((QString &) name))
  {
    i = prof->getProfileConnInfo ((QString &) name);
    delete prof;
  }
  else
  {
    //this will only happen with auto-connecting to a deleted profile;
    //Connect dialog won't allow this
    KMessageBox::sorry (this, i18n ("Sorry, requested profile does not exist."), i18n ("Profile does not exist"));
    delete prof;
    return;
  }
    
  //create cSession object and related stuff
  cSession *s = addSession ();

  //now establish that connection!
  s->connection()->establishConnection (i, sendNothing);

  //if it failed
  if (!(s->connection()->isConnected()))
  {
    //remove that session!
    removeSession (activeSession());
    return;
  }

  //update item in alias/trigger groups dialog
  cConnPrefs *cp = s->connection()->getConnPrefs();
  if (cp != 0)
  {
    if (agroups != 0)
      agroups->switchGroupList (cp->aliasGroupList());
    if (tgroups != 0)
      tgroups->switchGroupList (cp->triggerGroupList());
    //the same for running scripts dialog
    if (rdlg != 0)
      rdlg->switchRunningList(cp->runningList());
  }

}

void KMuddy::showAndHandleConnectDialog ()
{
  //so first we have to create the dialog...
  cdlg = new dlgConnect (this);

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (cdlg, SIGNAL (okClicked()), this, SLOT (doConnect()));
  connect (cdlg, SIGNAL (addPressed()), this, SLOT (addPressed()));
  connect (cdlg, SIGNAL (modifyPressed()), this, SLOT (modifyPressed()));
  connect (cdlg, SIGNAL (deletePressed()), this, SLOT (deletePressed()));
  connect (cdlg, SIGNAL (duplicatePressed()), this, SLOT (duplicatePressed()));

  //dialog is ready - show it!
  cdlg->exec ();

  //further action is handled by signal issued by OK button, so that we only
  //have to destroy the dialog...
  delete cdlg;
}

void KMuddy::showAndHandleQuickConnectDialog ()
{
  //so first we have to create the dialog...
  qdlg = new dlgQuickConnect (this);

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (qdlg, SIGNAL (okClicked()), this, SLOT (doQuickConnect()));

  //dialog is ready - show it!
  qdlg->exec ();

  //further action is handled by signal issued by OK button, so that we only
  //have to destroy the dialog...
  delete qdlg;
}

void KMuddy::addPressed ()
{
  //so first we have to create the dialog...
  mdlg = new dlgEditProfile (true, this);

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (mdlg, SIGNAL (okClicked ()), this, SLOT (doAdd ()));

  //fill in default login sequence
  QStringList ls;
  ls.append ("$name");
  ls.append ("$password");
  mdlg->setConnectionString (ls);
  
  //dialog is ready - show it!
  mdlg->exec ();

  //further action is handled by signal issued by OK button, so that we only
  //have to destroy the dialog...
  delete mdlg;
}

void KMuddy::doAdd ()
{
  cProfiles *prof = new cProfiles;
  QString name = mdlg->name();
  name = name.simplifyWhiteSpace ();
  cConnectionInfo *cci = new cConnectionInfo;
  cci->setName(name);
  cci->setServer(mdlg->server());
  cci->setPort(mdlg->port());
  cci->setLogin(mdlg->login());
  cci->setPassword(mdlg->password());
  cci->setConnStr(mdlg->connectionString());
  if (prof->exists(name))   //profile already exists
  {
    if (KMessageBox::questionYesNoCancel (this,
        i18n("Profile named") + " " + name + " " + i18n ("already exists.") +
        "\n" + i18n ("Do you want to modify existing profile?"),
        i18n("Add profile")) != KMessageBox::Yes)
      return;
    //they want to modify existing profile, DO IT!
    prof->modifyProfile(name, cci);
  }
  else
    prof->addProfile(name, cci);
  delete prof;
  delete cci;

  cdlg->updateList ();
  cdlg->setSelectedItem (name);
}

void KMuddy::modifyPressed ()
{
  //fetch profile name
  QString name = cdlg->selName();
  if (name == QString::null)
    return;

  //so first we have to create the dialog...
  mdlg = new dlgEditProfile (false, this);

  //then we fill in the data
  cProfiles *prof = new cProfiles;
  cConnectionInfo *cci = prof->getProfileConnInfo (name);
  delete prof;
  mdlg->setName(cci->name());
  mdlg->setServer(cci->server());
  mdlg->setPort(cci->port());
  mdlg->setLogin(cci->login());
  mdlg->setPassword(cci->password());
  mdlg->setConnectionString(cci->connStr());
  delete cci;

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (mdlg, SIGNAL (okClicked ()), this, SLOT (doModify ()));

  //dialog is ready - show it!
  mdlg->exec ();

  //further action is handled by signal issued by OK button, so that we only
  //have to destroy the dialog...
  delete mdlg;
}

void KMuddy::doModify ()
{
  cProfiles *prof = new cProfiles;
  QString name = mdlg->name();
  name = name.simplifyWhiteSpace ();
  cConnectionInfo * cci = new cConnectionInfo;
  cci->setName(name);
  cci->setServer(mdlg->server());
  cci->setPort(mdlg->port());
  cci->setLogin(mdlg->login());
  cci->setPassword(mdlg->password());
  cci->setConnStr(mdlg->connectionString());
  prof->modifyProfile (name, cci);

  delete prof;
  delete cci;

  cdlg->updateList ();
}

void KMuddy::deletePressed ()
{
  //fetch profile name
  QString name = cdlg->selName();
  if (name == QString::null)
    return;

  //ask if he really wants to do this
  if (KMessageBox::questionYesNoCancel (cdlg,
        i18n ("Do you really want to delete profile " + name + " ?"),
        i18n ("Delete profile")) == KMessageBox::Yes)
  {
    //if he does, DO IT!!!
    cProfiles *prof = new cProfiles ();
    prof->deleteProfile(name);
      //maybe I'll should add some notification if we fail...
    delete prof;
    cdlg->updateList();
  }
}

void KMuddy::duplicatePressed ()
{
  //fetch profile name
  QString name = cdlg->selName();
  if (name == QString::null)
    return;

  //so first we have to ask for a new name...
  bool ok;
  QString newname = KLineEditDlg::getText (i18n ("Please enter name for the duplicated profile"), QString::null, &ok, cdlg);
  if (ok)
  {
    cProfiles *prof = new cProfiles;
    if (prof->exists (newname))
      KMessageBox::sorry (cdlg, i18n ("Sorry, you have to enter a profile name that does not exist."), i18n ("Profile already exists"));
    else
    {
      prof->duplicateProfile (name, newname);
      //maybe I'll should add some notification if we fail...
      delete prof;
      cdlg->updateList();
    }
  }
}

/** lots of slots that handle menus; were handled by slots in objects,
now we have cSession, so we have to handle it here */
void KMuddy::addSelectionToClipboard ()
{
  getSession (activeSession())->console()->addSelectionToClipboard();
}

void KMuddy::setParsing (bool value)
{
  for (int i = 0; i < MAXSESSIONS; i++)
    if (session[i] != NULL)
      session[i]->cmdproc()->setParsing(value);
}

void KMuddy::clearCommandQueue ()
{
  getSession (activeSession())->sendcmd()->clearCommandQueue();
}

void KMuddy::handleAliasesDialog ()
{
  getSession (activeSession())->connection()->handleAliasesDialog();
}

void KMuddy::handleTriggersDialog ()
{
  getSession (activeSession())->connection()->handleTriggersDialog();
}

void KMuddy::handleActionsDialog ()
{
  getSession (activeSession())->connection()->handleActionsDialog();
}

void KMuddy::handleScriptsDialog ()
{
  getSession (activeSession())->connection()->handleScriptsDialog();
}

void KMuddy::handleTimersDialog ()
{
  getSession (activeSession())->connection()->handleTimersDialog();
}

void KMuddy::handleShortcutsDialog ()
{
  getSession (activeSession())->connection()->handleShortcutsDialog();
}

void KMuddy::handleVariablesDialog ()
{
  getSession (activeSession())->connection()->handleVariablesDialog();
}

void KMuddy::handleVarTrigsDialog ()
{
  getSession (activeSession())->connection()->handleVarTrigsDialog();
}

void KMuddy::handleWindowsDialog ()
{
  getSession (activeSession())->connection()->handleWindowsDialog();
}

void KMuddy::switchAliases (bool value)
{
  getSession (activeSession())->connection()->switchAliases(value);
}

void KMuddy::switchTriggers (bool value)
{
  getSession (activeSession())->connection()->switchTriggers(value);
}

void KMuddy::switchTimers (bool value)
{
  getSession (activeSession())->connection()->switchTimers(value);
}

void KMuddy::switchShortcuts (bool value)
{
  getSession (activeSession())->connection()->switchShortcuts(value);
}

void KMuddy::showConnPrefsDialog ()
{
  getSession (activeSession())->connection()->showConnPrefsDialog();
}

void KMuddy::configureTranscript ()
{
  getSession (activeSession())->transcript()->configure();
}

void KMuddy::dumpBuffer ()
{
  getSession (activeSession())->transcript()->dumpBuffer();
}

void KMuddy::makeDecision ()
{
  getSession (activeSession())->console()->makeDecision();
}

void KMuddy::importProfile ()
{
  dlgImportExportProfile *iep = new dlgImportExportProfile (true, this);
  iep->doThings ();
  delete iep;
}

void KMuddy::exportProfile ()
{
  dlgImportExportProfile *iep = new dlgImportExportProfile (false, this);
  iep->doThings ();
  delete iep;
}

void KMuddy::lineUp ()
{
  getSession (activeSession())->console()->lineUp();
}

void KMuddy::lineDown ()
{
  getSession (activeSession())->console()->lineDown();
}

void KMuddy::pageUp ()
{
  getSession (activeSession())->console()->pageUp();
}

void KMuddy::pageDown ()
{
  getSession (activeSession())->console()->pageDown();
}

void KMuddy::aconUp ()
{
  getSession (activeSession())->console()->aconUp();
}

void KMuddy::aconDown ()
{
  getSession (activeSession())->console()->aconDown();
}

bool KMuddy::eventFilter (QObject *o, QEvent *e)
{
  //only handle keypresses, nothing else...
  if (e->type() == QEvent::KeyPress)
  {
    bool report = true;
    QKeyEvent *qke = (QKeyEvent *) e;
    int key = qke->key ();
    int state = qke->state ();

    //do not report modifiers
    switch (key) {
      case Qt::Key_Shift:
      case Qt::Key_Control:
      case Qt::Key_Meta:
      case Qt::Key_Alt:
        report = false;
        break;
      default:
        report = true;
        break;
    };
    
    if (report)
    {
      if (grabdlg)
      {
        grabdlg->gotKey (key, state);
        return true;
      }
      else
      {
        cSession *s = getSession (activeSession());
        if (s && (s->connection()))
        {
          cConnPrefs *cp = s->connection()->getConnPrefs();
          if (cp && (cp->shortcutList()))
          {
            if (cp->shortcutList()->handleKey (key, state))
              return true;
          }
        }
      }
    }
  }
  return KMainWindow::eventFilter (o, e);
}

bool KMuddy::event (QEvent *e)
{
  //window activated - stop the flash!
  if (e->type() == QEvent::WindowActivate)
  {
    if (globalFlashingNOW)
    {
      //if global flashing was active, then app title may be wrong - FIX IT!
      QString nn = getSession (activeSession())->name();
      /*QWidget::*/setPlainCaption (kapp->makeStdCaption (nn));
    }
    globalFlashingNOW = false;
  }
  //call inherited function
  return KMainWindow::event (e);
}

void KMuddy::handleFlashing ()
{
  flashVisible = flashVisible ? false : true;
  //global flashing
  if (kapp->activeWindow() != 0)
    globalFlashingNOW = false;
  if (globalFlashingNOW)
  {
    //construct new title!
    QString nn = getSession (activeSession())->name();
    //I have to keep prepending these chars, because the Modified icon is
    //not shown if you use a translated version of KDE and an English version
    //of KMuddy; only solution would probably be to translate KMuddy to every
    //language in which KDE is shipped :-(
    if (flashVisible)
      nn.prepend ("*** ");
    /*QWidget::*/setPlainCaption (kapp->makeStdCaption (nn, true, flashVisible));
  }

  //local flashing
  for (int i = 0; i < MAXSESSIONS; i++)
    if ((session[i] != 0) && (session[i]->flashing()))
    {
      QTab *tab = tabbar->tab (tabID[i]);
      //flashing could occur because a goodbye msg may have come
      bool connClosed = session[i]->connection()->connectionClosed();
      if (flashVisible)
        tab->setIconSet(iconFlag);
      else
        tab->setIconSet(connClosed ? iconNo : iconOk);
      tabbar->update();
    }
}

bool KMuddy::queryClose ()
{
  for ( int i = 0; i < MAXSESSIONS; i++ ) {
    if ( session[i] ) {
      //switch to that session, so that he knows which one is about to close
      setSession (i);
      kapp->processEvents ();
      //disconnect, asking the user whether he really wants to
      if ( !session[i]->connection()->handleDisconnect() ) {
        return false;
      }
    }
  }

  return true;
}
